#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

using namespace QSAS;

extern "C" QplugReturnStatus CharacteristicFrequencies(QplugArgList *);

QplugReturnStatus CharacteristicFrequencies(QplugArgList *call_list)
{
    bool hasB=false;
    bool hasNe=false;
    
    QdObject_var temp=(*call_list)[0];
    
    QdRScalarSeq_var B=QdRScalarSeq_var::narrow(temp);
    QdRScalarSeq_var B_Tesla;
            
    if(B.is_nil())
    {
        QplugAppendTextDisplay("Warning: B is not a scalar sequence - considering B is unknown.");
    }
    else
    {
        hasB=true;
    }
    
    temp=(*call_list)[1];
    
    QdRScalarSeq_var Ne=QdRScalarSeq_var::narrow(temp);
    QdRScalarSeq_var Ne_SI;
    
    if(Ne.is_nil())
    {
        QplugAppendTextDisplay("Warning: Ne is not a scalar sequence - considering Ne is unknown.");
    }
    else
    {
        hasNe=true;
    }
    
    if(!hasB && !hasNe)
    {
        QplugAppendTextDisplay("No data available: returning.");
        return QPLUG_FAILURE;
    }
    
    if(hasB && hasNe)
    {
        int qar_status=QarTestConformal(B, Ne);
        if(qar_status==QAR_ERR)
        {
            QplugAppendTextDisplay("Please join input time series.");
            return QPLUG_FAILURE;
        }
    }

    //Check units
    
    char * units_SI;
    if(hasB)
    {
        units_SI=QuGetAttrText("SI_conversion", B);
        if (strcmp(units_SI, "") == 0)
        {
            QplugAppendTextDisplay("Problem checking input units.");
            return QPLUG_FAILURE;
        }
        
        temp=QarConvertUnitsTo(B, "1 > kg s^-2 A^-1");
        B_Tesla=QdRScalarSeq_var::narrow(temp);
        
        if(B_Tesla.is_nil())
        {
            QplugAppendTextDisplay("Conversion to Tesla failed.");
            return QPLUG_FAILURE;
        }
    }
    
    if(hasNe)
    {
        units_SI=QuGetAttrText("SI_conversion", Ne);
        if (strcmp(units_SI, "") == 0)
        {
            QplugAppendTextDisplay("Problem checking input units.");
            return QPLUG_FAILURE;
        }
        
        temp=QarConvertUnitsTo(Ne, "1 > m^-3");
        Ne_SI=QdRScalarSeq_var::narrow(temp);
        
        if(Ne_SI.is_nil())
        {
            QplugAppendTextDisplay("Conversion to m^-3 failed.");
            return QPLUG_FAILURE;
        }
    }
    
    // Constants
    
    QdObject_var q_e; double q_e_value;
    QdObject_var m_e; double m_e_value;
    QdObject_var m_p; double m_p_value;
    QdObject_var eps_0; double eps_0_value;
    
    QdObject_var pi_2; double pi_value;
    pi_value = (double)  4.0 *atan(1.0);
    pi_2  = new QdRScalar (2.*pi_value);
    
  // Electron Charge
    q_e_value = (double)  1.60210e-19;
    q_e = new QdRScalar((double) q_e_value);
    QuSetTxtAttr("SI_conversion", "1>C", q_e);
    QuSetTxtAttr("FIELDNAM", "Electron charge", q_e);
    QuSetFrame( q_e, "scalar>na");
    QuSetTxtAttr("UNITS", "C", q_e);
    
 // electron mass
    m_e_value = (double)  9.1091e-31;
    m_e = new QdRScalar((double) m_e_value);
    QuSetTxtAttr("SI_conversion", "1.0E-3>g", m_e);
    QuSetTxtAttr("FIELDNAM", "mass of electron", m_e);
    QuSetFrame( m_e, "scalar>na");
    QuSetTxtAttr("UNITS", "kg", m_e);
 
  //  proton mass
    m_p_value = (double)  1.6726e-27;
    m_p = new QdRScalar((double) m_p_value);
    QuSetTxtAttr("SI_conversion", "1.0E-3>g", m_p);
    QuSetTxtAttr("FIELDNAM", "mass of proton", m_p);
    QuSetFrame( m_p, "scalar>na");
    QuSetTxtAttr("UNITS", "kg", m_p);
    
  //  epsilon 0
    eps_0_value = (double)  8.854187817e-12;
    eps_0 = new QdRScalar((double) eps_0_value);
    QuSetTxtAttr("SI_conversion", "1>C^2 J^-1 m", eps_0);
    QuSetTxtAttr("FIELDNAM", "permittivity of vacuum", eps_0);
    QuSetFrame( eps_0, "scalar>na");
    QuSetTxtAttr("UNITS", "C^2 J^-1 m", eps_0);
    
    // Outputs

    QdRScalarSeq_var wce;
    QdRScalarSeq_var wci;
    QdRScalarSeq_var wpe;
    
    QdRScalarSeq_var out_fce;
    QdRScalarSeq_var out_fci;
    QdRScalarSeq_var out_fpe;
    QdRScalarSeq_var out_fuh;
    QdRScalarSeq_var out_flh;
    QdRScalarSeq_var out_fl;
    QdRScalarSeq_var out_fr;
    char str_out[4000];
    
    int num=(hasB? B->sequence_size() : Ne->sequence_size());
    QdRScalarSeq foo(num, -1);
    
    
    try{
        if(hasB)
        {
            wce = QdRScalarSeq_var::narrow(q_e*B_Tesla/m_e);
            wci = QdRScalarSeq_var::narrow(q_e*B_Tesla/m_p);
            
            out_fce=QdRScalarSeq_var::narrow(wce/pi_2);
            out_fci=QdRScalarSeq_var::narrow(wci/pi_2);
        }
        
        if(hasNe)
        {
            wpe = QdRScalarSeq_var::narrow(sqrt(Ne_SI*q_e*q_e/(m_e*eps_0)));
            out_fpe=QdRScalarSeq_var::narrow(wpe/pi_2);
        }
        
        if(hasNe && hasB)
        {
            out_fuh=QdRScalarSeq_var::narrow(sqrt(wpe*wpe+wce*wce)/pi_2);
            out_flh=QdRScalarSeq_var::narrow(sqrt((wpe*wpe+wce*wci)/(wpe*wpe+wce*wce)*wce*wci)/pi_2);
            out_fl=QdRScalarSeq_var::narrow((-wce+sqrt(wce*wce+new QdRScalar(4.)*wpe*wpe))/pi_2/new QdRScalar(2.));
            out_fr=QdRScalarSeq_var::narrow((wce+sqrt(wce*wce+new QdRScalar(4.)*wpe*wpe))/pi_2/new QdRScalar(2.));
        }
        
    }
    catch(Exception &e){
        e.print_msg();
        return QPLUG_FAILURE;
    }
    
    if(hasB)
    {
        out_fce->copy_xrefs_from(B);
        out_fci->copy_xrefs_from(B);
        if(hasNe)
        {
            out_fpe->copy_xrefs_from(B);
            out_fuh->copy_xrefs_from(B);
            out_flh->copy_xrefs_from(B);
            out_fl->copy_xrefs_from(B);
            out_fr->copy_xrefs_from(B);
        }
    }
    else
    {
        out_fpe->copy_xrefs_from(Ne);
    }
    
    if(hasB)
    {
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_fce);
        sprintf(str_out,"f_ce (from %s)", QuGetAttrText("FIELDNAM", B));
        QuSetTxtAttr("LABLAXIS", "f_ce", (QdObject_var) out_fce);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_fce);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_fce);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_fce);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_fce);
    
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_fci);
        sprintf(str_out,"f_ci (from %s)", QuGetAttrText("FIELDNAM", B));
        QuSetTxtAttr("LABLAXIS", "f_ci", (QdObject_var) out_fci);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_fci);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_fci);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_fci);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_fci);
    }
    
    if(hasNe)
    {
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_fpe);
        sprintf(str_out,"f_pe (from %s)", QuGetAttrText("FIELDNAM", Ne));
        QuSetTxtAttr("LABLAXIS", "f_pe", (QdObject_var) out_fpe);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_fpe);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_fpe);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_fpe);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_fpe);
    }

    if(hasB && hasNe)
    {
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_fuh);
        sprintf(str_out,"f_uh (from %s and %s)", QuGetAttrText("FIELDNAM", Ne), QuGetAttrText("FIELDNAM", B));
        QuSetTxtAttr("LABLAXIS", "f_uh", (QdObject_var) out_fuh);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_fuh);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_fuh);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_fuh);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_fuh);
    
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_flh);
        sprintf(str_out,"f_lh (from %s and %s)", QuGetAttrText("FIELDNAM", Ne), QuGetAttrText("FIELDNAM", B));
        QuSetTxtAttr("LABLAXIS", "f_lh", (QdObject_var) out_flh);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_flh);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_flh);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_flh);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_flh);
        
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_fl);
        sprintf(str_out,"f_l (from %s and %s)", QuGetAttrText("FIELDNAM", Ne), QuGetAttrText("FIELDNAM", B));
        QuSetTxtAttr("LABLAXIS", "f_l", (QdObject_var) out_fl);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_fl);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_fl);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_fl);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_fl);
    
        QuSetTxtAttr("SI_conversion", "1>Hz", (QdObject_var) out_fr);
        sprintf(str_out,"f_r (from %s and %s)", QuGetAttrText("FIELDNAM", Ne), QuGetAttrText("FIELDNAM", B));
        QuSetTxtAttr("LABLAXIS", "f_r", (QdObject_var) out_fr);
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) out_fr);
        QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_fr);
        QuSetTxtAttr("Frame", "scalar>na", (QdObject_var) out_fr);
        QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_fr);
    }
    

    call_list->push_back((QdObject_var) out_fce);
    call_list->push_back((QdObject_var) out_fci);
    call_list->push_back((QdObject_var) out_fpe);
    call_list->push_back((QdObject_var) out_fuh);
    call_list->push_back((QdObject_var) out_flh);
    call_list->push_back((QdObject_var) out_fl);
    call_list->push_back((QdObject_var) out_fr);
        
    return QPLUG_SUCCESS;

    
}
