/**********************************************************************
* SCCS id: @(#)IndexTSSubset.c	1.1 14:32:59 12/17/98 
***********************************************************************
* PROJECT:   Cluster UK CDHF
* COMPONENT: TS Subsetting
* MODULE:   IndexTSSubset.c
* LANGUAGE:  C  
* PURPOSE:    Find time tag array locations for tstart and tend
* DESCRIPTION:  perform bi-section search
*
* Arguments:
*   IN
*       double tstart,           // input start and end times for subset interval
*       double tend, 
*       double *in_tags,         // pointer to input time tags
*       int ndata,               // input number of data points
*
*   OUT
*       int *istart_p,              // output indices of start and end tags
*       int *iend_p,
*       char *out_message          //  string containing error messages, etc.
* Modules Used:   nle_bisection  C routine
*
* HISTORY: SJ Schwartz and Gareth Chisham, Oct 97
*/

/* *************************************************************************** */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "IndexTSSubset.h"

IndexTSSubsetErrorStatus IndexTSSubset (
        double tstart,       /* input start and end times for subset interval*/
        double tend, 
        double *in_tags,     /* pointer to input time tags  */
        int ndata,           /* input number of data points */
        int *istart_p,       /* output indices of start and end tags */
        int *iend_p,
        char *out_message
        ){


IndexTSSubsetErrorStatus	okstart, okend;
int istart, iend;

if(tstart > tend){              /* start after end time; bale out now! */
	sprintf(out_message, "\nIndexStartEnd>> start = %d,  end= %d\n",
            tstart,tend);
    return INDEX_TSSUBSET_ERROR_START_GT_DATA_END; 
}

istart = nle_bisection (tstart, in_tags, ndata); /* Find index to start tag */

switch (istart){
  case ( -2 ): 
      okstart= INDEX_TSSUBSET_ERROR_START_GT_DATA_END;    
                                  /* requested start after end time  */
      break;                      /*               of input data */
  case ( -1 ):                    /* requested start before beginning; */
      istart=0; 		  /* set to earliest time in data */
      okstart=INDEX_TSSUBSET_OK;
      break;           
  default:  			  /* drop through by default */
      if (istart >= 0){
         if (in_tags[istart] < tstart ) istart++; /* incr so 1st pt >=tstart */
         okstart=INDEX_TSSUBSET_OK;
      } else {
         okstart=INDEX_TSSUBSET_ERROR_START_UNDEFINED;
      }
      break;                
}                      /* end switch (istart) */

iend = nle_bisection (tend, in_tags, ndata);  /* Find index to end time */
                                  /* index to nearest t < tend */

switch (iend){
  case ( -2 ):                    /* requested stop after end time     */
                                  /* of input data; set to latest time */
      iend=ndata-1; 
      okend=INDEX_TSSUBSET_OK;
      break;
  case ( -1 ):                    /* requested stop before beginning */ 
      okend= INDEX_TSSUBSET_ERROR_END_LT_DATA_START;
      break;
  default: 
      if (iend >= 0){            /* check if one more pt is within range */
				 /* (nle_bisection returns lower index   */
				 /* when bisectioning ends at pair)      */
         if ( (iend < ndata-1) && (in_tags[iend+1]<=tend) )iend++;  
         okend=INDEX_TSSUBSET_OK;
      } else {
         okend=INDEX_TSSUBSET_ERROR_END_UNDEFINED;
      }
      break;                      /*  t(iend) will be less than tend */
}                                 /* end switch (iend) */

if ( (okstart !=INDEX_TSSUBSET_OK) || (okend !=INDEX_TSSUBSET_OK)){
    sprintf (out_message, "\nIndexTSSubset>> ERROR:\n%s\n%s\n", 
                          IndexTSSubsetErrorMessage[okstart], 
                          IndexTSSubsetErrorMessage[okend]);
    return INDEX_TSSUBSET_ERROR_RETURN;
}  /* end if  */
if (iend < istart ){
    sprintf (out_message, 
      "IndexTSSubset>> ERROR - end index < start! (no data in interval)\n");
    return INDEX_TSSUBSET_ERROR_NO_DATA;
}
sprintf (out_message, "IndexTSSubset >> Success: istart, iend = %i, %i",istart,iend);
*istart_p = istart;
*iend_p = iend;
return INDEX_TSSUBSET_OK;
} 


/* ***********************************************************************
*********** nle_bisection *********************************************/
/* PROJECT:   Cluster UK CDHF
* COMPONENT: Boxcar (sliding average)
* MODULE:    bi-section search on monotonic array of doubles
*            function nle_bisection (double xsearch, double x[], int nx)
* LANGUAGE:  C  
* PURPOSE:   Locate, e.g., array index corresponding to given time, e.g.,
*            for sub-setting
* DESCRIPTION: Bi-section search, finding index of nearest element lower in
*              value than the search value. For descending series, this is
*              the element with the higher index.
* ARGUMENTS: 
* INPUTS:
*    xsearch      search value - must lie within range x[0] to x[nx-1]
*    x[]          array to be searched - must be monotonic
*    nx           number of elements in array to be search
* OUTPUTS:    only return value
* GLOBALS:    None
* RETURN:
*    TYPE:    int
*    VALUES:
*      on success    array subscript for value x[i] where x[i] <= xsearch
*      Error         -1 if xsearch < min(x)
*      Error         -2 if xsearch > max (x)
* FUNCTIONS USED: None
* ALGORITHM: 
*   Simple bisection. NO CHECK is made for non-MONOTONIC arrays.
*   Such checks should be done by calling routine
* HISTORY: SJ Schwartz February 1996
*/

int nle_bisection (double xsearch, double x[], int nx) {

int low, mid, high, ascnd;
if  (nx < 1) nx=1;                    /* catch silly input error and correct */
ascnd = ( x[nx-1] > x[0]);            /* true if series is in ascending order */
if (ascnd){
  if (x[0] > xsearch) return -1;      /* error if xsearch < lowest value */
  if (x[nx-1] < xsearch) return -2;   /* error if xsearch > highest value  */
}
else {               /* cope with descending series  */
  if (x[0] < xsearch) return -2;      /* error if xsearch > highest value */
  if (x[nx-1] > xsearch) return -1;   /* error if xsearch < lowest value  */
}        /* end if  */
/* so xsearch lies within range spanned by x[] */


low=0;
high = nx-1;
mid = high/2;

while (low < high -1) {
  if ( ( x[mid] < xsearch ) == ascnd )    /* to cope with monotonic decrease*/ 
    low = mid;
  else 
    high = mid;
  mid = ( high + low ) /2 ;
}                            /* end while */
if (ascnd) return low;       /* successful completion */
else return high;
}                            /* end function nle_bisection  */

