// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for Massflow
//
// Inputs :
//     B = magnetic field  (n x 3)     [nT]
//     V = plasma velocity (n x 3)     [km/s]
//     N = plasma density  (n)         [1/cm^3]  (proton mass used to get mass density)
//
// For simplification, all units are converted to SI units
//
//
// SEH, Mar 27, 2003
//      Sep 12, 2003 - modified for use with the GenericResidue plugin
//
// -----------------------------------------------------------------------------------------

#include "Qdos.h"
#include "qplug_if.h"
#include "qdutil.h"
#include "GenericResidue.h"

QdRMatrix_var Q_MMR_calc(QdRMatrixSeq_var V, QdRScalarSeq_var N, QdRMatrix_var U, int nestsize)
{
 int i,j,k;
 double ave_eta=0.0, d_eta2=0.00;
 double normfac = (double) nestsize;
 double m_p = 1.67e-27;                     // mass of proton [kg] )we assume ions are == protons)
 bool PO = false;                          // PrintOut to screen ? 

 QdRScalarSeq_var eta =  new QdRScalarSeq(nestsize);
 QdRMatrixSeq_var q_i = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 QdRMatrix_var    ave_q_i = new QdRMatrix(Dim(3), 0.0);
 QdRMatrix_var Q  = new QdRMatrix(Dim(3,3), 0.0);


 // --- construct q_i and eta -----

 for (k=0; k<nestsize; k++) {
//    m_p = 1.67e-27;
//    if (k > 10) m_p *= 2.0;   
   (*eta)[k] = m_p * (*N)[k] * 1E6; 
   ave_eta += (double) (*eta)[k];  
   for (i=0; i<3; i++) {
     (*q_i)[k][i] = (*eta)[k] * (*V)[k][i] * 1E3;
     (*ave_q_i)[i] += (*q_i)[k][i];
   }
 }


  // ----- initialize U, get <q_ij> and <eta>--------

  for (i=0; i<3; i++) {
    (*ave_q_i)[i] /= normfac;
    (*U)[i] = 0.0;
  }
  ave_eta /= normfac;

  if (PO) printf("In Q_MMR:\nq_i vector [kg/(m^2 s)] : \n");
  if (PO) printf("<q_j>     = [%E  %E  %E]; \n",(double)(*ave_q_i)[0], (double)(*ave_q_i)[1], (double)(*ave_q_i)[2]);
  if (PO) printf("\n<eta>      = %E [kg/m^3]\n", ave_eta);
  if (PO) fflush(stdout);

  // ---   get delta_q_i = q_i - <q_i> and delta_eta --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) (*q_i)[k][i] -= (*ave_q_i)[i];
    (*eta)[k] -= ave_eta;
    d_eta2 += ((double) (*eta)[k] * (double) (*eta)[k]);
  }

 
  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }

  // ---- construct U* vector ---------------

  if (d_eta2 != 0.0) {
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) (*U)[i] += (double)(*eta)[k] * (double)(*q_i)[k][i] / d_eta2; 
      }  
  } else {
     printf("MMR: singularity - setting U* = <V_cis>\n");
	 fflush(stdout);
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) (*U)[i] +=  (double)(*V)[k][i] * 1E3 / normfac;       
      }
  }

  if (PO)
  {
	printf("\nU*         = [%E  %E  %E];  [m/s] (eq 12 in manuscript)\n\n", (double)(*U)[0], (double)(*U)[1], (double)(*U)[2]);
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------  

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	(*Q)[i][j] += ((*q_i)[k][i] - (*U)[i]*(*eta)[k]) * ((*q_i)[k][j] - (*U)[j]*(*eta)[k]);
      }
    }
  }
  

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      (*Q)[i][j] /= normfac;
    }
  }

  if (PO)  printf("Q_ij  [kg/(s^3)] (eq 15 in manuscript)\n");
  if (PO)  for (i=0; i<3; i++) {
    printf("Q_i%d       = [%E  %E  %E];\n",i, (double)(*Q)[i][0], (double)(*Q)[i][1], (double)(*Q)[i][2]);
	fflush(stdout);
  } 
 
 return(Q);
}

