// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Solar Wind Parameter calculator
// 
// This plugin takes a magnetic field vector and plasma data as inputs
// and calculates the solar wind epsilon paramtere, accumulated
// energy, dynamic pressure and the magnetopause standoff distance.
//
// Inputs :
// 	bfield		- vector time series with magnetic field
// 	velocity	- time series with plasma velocity
// 	density	        - scalar time series with plasma density
// 	timeinterval	- time interval
//
// Outputs
//      epsilon parameter - scalar time series with the Akasofu epsilon param
//      energy   - scalar time series with accumulated energy input
//      pdyn  - scalar time series solar wind dynamic pressure
//      Rst  - scalar time series the Magnetopayuse standoff distance
//      clockangle  - Bimf clock angle (0 deg = northward, 180 deg = southward)
//
// SEH, 03 Mar, 2002
//      25 Nov, 2002 - allowed for scalar Vsw + clock angle output.
//      09 Dec, 2002 - V may be vector series, scalarseries or constant
//                     and n may be scalar series or constant.
//      10 Dec, 2002 - allow for time shift + corrected bug when changing time
//      11 Dec, 2002 - V_bulk (shifted) as output 
//      27 Oct, 2004 - B_IMF (shifted) as output + variable time delay option
// AJA  03 Oct, 2005 - Added to QSAS distribution and defensive programing and units conversion added
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"


extern "C" QplugReturnStatus  ISSI_Solarwind  ( QplugArgList *);

typedef struct VEC {double x,y,z;} vec;
double vecsum(double x, double y, double z);
float sqr(float x);

using namespace QSAS;

QplugReturnStatus  ISSI_Solarwind  ( QplugArgList *call_list )

{

 QdRMatrixSeq_var  bfield, ts_inV;
 QdRScalarSeq_var  epsilon, ts_inS;
 QdRScalarSeq_var  energy, vout;
 QdRScalarSeq_var  pdyn;
 QdRScalarSeq_var  rstandoff;
 QdRScalarSeq_var  clockangle;
 QdRMatrixSeq_var  bout;

// -- variables ----
// ionmass = 1.67E-6 = 1.67*1E-27 [kg] *(1E6) * (1E9) * (1E6)
//	 ...1E6 = 1/ccm -> 1/m^3 * ...
//       ...1E9 = Pa -> nPa
//       ...1E6 = (km/s)^2 -> (m/s)^2

 long    n_vecs,nestsize;
 double  L0, btot2,theta,deltat,avg_width,Vsw,TotalEnergy=0.0,Bt;
 double  ionMass= 1.67E-6;                                         
 int     nseq =3, triage_cnt=0;
 vec B,V;
 char    txtbuf[200];

 
 // ---- read other inputs ----

 QdTimeSeq_var tt_B = get_timetags((*call_list)[0]);
 if (tt_B.is_nil()) {
   QplugAppendTextDisplay("Bfield has no time tags\n");
   return(QPLUG_FAILURE);
 }
 long nb = tt_B->sequence_size();
 
 bfield = new QdRMatrixSeq(nb,RMatrix(Dim(3)));

 QdRMatrixSeq_var b_in = QdRMatrixSeq_var::narrow((*call_list)[0]);
 if (b_in.is_nil()) {
   QplugAppendTextDisplay("Bfield wrong or not vector time series\n");
   return(QPLUG_FAILURE);
 }
 
 bfield = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(b_in, "1.0e-9>T")); // ensure units are in Cluster units assumed
 if(bfield.is_nil()){
   bfield = new QdRMatrixSeq(nb,RMatrix(Dim(3)));
   // ACE does not have CSDS/CAA attributes needed by QarConvertUnitsTo()
   if (strcmp(QuGetAttrText("UNITS", b_in),"0.1nT") == 0) {
     for (int i=0;i< tt_B->sequence_size(); i++) {
       for (int j=0; j<3; j++) {
         bfield[i][j] = 0.1 * (float)(*b_in)[i][j];
       }
     }
   }  
   else {
     for (int i=0;i< tt_B->sequence_size(); i++) {
       for (int j=0; j<3; j++) {
         bfield[i][j] = (float)(*b_in)[i][j];
       }
     }
   }
 }
 set_timetags(bfield,tt_B);


 //  --- check if V is vector -----
 
 QdRScalarSeq_var ts_in;
 ts_inV = QdRMatrixSeq_var::narrow((*call_list)[1]);
 
 if ( !ts_inV.is_nil() ) {
    QplugAppendTextDisplay("Velocity is vector .. calculating |V|...\n");
    QdTimeSeq_var tt_V = get_timetags(ts_inV);
    if (tt_V.is_nil()) {
      QplugAppendTextDisplay("Velocity has no time tags\n");
      return(QPLUG_FAILURE);
    
    }
    ts_in = new QdRScalarSeq(ts_inV->sequence_size());
    set_timetags(ts_in,tt_V);
    for (int i=0; i<ts_inV->sequence_size(); i++)  {
      V.x = (double) ts_inV[i][0];
      V.y = (double) ts_inV[i][1];
      V.z = (double) ts_inV[i][2];     
      ts_in[i] = vecsum(V.x,V.y,V.z);
    } 
    ts_in->copy_xrefs_from((* call_list)[1] );
 }
 else  {
   ts_in = QdRScalarSeq_var::narrow((*call_list)[1]);
   if( !ts_in.is_nil()){
      QplugAppendTextDisplay("Velocity as scalar series\n");
   } 
   else {
     QdRScalar_var vsw =  QdRScalar_var::narrow((*call_list)[1]);
	if(!vsw.is_nil()){
       ts_in = new QdRScalarSeq(nb);
       for (int i=0; i<nb; i++) ts_in[i] = fabs((float) (*vsw));
       ts_in->copy_xrefs_from((* call_list)[1] );
       set_timetags(ts_in,tt_B);  // has none of its own
       sprintf(txtbuf,"NB! Velocity is constant %7.2f\n",(float) (*vsw));
       QplugAppendTextDisplay(txtbuf);
	}
   }
 }
   
 if(ts_in.is_nil()) {
    QplugAppendTextDisplay("Cannot find valid Velocity\n");
    return(QPLUG_FAILURE);
 }
 // Now convert to required units if possible

 ts_inS =  QdRScalarSeq_var::narrow(QarConvertUnitsTo(ts_in, "1.0e3>m s^-1")); // ensure units are in Cluster units assumed
 if(ts_inS.is_nil()){
   // case when SIconversion not found
   QplugAppendTextDisplay("Velocity has no SI_conversion, assumed in Km/s\n");
   ts_inS = ts_in;
 }
 
 // ---- check density. If not scalar series -> use given input -----
 
 QdRScalarSeq_var n_in = QdRScalarSeq_var::narrow((*call_list)[2]);
 if (!n_in.is_nil()) {
   printf("OK - density is scalar time series\n");
   fflush(stdout);
 } 
 else {
   QdRScalar_var nsw = QdRScalar_var::narrow((*call_list)[2]);
   if( !nsw.is_nil()){
     printf("OK - density is given constant\n");
	 fflush(stdout);
     n_in = new QdRScalarSeq(nb);
     printf("%d\n",n_in->sequence_size());
	 fflush(stdout);
     for (int i=0; i<nb; i++) n_in[i] = fabs((float) (*nsw));
     n_in->copy_xrefs_from((* call_list)[2] );
     set_timetags(n_in,tt_B);
     sprintf(txtbuf,"NB! Velocity is constant %7.2f\n",(float) (*nsw));
     QplugAppendTextDisplay(txtbuf);
   }
   else {
    QplugAppendTextDisplay("Cannot find valid density\n");
    return(QPLUG_FAILURE);
   }
 }
  

 QdRScalarSeq_var n_inS = QdRScalarSeq_var::narrow(QarConvertUnitsTo(n_in, "1.0e6>m^-3")); // ensure units are per cc
 if(n_inS.is_nil()){
   // case when SIconversion not found
   QplugAppendTextDisplay("Density has no SI_conversion, assumed in cm^-3\n");
   n_inS = n_in;
 }
 //  --- first we join B measurements to plasma measurements -----
 QdObjectSeq_var ttIn    = new QdObjectSeq(nseq);
 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 QdObjectSeq_var restSig = new QdObjectSeq(nseq);
 QdObjectSeq_var joinMethod = new QdObjectSeq(nseq);
 QdObjectSeq_var gapValue   = new QdObjectSeq(nseq);
 QdObjectSeq_var gapHandler = new QdObjectSeq(nseq);
 QdRMatrixSeq_var  bfield_nest;
 QdRScalarSeq_var  velocity_nest;                            // V = scalar series at this stage
 QdRScalarSeq_var  density_nest;
 QdTimeSeq_var ttJoined = tt_B;                             // tt of joined N,V and B
 QdTimeSeq_var tt_out;                                      // tt of masked N,V and B


 QuIsTSRegular(ttJoined, &deltat); // need spacing but not regular tt
 avg_width = 1.5 * deltat;
 printf("delta t = %f  ->  gap_fill if t > %f\n",deltat,avg_width);
 fflush(stdout);

 toJoin[0] = bfield; //((*call_list)[0]);
 toJoin[1] = ts_inS;
 toJoin[2] = n_inS;
 ttIn[0] = tt_B;
 ttIn[1] = get_timetags(ts_inS);
 ttIn[2] = get_timetags(n_inS);

 for (int i=0; i<nseq; i++) {
   //   toJoin[i] = ((*call_list)[i]);
   //   ttIn[i] = get_timetags((*call_list)[i]);
   joined[i]  = toJoin[i]->create();

   joinMethod[i] = new QdString(LINEAR);
   gapValue[i] = new QdRScalar(avg_width);
   gapHandler[i] = new QdString(LINEAR);   
 }

 try {
   printf("%f\n",avg_width);
   fflush(stdout);
   multijoin(ttIn, toJoin, ttJoined, joined, 
      joinMethod, gapValue, gapHandler, triage_cnt);
   printf("%f\n",avg_width);
   fflush(stdout);
 }
 catch(Exception &e) {
   e.print_msg();
   printf("%f\n",avg_width);
   fflush(stdout);
   return QPLUG_FAILURE;
 }

 //  ---- get Time interval and mask ----------------

 QdTimeInterval_var start_end = get_timeinterval((* call_list)[3]);
 if(start_end.is_nil()){
    QplugAppendTextDisplay("Cannot find time interval input\n");
    return(QPLUG_FAILURE);
 }
 SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 tt_out = new QdTimeSeq();
 ttJoined->get_subsequence(*tt_out, msk);
 if (tt_out.is_nil()) {    
   QplugAppendTextDisplay("No time tags in the selected interval\n");
   return(QPLUG_FAILURE);
 }

 QdRMatrixSeq_var joined_B = QdRMatrixSeq_var::narrow(joined[0]);
 if (joined_B.is_nil()) {    
   QplugAppendTextDisplay("B not a vector series?\n");
   return(QPLUG_FAILURE);
 }
 
 QdRMatrixSeq_var masked_B = new QdRMatrixSeq();
 joined_B->get_subsequence(*masked_B,msk); 
 if (masked_B.is_nil() || masked_B->sequence_size() <=1) {    
   QplugAppendTextDisplay("No B data in interval\n");
   return(QPLUG_FAILURE);
 }

 bfield_nest = (QdRMatrixSeq_var)  masked_B;

 nestsize = bfield_nest->sequence_size();
 printf("Nestsize B = %d\n",nestsize);
 fflush(stdout);
 if (nestsize<=1) {    
   QplugAppendTextDisplay("Nestsize B = 0. Is time correct ?...\n");
   return(QPLUG_FAILURE);
 }
 
 QdRScalarSeq_var joined_V = QdRScalarSeq_var::narrow(joined[1]); 
 if (joined_V.is_nil() || joined_V->sequence_size() <=1) {    
   QplugAppendTextDisplay("V data not a scalar sequence?\n");
   return(QPLUG_FAILURE);
 }

 QdRScalarSeq_var masked_V = new QdRScalarSeq();
 joined_V->get_subsequence(*masked_V,msk); 
 if (masked_V.is_nil() || masked_V->sequence_size() <=1) {    
   QplugAppendTextDisplay("No V data in interval\n");
   return(QPLUG_FAILURE);
 }

 velocity_nest = (QdRScalarSeq_var)  masked_V;

 nestsize = velocity_nest->sequence_size();
 printf("Nestsize V = %d\n",nestsize);
 fflush(stdout);
 if (nestsize<=1) {    
   QplugAppendTextDisplay("Nestsize V = 0. Is time correct ?...\n");
   return(QPLUG_FAILURE);
 }


 QdRScalarSeq_var joined_N = QdRScalarSeq_var::narrow(joined[2]);
 if (joined_N.is_nil() || joined_N->sequence_size() <=1) {    
   QplugAppendTextDisplay("Density data not a scalar sequence?\n");
   return(QPLUG_FAILURE);
 }

 QdRScalarSeq_var masked_N = new QdRScalarSeq();
 joined_N->get_subsequence(*masked_N,msk);
 if (masked_N.is_nil() || masked_N->sequence_size() <=1) {    
   QplugAppendTextDisplay("No Density data in interval\n");
   return(QPLUG_FAILURE);
 }

 density_nest = (QdRScalarSeq_var)  masked_N;

 nestsize = density_nest->sequence_size();
 printf("Nestsize N = %d\n",nestsize);
 fflush(stdout);
 if (nestsize<=1) {    
   QplugAppendTextDisplay("Nestsize n = 0. Is time correct ?...\n");
   return(QPLUG_FAILURE);
 }


 // ----- check if time shift is given -------

 bool shift_given = false;
 double shift = 0.0;
 QdRScalar_var shift_ws = QdRScalar_var::narrow((*call_list)[4]);
 if (!shift_ws.is_nil()) {
   shift_given = true;
   shift = (double) (*shift_ws);
 }


// ---- declare and calculate the output objects -----

 epsilon    = new QdRScalarSeq(nestsize);
 energy     = new QdRScalarSeq(nestsize);
 pdyn       = new QdRScalarSeq(nestsize);
 rstandoff  = new QdRScalarSeq(nestsize);
 clockangle = new QdRScalarSeq(nestsize);
 vout       = new QdRScalarSeq(nestsize);
 bout       = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));

 L0 = 7.0 * 6370000;  
 double ACEpos = 1.582e6-63700.0;  // ACE position rel to BS at 10Re
 double aveshift = 0.00;

 // calculate epsilon and Pdyn
 // NB ! we use SI units B [T], V [km/s]
 // epilon [W], pdyn [nPa]
 // TotalEnergy in [10^12 Ws = 10^12 J]
 // rstandoff = MP Rst [Re] after (Walker & Russell, 1995)

 float dir = 1.00;

 for(int i=0; i<nestsize; i++)  {
    B.x = (double) bfield_nest[i][0] * 1.0E-9;
    B.y = (double) bfield_nest[i][1] * 1.0E-9;
    B.z = (double) bfield_nest[i][2] * 1.0E-9;
    Vsw = (double) velocity_nest[i];
    theta = acos(B.z / sqrt(B.y*B.y + B.z*B.z));
    btot2 = (B.x*B.x + B.y*B.y + B.z*B.z);
    Bt = sqrt(B.y*B.y + B.z*B.z);
    dir = 1.0;
    if (B.y <= 0.0) dir = -1.0;
    bout[i][0] = (double) bfield_nest[i][0];
    bout[i][1] = (double) bfield_nest[i][1];
    bout[i][2] = (double) bfield_nest[i][2];
    epsilon[i] = 1.0E7 * sqr(L0) * btot2 * pow(sin(theta/2.0),4.0) * Vsw*1E3;   
    energy[i] = TotalEnergy + (double)epsilon[i] * (double)deltat * 1.0E-12; 
    pdyn[i] = density_nest[i]*ionMass*Vsw*Vsw;
    rstandoff[i] = 107.4 * pow((density_nest[i] * Vsw*Vsw),-0.16667);
    clockangle[i] = dir * 57.29 * acos(B.z/Bt);
    vout[i] = Vsw;
    if (!shift_given) shift = (ACEpos/Vsw) + 300.0;
    (*tt_out)[i] += Seconds(shift);
    if (i > 0) {
      if ((*tt_out)[i] < (*tt_out)[i-1]) (*tt_out)[i] = (*tt_out)[i-1] + Seconds(0.1);
    }
    TotalEnergy = (double) energy[i];
    aveshift += shift;
    //    printf(" i, Eps, Pdyn, Rst, TotE %d %10.2f %10.2f %10.2f = %10.2f\n",
    //	     i,(double)epsilon[i],(double)pdyn[i],(double)rstandoff[i],TotalEnergy);
 }    

 aveshift /= (double) nestsize;

 // printf("Completed calculations..\n");

 // --- send some output to QSAS window ---
 sprintf(txtbuf,"\nNest size (# of data points) = %d\n",nestsize);
 QplugAppendTextDisplay(txtbuf);

 sprintf(txtbuf,"Total accumulated energy in interval = %8.2f x 10^12 J\n\n",TotalEnergy);
 QplugAppendTextDisplay(txtbuf);


 // --- insert timetags + attributes and push to WL ---

 QdTimeSeq_var timetags = get_timetags(velocity_nest);
 
 
 sprintf(txtbuf,"Solar wind bulk speed |V|, shifted %d seconds",(int) aveshift);
 set_timetags(vout,tt_out);
 QuSetTxtAttr("SI_conversion","1.0e3>m s^-1",(QdObject_var) vout);
 QuSetTxtAttr("UNITS","km/s",(QdObject_var) vout);
 QuSetTxtAttr("LABLAXIS","V_bulk",(QdObject_var) vout);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) vout);
 if (shift != 0.0) {
   sprintf(txtbuf,"V_SW_Bulk_sh_%ds",(int) aveshift);
   SetSlotText(0,txtbuf);
 }

 sprintf(txtbuf,"BIMF, shifted %d seconds",(int) aveshift);
 set_timetags(bout,tt_out);
 QuSetTxtAttr("SI_conversion","1.0e-9>T",(QdObject_var) bout);
 QuSetTxtAttr("UNITS","nT",(QdObject_var) bout);
 QuSetFrame((QdObject_var) bout, "vector>gse_xyz");
 QuSetTxtAttr("LABLAXIS","B_IMF",(QdObject_var) bout);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) bout);
 if (shift != 0.0) {
   sprintf(txtbuf,"B_IMF_sh_%ds",(int) aveshift);
   SetSlotText(1,txtbuf);
 }


 sprintf(txtbuf,"Epsilon parameter, shifted %d seconds",(int) aveshift);
 set_timetags(epsilon,tt_out);
 QuSetTxtAttr("SI_conversion","1>W",(QdObject_var) epsilon);
 QuSetTxtAttr("UNITS","W",(QdObject_var) epsilon);
 QuSetTxtAttr("LABLAXIS","Epsilon",(QdObject_var) epsilon);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) epsilon);
 if (shift != 0.0) {
   sprintf(txtbuf,"Epsilon_sh_%ds",(int) aveshift);
   SetSlotText(2,txtbuf);
 }


 sprintf(txtbuf,"Accumulated energy , shifted %d seconds",(int) aveshift);
 set_timetags(energy,tt_out);
 QuSetTxtAttr("SI_conversion","1.0e12>J",(QdObject_var) energy);
 QuSetTxtAttr("UNITS","10^12 J",(QdObject_var) energy);
 QuSetTxtAttr("LABLAXIS","Acc_energy",(QdObject_var) energy);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) energy);
 if (shift != 0.0) {
   sprintf(txtbuf,"Energy_sh_%ds",(int) aveshift);
   SetSlotText(3,txtbuf);
 }

 sprintf(txtbuf,"Solar wind dynamic pressure , shifted %d seconds",(int) aveshift);
 set_timetags(pdyn,tt_out);
 QuSetTxtAttr("SI_conversion","1e-09>Pa",(QdObject_var) pdyn);
 QuSetTxtAttr("UNITS","nPa",(QdObject_var) pdyn);
 QuSetTxtAttr("LABLAXIS","Pdyn",(QdObject_var) pdyn);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) pdyn);
 if (shift != 0.0) {
   sprintf(txtbuf,"P_dyn_sh_%ds",(int) aveshift);
   SetSlotText(4,txtbuf);
 }

 sprintf(txtbuf,"magnetopause standoff distance Rst = 107.4*[nV^2]^(-1/6), shifted %d seconds",(int) aveshift);
 set_timetags(rstandoff,tt_out);
 QuSetTxtAttr("UNITS","Re",(QdObject_var) rstandoff);
 QuSetTxtAttr("LABLAXIS","Rst",(QdObject_var) rstandoff);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) rstandoff);
 if (shift != 0.0) {
   sprintf(txtbuf,"Rst_sh_%ds",(int) aveshift);
   SetSlotText(5,txtbuf);
 }

 sprintf(txtbuf,"IMF clock angle, shifted %d seconds",(int) aveshift);
 set_timetags(clockangle,tt_out);
 QuSetTxtAttr("SI_conversion","1>degree",(QdObject_var) clockangle);
 QuSetTxtAttr("UNITS","Degrees",(QdObject_var) clockangle);
 QuSetTxtAttr("LABLAXIS","Theta",(QdObject_var) clockangle);
 QuSetTxtAttr("CATDESC",txtbuf,(QdObject_var) clockangle);
 if (shift != 0.0) {
   sprintf(txtbuf,"Theta_sh_%ds",(int) aveshift);
   SetSlotText(6,txtbuf);
 }

 call_list->push_back((QdObject_var) vout);
 call_list->push_back((QdObject_var) bout);
 call_list->push_back((QdObject_var) epsilon);
 call_list->push_back((QdObject_var) energy);
 call_list->push_back((QdObject_var) pdyn);
 call_list->push_back((QdObject_var) rstandoff);
 call_list->push_back((QdObject_var) clockangle);

 QplugAppendTextDisplay ("ISSI_Solarwind >> data object created - exiting successfully\n");

 return QPLUG_SUCCESS;

}



double vecsum(double x, double y, double z)
{
    return sqrt(x*x +y*y +z*z);
}


float sqr(float x)
{
    return(x*x);
}


int main()
{
  // do nothing
}


