/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    ISSI_ClusterVectorCalc
* 
* MODULE:	QSAS Manipulation of Cluster Vector Quantities
*
* LANGUAGE:	C++
*
* FUNCTION:
*
* PURPOSE:      Calculate the curl or divergence of a vector quantity
*               measured by the Cluster satellites
* 
* DESCRIPTION:
*
* ARGUMENTS:
*    IN:
*     VECTOR_TIME_SERIES    V_C1..V_C4           input vector quantity
*     VECTOR_TIME_SERIES    Pos_C1..Pos_C4       position vectors for the four SC's
*     TIME_INTERVAL         Time                 time to calculate curl or div for
* 
*    OUT:
*     ANY_DATA_OBJECT       curlV, divV, gradV   results of the calculation;
*                                                a scalar for div, a vector for curl
*
* GLOBALS:
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
*
*
* ALGORITHM:   See chap 12 & 14 of Blue ISSI book.
*
* HISTORY:	SEH, 27 Aug 2003
*                    26 Dec 2003 - QSAS 2 version
*                     2 May 2005, flipped curl sign in function CurlVec()
*                    27 May 2005, fixed units for div, grad
*             AJA, 27 Sept 2005, incorporated into QSAS release
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "qar.h"
#include "Qdos.h"
#include "ConstrainedMinVariance.h"

extern  "C" QplugReturnStatus  ISSI_ClusterVectorCalc  (QplugArgList *);

double maxVal(QdRMatrixSeq_var V);
double sum_SCs(QdRMatrix_var V1, QdRMatrix_var V2, int compV1, int compV2);
QdRMatrix_var CurlVec(QdRMatrix_var V, QdRMatrix_var R);
QdRMatrix_var GradVec(QdRMatrix_var V, QdRMatrix_var R);
QdRMatrix_var GradScalar(QdRMatrix_var V, QdRMatrix_var R);  
double DivergenceVec(QdRMatrix_var V, QdRMatrix_var R);
double vecsum(QdRMatrix_var V);


using namespace QSAS;
QplugReturnStatus  ISSI_ClusterVectorCalc  ( QplugArgList *call_list )
{
 int    nestsize, nseq = 8, triage_cnt=0, VERBOSE=0;     
 double deltat, avg_width;

 QdRMatrixSeq_var  V1,V2,V3,V4,Pos1,Pos2,Pos3,Pos4;
 QdObjectSeq_var ttIn    = new QdObjectSeq(nseq);
 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 QdObjectSeq_var joinMethod = new QdObjectSeq(nseq);
 QdObjectSeq_var gapValue   = new QdObjectSeq(nseq);
 QdObjectSeq_var gapHandler = new QdObjectSeq(nseq);
 QdTimeSeq_var ttJoined = get_timetags((*call_list)[2]);    // tt of joined V and Pos form SC3 vec
 QdTimeSeq_var tt_out;                                      // tt of masked V and Pos


 // --- SEH 2 May 2005 : set this to 1.0 to remove offsets in e.g. E-fields
 double offsremove = 0.00;

 //	..first, get input vectors from each SC and join ..

 QuIsTSRegular(ttJoined,&deltat); // AJA - don't care about regular, but need estimate of spacing
 avg_width = 1.5 * deltat;
 printf("delta t = %f  ->  gap_fill if t > %f\n",deltat,avg_width);
 fflush(stdout);


 
 char *frame_V;
 frame_V = QplugNewSTR( QuGetFrameAttr( ((* call_list)[0])) ); 
 const char *frame_i;
 for (int i=1; i<nseq; i++){
   frame_i = QuGetFrameAttr( ((* call_list)[i]));
   if( strcmp(frame_i, "" ) == 0) continue;
   if( strcmp(frame_i, frame_V) != 0) {   
     QplugAppendTextDisplay("Inputs not in same vector frame:");
     QplugAppendTextDisplay(frame_i);
     QplugAppendTextDisplay(" and ");
     QplugAppendTextDisplay(frame_V);
     return QPLUG_FAILURE;
   }
 }
 
 bool is_Bfield = false;
 char *SI_V;
 SI_V = QuGetAttrText("SI_conversion", ((* call_list)[0])); 
 if(QarSameBaseSI(SI_V, "1.0e-9>T") ) is_Bfield = true;

 char *SI_i;
 for (int i=1; i<4; i++){
   SI_i = QuGetAttrText("SI_conversion", ((* call_list)[i]));
   if( strcmp(SI_i, "" ) == 0) continue;
   if( !QarSameBaseSI(SI_V, SI_i) ) {   
     QplugAppendTextDisplay("Inputs not in same base units:");
     QplugAppendTextDisplay(SI_i);
     QplugAppendTextDisplay(" and ");
     QplugAppendTextDisplay(SI_V);
     return QPLUG_FAILURE;
   }
 }
 
 char *SI_R;
 SI_R = QuGetAttrText("SI_conversion", ((* call_list)[4])); 
 for (int i=5; i<8; i++){
   SI_i = QuGetAttrText("SI_conversion", ((* call_list)[i]));
   if( strcmp(SI_i, "" ) == 0) continue;
   if( !QarSameBaseSI(SI_R, SI_i) ) {   
     QplugAppendTextDisplay("Inputs not in same base units:");
     QplugAppendTextDisplay(SI_i);
     QplugAppendTextDisplay(" and ");
     QplugAppendTextDisplay(SI_R);
     return QPLUG_FAILURE;
   }
 }
 
 for (int i=0; i<nseq; i++) {
   toJoin[i] = QuGetVectorXYZ ((*call_list)[i]);  // this is no slower then the test on representation as xyz vectors are returned untouched
   if ( (toJoin[i]).is_nil() ){
     QplugAppendTextDisplay("Failed to get vector in cartesian form");
   }
   ttIn[i] = get_timetags((*call_list)[i]);
   joined[i]  = toJoin[i]->create();

   joinMethod[i] = new QdString(LINEAR);
   gapValue[i] = new QdRScalar(avg_width);
   gapHandler[i] = new QdString(LINEAR);
 }
 
try {
   multijoin(ttIn, toJoin, ttJoined, joined,
      joinMethod, gapValue, gapHandler, triage_cnt);
 }
 catch(Exception &e) {
   printf("Exception : avg_width = %f\n",avg_width);
   fflush(stdout);
   e.print_msg();
   return QPLUG_FAILURE;
 }


 QdTimeInterval_var start_end = get_timeinterval((* call_list)[8]);
 SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 tt_out = new QdTimeSeq();
 ttJoined->get_subsequence(*tt_out, msk);


 // ---- get Vector quantity for given time interval - store in V ----

 QdRMatrixSeq_var V1j = QdRMatrixSeq_var::narrow(joined[0]);     // V*j = joined vector
 if(V1j.is_nil()) {
   QplugAppendTextDisplay("Input 1 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var V1m =  new QdRMatrixSeq();     
 V1j->get_subsequence(*V1m,msk);                                 // V*m = masked vector
 V1m->copy_xrefs_from((* call_list)[0] );
 if(is_Bfield){
    V1 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(V1m, "1.0e-9>T")); // ensure units are in Cluster units assumed
 }
 else{
    V1 = (QdRMatrixSeq_var) V1m;                                    // V* = vector TS to use
  }

   
 QdRMatrixSeq_var V2j = QdRMatrixSeq_var::narrow(joined[1]);
  if(V2j.is_nil()) {
   QplugAppendTextDisplay("Input 2 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var V2m =  new QdRMatrixSeq();
 V2j->get_subsequence(*V2m,msk);
 V2m->copy_xrefs_from((* call_list)[1] );
 if(is_Bfield){
    V2 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(V2m, "1.0e-9>T")); // ensure units are in Cluster units assumed
 }
 else{
    V2 = (QdRMatrixSeq_var) V2m;                                    // V* = vector TS to use
 }


 QdRMatrixSeq_var V3j = QdRMatrixSeq_var::narrow(joined[2]);
  if(V3j.is_nil()) {
   QplugAppendTextDisplay("Input 3 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var V3m =  new QdRMatrixSeq();
 V3j->get_subsequence(*V3m,msk);
 V3m->copy_xrefs_from((* call_list)[2] );
 if(is_Bfield){
    V3 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(V3m, "1.0e-9>T")); // ensure units are in Cluster units assumed
 }
 else{
    V3 = (QdRMatrixSeq_var) V3m;                                    // V* = vector TS to use
 }
 

 QdRMatrixSeq_var V4j = QdRMatrixSeq_var::narrow(joined[3]);
  if(V4j.is_nil()) {
   QplugAppendTextDisplay("Input 4 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var V4m =  new QdRMatrixSeq();
 V4j->get_subsequence(*V4m,msk);
 V4m->copy_xrefs_from((* call_list)[3] );
 if(is_Bfield){
    V4 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(V4m, "1.0e-9>T")); // ensure units are in Cluster units assumed
 }
 else{
    V4 = (QdRMatrixSeq_var) V4m;                                   // V* = vector TS to use
 }
 
 nestsize = V4->sequence_size();

 QdRMatrixSeq_var V = new QdRMatrixSeq(nestsize, RMatrix(Dim(4,3)));

 for (int t=0; t<nestsize; t++) {
   for (int i=0; i<3; i++) {
     (*V)[t][0][i] = (double)(*V1)[t][i] - ( offsremove * (double)(*V1)[0][i]) ;
     (*V)[t][1][i] = (double)(*V2)[t][i] - ( offsremove * (double)(*V2)[0][i]) ;
     (*V)[t][2][i] = (double)(*V3)[t][i] - ( offsremove * (double)(*V3)[0][i]) ;
     (*V)[t][3][i] = (double)(*V4)[t][i] - ( offsremove * (double)(*V4)[0][i]) ;
   }
 }

// ---- get position for given time interval - store in P -------

 QdRMatrixSeq_var P = new QdRMatrixSeq(nestsize, RMatrix(Dim(4,3)));


 QdRMatrixSeq_var P1j = QdRMatrixSeq_var::narrow(joined[4]);     // P*j = joined position vector
  if(P1j.is_nil()) {
   QplugAppendTextDisplay("Position 1 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var P1m =  new QdRMatrixSeq();
 P1j->get_subsequence(*P1m,msk);                                 // P*m = masked position vector
 P1m->copy_xrefs_from((* call_list)[4] );
 Pos1 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(P1m, "1.0e3>m"));       // Pos* = position vector for this time interval in Cluster units
 if(Pos1.is_nil()){
   QplugAppendTextDisplay("Position 1 is not in units of distance");
   return QPLUG_FAILURE;
 }

 QdRMatrixSeq_var P2j = QdRMatrixSeq_var::narrow(joined[5]);   
  if(P2j.is_nil()) {
   QplugAppendTextDisplay("Position 2 is not vector sequence");
   return QPLUG_FAILURE;
 } 
 QdRMatrixSeq_var P2m =  new QdRMatrixSeq();
 P2j->get_subsequence(*P2m,msk);
 P2m->copy_xrefs_from((* call_list)[5] );
 Pos2 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(P2m, "1.0e3>m")); 
 if(Pos2.is_nil()){
   QplugAppendTextDisplay("Position 2 is not in units of distance");
   return QPLUG_FAILURE;
 }

 QdRMatrixSeq_var P3j = QdRMatrixSeq_var::narrow(joined[6]);     
 if(P3j.is_nil()) {
   QplugAppendTextDisplay("Position 3 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var P3m =  new QdRMatrixSeq();
 P3j->get_subsequence(*P3m,msk);
 P3m->copy_xrefs_from((* call_list)[6] );
 Pos3 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(P3m, "1.0e3>m")); 
 if(Pos3.is_nil()){
   QplugAppendTextDisplay("Position 3 is not in units of distance");
   return QPLUG_FAILURE;
 }

 QdRMatrixSeq_var P4j = QdRMatrixSeq_var::narrow(joined[7]);     // P*j = joined position vector
  if(P4j.is_nil()) {
   QplugAppendTextDisplay("Position 4 is not vector sequence");
   return QPLUG_FAILURE;
 }
 QdRMatrixSeq_var P4m =  new QdRMatrixSeq();
 P4j->get_subsequence(*P4m,msk);
 P4m->copy_xrefs_from((* call_list)[7] );
 Pos4 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(P4m, "1.0e3>m")); 
 if(Pos4.is_nil()){
   QplugAppendTextDisplay("Position 4 is not in units of distance");
   return QPLUG_FAILURE;
 }

 nestsize = Pos4->sequence_size();

 for (int t=0; t<nestsize; t++) {
   for (int i=0; i<3; i++) {
     (*P)[t][0][i] = (double)(*Pos1)[t][i];
     (*P)[t][1][i] = (double)(*Pos2)[t][i];
     (*P)[t][2][i] = (double)(*Pos3)[t][i];
     (*P)[t][3][i] = (double)(*Pos4)[t][i];
   }
 }

 // --- get scalefactor for Curl and check if this is a current (scale = mu_0 so that J = curl B / mu_0) ----

 char txtbuf[63] = "";
 char curlunit[63] = "";
 char divunit[63] = "";
 char scaleunit[63] = "";
 double scalefac=1.0;
 bool iscurrent=FALSE;

 QdRScalar_var sfac = QdRScalar_var::narrow((*call_list)[9]);

 sprintf(txtbuf,"%s/%s",QuGetAttrText("UNITS", (*call_list)[2]),QuGetAttrText("UNITS", (*call_list)[6]));
 sprintf(curlunit,"%s/%s",QuGetAttrText("UNITS", (*call_list)[2]),QuGetAttrText("UNITS", (*call_list)[6]));
 sprintf(divunit,"%s/%s",QuGetAttrText("UNITS", (*call_list)[2]),QuGetAttrText("UNITS", (*call_list)[6]));

 if (!sfac.is_nil()) {
   scalefac = ((double) *sfac);
   sprintf(scaleunit,"%s",(QuGetAttrText("UNITS", (*call_list)[9])));
   if (strlen(scaleunit) > 1) strcat(curlunit,"/");
   strcat(curlunit,scaleunit);
   printf("> Scalefactor supplied : %6.4g %s\n",scalefac, scaleunit);
   fflush(stdout);
   if (!strncasecmp(curlunit,"nT/km/H/m",9)) {
     iscurrent = TRUE;
     sprintf(curlunit,"A/m^2");
     scalefac *= 1.0E12;                  // divide (10^-9 nT / 10^3 m) to get current in A/m2
   }
 }
 else {
   printf("> Curl unscaled\n");
   fflush(stdout);
 }

 // printf("Nestsize = %d\n",nestsize);

 // --- now, calculate curl, div and grad for each time step -----
  

 int c[]={0,1,2,0,1};
 int i, j, sc, c1=0, c2=1, c3=2, c4=3;
 double totvol=0.0, vol[]={0.0, 0.0, 0.0, 0.0};

 QdRMatrix_var R = new QdRMatrix(Dim(3,3),0.0);    // Volumetric tensor - see Blue book sec 12.2.1 + 14.2.1
 QdRMatrix_var K = new QdRMatrix(Dim(4,3),0.0);    // Reciprocalvector - see Blue book, sec 12.4
 QdRMatrix_var Vs = new QdRMatrix(Dim(4,3),0.0);

 QdRMatrix_var Curl =  new QdRMatrix(Dim(3),0.0);
 QdRMatrix_var Grad =  new QdRMatrix(Dim(3),0.0);

 QdRMatrixSeq_var Curl_TS  = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 QdRMatrixSeq_var Grad_TS  = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
 QdRScalarSeq_var Div_TS   = new QdRScalarSeq(nestsize);
 QdRScalarSeq_var Vol_TS   = new QdRScalarSeq(nestsize);
 QdRMatrixSeq_var Bary_TS  = new QdRMatrixSeq(nestsize,RMatrix(Dim(3),0.0));


 for (int t=0; t<nestsize; t++) {                  // Reciprocal vectors according to Eq. (14.7) in Blue Book
   for (i=0; i<3; i++) {
     (*K)[c1][i] =   ((double)(*P)[t][c3][c[i+1]] - (double)(*P)[t][c2][c[i+1]]) *  ((double)(*P)[t][c4][c[i+2]] - (double)(*P)[t][c2][c[i+2]])  
       - ((double)(*P)[t][c3][c[i+2]] - (double)(*P)[t][c2][c[i+2]]) *  ((double)(*P)[t][c4][c[i+1]] - (double)(*P)[t][c2][c[i+1]]);
     (*K)[c2][i] = -(((double)(*P)[t][c4][c[i+1]] - (double)(*P)[t][c3][c[i+1]]) *  ((double)(*P)[t][c3][c[i+2]] - (double)(*P)[t][c1][c[i+2]]) 
		     - ((double)(*P)[t][c4][c[i+2]] - (double)(*P)[t][c3][c[i+2]]) *  ((double)(*P)[t][c3][c[i+1]] - (double)(*P)[t][c1][c[i+1]]));
     (*K)[c3][i] =   ((double)(*P)[t][c4][c[i+1]] - (double)(*P)[t][c1][c[i+1]]) *  ((double)(*P)[t][c4][c[i+2]] - (double)(*P)[t][c2][c[i+2]])
       - ((double)(*P)[t][c4][c[i+2]] - (double)(*P)[t][c1][c[i+2]]) *  ((double)(*P)[t][c4][c[i+1]] - (double)(*P)[t][c2][c[i+1]]);
     (*K)[c4][i] =   ((double)(*P)[t][c2][c[i+1]] - (double)(*P)[t][c1][c[i+1]]) *  ((double)(*P)[t][c3][c[i+2]] - (double)(*P)[t][c1][c[i+2]])
       - ((double)(*P)[t][c2][c[i+2]] - (double)(*P)[t][c1][c[i+2]]) *  ((double)(*P)[t][c3][c[i+1]] - (double)(*P)[t][c1][c[i+1]]);
   }

   vol[c1] = -((double)(*P)[t][c2][0]-(double)(*P)[t][c1][0]) * (double)(*K)[c1][0] -  ((double)(*P)[t][c2][1]-(double)(*P)[t][c1][1]) * (double)(*K)[c1][1] - ((double)(*P)[t][c2][2]-(double)(*P)[t][c1][2]) * (double)(*K)[c1][2];
   vol[c2] = -((double)(*P)[t][c3][0]-(double)(*P)[t][c2][0]) * (double)(*K)[c2][0] -  ((double)(*P)[t][c3][1]-(double)(*P)[t][c2][1]) * (double)(*K)[c2][1] - ((double)(*P)[t][c3][2]-(double)(*P)[t][c2][2]) * (double)(*K)[c2][2];
   vol[c3] = -((double)(*P)[t][c4][0]-(double)(*P)[t][c3][0]) * (double)(*K)[c3][0] -  ((double)(*P)[t][c4][1]-(double)(*P)[t][c3][1]) * (double)(*K)[c3][1] - ((double)(*P)[t][c4][2]-(double)(*P)[t][c3][2]) * (double)(*K)[c3][2];
   vol[c4] = +((double)(*P)[t][c4][0]-(double)(*P)[t][c1][0]) * (double)(*K)[c4][0] +  ((double)(*P)[t][c4][1]-(double)(*P)[t][c1][1]) * (double)(*K)[c4][1] + ((double)(*P)[t][c4][2]-(double)(*P)[t][c1][2]) * (double)(*K)[c4][2];

   for (sc=0;sc<4; sc++) {
     for (i=0; i<3; i++) { 	
       (*Bary_TS)[t][i] += 0.25 * (*V)[t][sc][i];
       (*K)[sc][i] /= vol[sc];
       (*Vs)[sc][i] = (*V)[t][sc][i];
       for(j=0; j<3; j++) {
	 (*R)[i][j]  += (double)(*P)[t][sc][i] * (double)(*P)[t][sc][j]; 													   
       }
     totvol += 0.25*fabs(vol[sc]);
     }
   } 


   if (VERBOSE) {
     for (i=0; i<3; i++)  printf("[%10.7f  %10.7f  %10.7f]\n",(double)(*R)[i][0],(double)(*R)[i][1],(double)(*R)[i][2]);
     printf("\n");
     for (i=0; i<4; i++)  printf("[%14.11f  %14.11f  %14.11f]\n",(double)(*K)[i][0],(double)(*K)[i][1],(double)(*K)[i][2]);
	 fflush(stdout);
   }


   Curl =  CurlVec(Vs,K);
   Grad =  GradScalar(Vs,K);
   (*Div_TS)[t] = DivergenceVec(Vs,K);
   (*Vol_TS)[t] = totvol;

   //   printf("t = %d , CurlV = %f  \n",t, (double)(*Curl)[0]);

   for (i=0; i<3; i++) {
     (*Curl_TS)[t][i] = (double)(*Curl)[i] / scalefac;
     (*Grad_TS)[t][i] = (double)(*Grad)[i];
   }

   if (VERBOSE) {
     printf("t = %d\n",t);
     printf("curl V = %f  %f  %f\n", (double)(*Curl)[0], (double)(*Curl)[1], (double)(*Curl)[2]);
     printf("div V = %f \n", (double)(*Div_TS)[t]);
     printf("grad V = [%10.7f  %10.7f  %10.7f]\n",(double)(*Grad)[0],(double)(*Grad)[1],(double)(*Grad)[2]);
     printf("vol = %f\n\n", totvol);
	 fflush(stdout);
   }
 }

 

 //  ----- do a MinVar of Curl Vector vector -------

 QdRMatrix_var Rmatx = new QdRMatrix(Dim(3,3), 0.0);  
 QdRMatrix_var RmatxB = new QdRMatrix(Dim(3,3), 0.0);  
 QdRMatrix_var Rmatx_tmp = new QdRMatrix(Dim(3,3), 0.0);  
 QdRMatrix_var Eigenvalues = new QdRMatrix(Dim(3), 0.0) ;
 char evalsevecs[80];
 QString str("Eigenvalues, eigenvectors of MVAJ are :\n");
 const int constrained = 0;
 
 try{
     ConstrainedMinVariance minvar(*Curl_TS,*Curl_TS,constrained);   // qdos minvariance call
     *Rmatx = minvar.eigenvectors().duplicate(SHALLOW);
     *Eigenvalues = minvar.eigenvalues().duplicate(SHALLOW); 
 }
 catch(Exception& e) {
   e.print_msg();
   return QPLUG_FAILURE; 
 }    
 if ((double)(*Rmatx)[0][0] < 0) (*Rmatx) *= -1.0;
 

 for (int i=0; i<3; i++){
   printf("%10.5g  [%10.6g, %10.6g, %10.6g ]\n", 
	   (double)(*Eigenvalues)[i], (double)(*Rmatx)[i][0],
	   (double) (*Rmatx)[i][1], (double)(*Rmatx)[i][2]);
   sprintf(evalsevecs,"%10.6g  [%10.6g, %10.6g, %10.6g ]\n", 
	   (double)(*Eigenvalues)[i], (double)(*Rmatx)[i][0],
	   (double) (*Rmatx)[i][1], (double)(*Rmatx)[i][2]);
   str += evalsevecs;
 }   // end for
fflush(stdout);
 // ----- rotate J into the MinVar system -------


 double mu0 = 1.25664e-6;
 QdRMatrixSeq_var Curl_TS_MVA  = new QdRMatrixSeq(nestsize,RMatrix(Dim(3), 0.0));
 QdRMatrixSeq_var Bary_TS_MVA  = new QdRMatrixSeq(nestsize,RMatrix(Dim(3),0.0));
 QdRMatrixSeq_var Velo_TS_MVA  = new QdRMatrixSeq(nestsize,RMatrix(Dim(3),0.0));
 for (int k=0; k<nestsize; k++) {
   Curl_TS_MVA[k] = (*Rmatx) * Curl_TS[k];
   Bary_TS_MVA[k] = (*Rmatx) * Bary_TS[k];

   //   printf("curl_TS[0] = %f \n",(double)(*Curl_TS)[k][0]);
 }


 QdRMatrix_var Jint =  new QdRMatrix(Dim(3),0.0);
 QdRMatrix_var Jave =  new QdRMatrix(Dim(3),0.0);
 QdRMatrix_var Jint_GSE =  new QdRMatrix(Dim(3),0.0);
 QdRMatrix_var Jave_GSE =  new QdRMatrix(Dim(3),0.0);
 
 QdRMatrix_var Bint =  new QdRMatrix(Dim(3),0.0);
 QdRMatrix_var v_n = new QdRMatrix(Dim(3),0.0);
 QdRMatrix_var deltaB = new QdRMatrix(Dim(3),0.0);
 double nf = (double) nestsize;

 for (i=0; i<3; i++) {
   (*deltaB)[i] = (double)(*Bary_TS_MVA)[nestsize-1][i] - (double)(*Bary_TS_MVA)[0][i];
   sprintf(txtbuf,"\nDelta B_%d = %f",i,(double)(*deltaB)[i]);
   str += txtbuf;
 }

 for (int k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     (*Bint)[i] += (double)(*Bary_TS_MVA)[k][i];
     (*Jint)[i] += (double)(*Curl_TS_MVA)[k][i] * deltat;
     (*Jave)[i] += (double)(*Curl_TS_MVA)[k][i];
     (*Jint_GSE)[i] += (double)(*Curl_TS)[k][i] * deltat;
     (*Jave_GSE)[i] += (double)(*Curl_TS)[k][i];
   }
 }
 
 double bave = 
      (double)(*Rmatx)[constrained][0] * (double)(*Bint)[0]/nf + 
      (double)(*Rmatx)[constrained][1] * (double)(*Bint)[1]/nf + 
      (double)(*Rmatx)[constrained][2] * (double)(*Bint)[2]/nf;
 double df = 
      (double)(*Rmatx)[constrained][0] * (double)(*Jave_GSE)[0]/nf + 
      (double)(*Rmatx)[constrained][1] * (double)(*Jave_GSE)[1]/nf + 
      (double)(*Rmatx)[constrained][2] * (double)(*Jave_GSE)[2]/nf;

 printf("<J> = %d %g  %g  %g  (%g)\n",i,(double)(*Jave)[0]/nf,(double)(*Jave)[1]/nf,(double)(*Jave)[2]/nf, df);
 sprintf(txtbuf,"\n\n<J> = [ %g  %g  %g] (%g along n_MVAJ)\n",(double)(*Jave)[0]/nf,(double)(*Jave)[1]/nf,(double)(*Jave)[2]/nf, df);
 str += txtbuf;

 sprintf(txtbuf,"\n\n<J> = [ %g  %g  %g] (%g along n_MVAJ)\n",(double)(*Jave_GSE)[0]/nf,(double)(*Jave_GSE)[1]/nf,(double)(*Jave_GSE)[2]/nf, df);
 str += txtbuf;

 double maxJ = maxVal(Curl_TS);
 sprintf(txtbuf,"J_max = %10.3g\n",maxJ);
 str += txtbuf;
 
 printf("<B> = %d %g  %g  %g  (%g)\n",i,(double)(*Bint)[0]/nf,(double)(*Bint)[1]/nf,(double)(*Bint)[2]/nf, bave);
 sprintf(txtbuf,"<B> = [ %g  %g  %g] (%g along n_MVAJ)\n\n",(double)(*Bint)[0]/nf,(double)(*Bint)[1]/nf,(double)(*Bint)[2]/nf, bave);
 str += txtbuf;
fflush(stdout);
 
 for (int k=1; k<nestsize; k++) {
   (*Velo_TS_MVA)[k][0] = ((double)(*Bary_TS)[k-1][1] - (double)(*Bary_TS)[k][1]) * 1e-12 / (deltat * mu0 * (double)(*Curl_TS_MVA)[k][2]);
   (*Velo_TS_MVA)[k][1] = ((double)(*Bary_TS)[k-1][2] - (double)(*Bary_TS)[k][2]) * 1e-12 / (deltat * mu0 * (double)(*Curl_TS_MVA)[k][1]);
   (*Velo_TS_MVA)[k][2] = ((double)(*Velo_TS_MVA)[k][0] + (double)(*Velo_TS_MVA)[k][1]) / 2.0;  //avg of above
   //   printf("k = %d, %f  %f  -> %f  \n",k, (double)(*Velo_TS_MVA)[k][0],(double)(*Velo_TS_MVA)[k][1],(double)(*Velo_TS_MVA)[k][2] );
 }
 if (constrained == 1) {
   (*v_n)[0] =  (double)(*deltaB)[0] *1e-9 / (mu0 * (double)(*Jint)[2]);
   (*v_n)[1] = - (double)(*deltaB)[2] *1e-9 / (mu0 * (double)(*Jint)[0]);
   (*v_n)[2] =  ((double) (*v_n)[0] + (double) (*v_n)[1]) / 2.0; //avg of above
 } else {
   (*v_n)[0] =  (double)(*deltaB)[1] *1e-9 / (mu0 * (double)(*Jint)[2]);
   (*v_n)[1] = - (double)(*deltaB)[2] *1e-9 / (mu0 * (double)(*Jint)[1]);
   (*v_n)[2] =  ((double) (*v_n)[0] + (double) (*v_n)[1]) / 2.0;
 }

 QplugAppendTextDisplay ("--\n");
 for (i=0; i<2; i++) {
   //   printf("J_%d =  %g\n",i,(double)(*Jint)[i]);
   sprintf(txtbuf,"V_%d    = %g\n",i,(double)(*v_n)[i]);
   printf("%s\n",txtbuf);
   str += txtbuf;
 }
 sprintf(txtbuf,"V_ave = %g\n",(double)(*v_n)[2]);
 str += txtbuf;
 

 // ---- do MVA of barycentric B-field ----------- 

 str += "\nEigenvals and Eigenvecs for MVAB_bary :\n";
 try{
     ConstrainedMinVariance minvar(*Bary_TS,*Bary_TS, constrained);   // qdos minvariance call
     *RmatxB = minvar.eigenvectors().duplicate(SHALLOW);
     *Eigenvalues = minvar.eigenvalues().duplicate(SHALLOW); 
 }
 catch(Exception& e) {
   e.print_msg();
   return QPLUG_FAILURE; 
 }    
 if ((double)(*RmatxB)[0][0] < 0) (*RmatxB) *= -1.0;
 
 for (int i=0; i<3; i++){
   printf("%10.5g  [%10.6g, %10.6g, %10.6g ]\n", 
	   (double)(*Eigenvalues)[i], (double)(*RmatxB)[i][0],
	   (double) (*RmatxB)[i][1], (double)(*RmatxB)[i][2]);
   sprintf(evalsevecs,"%10.6g  [%10.6g, %10.6g, %10.6g ]\n", 
	   (double)(*Eigenvalues)[i], (double)(*RmatxB)[i][0],
	   (double) (*RmatxB)[i][1], (double)(*RmatxB)[i][2]);
   str += evalsevecs;
 }   // end for
 fflush(stdout);
 bave = 
      (double)(*RmatxB)[constrained][0] * (double)(*Bint)[0]/nf + 
      (double)(*RmatxB)[constrained][1] * (double)(*Bint)[1]/nf + 
      (double)(*RmatxB)[constrained][2] * (double)(*Bint)[2]/nf;
 df = 
      (double)(*RmatxB)[constrained][0] * (double)(*Jave)[0]/nf + 
      (double)(*RmatxB)[constrained][1] * (double)(*Jave)[1]/nf + 
      (double)(*RmatxB)[constrained][2] * (double)(*Jave)[2]/nf;

 sprintf(txtbuf,"\n\n<J> = [ %g  %g  %g] (%g along n_MVAB_bary)\n",(double)(*Jave)[0]/nf,(double)(*Jave)[1]/nf,(double)(*Jave)[2]/nf, df);
 str += txtbuf;

 printf("<B> = %d %g  %g  %g  (%g)\n",i,(double)(*Bint)[0]/nf,(double)(*Bint)[1]/nf,(double)(*Bint)[2]/nf, bave);
 fflush(stdout);
 sprintf(txtbuf,"<B> = [ %g  %g  %g] (%g along n_MVAB_bary)\n\n",(double)(*Bint)[0]/nf,(double)(*Bint)[1]/nf,(double)(*Bint)[2]/nf, bave);
 str += txtbuf;
 QplugAppendTextDisplay (str.latin1());


 // ---- send some outputs -------------------
  
   // find inverses first
  char * attr_inv_ptr; 
  attr_inv_ptr = QarSIConv_inverse((*call_list)[4]);
  QdObject_var inv_ptr = (*call_list)[4]->create(); // this is empty as we only need metadata
    
  if(attr_inv_ptr != NULL){
       QuSetTxtAttr("SI_conversion", attr_inv_ptr, inv_ptr);
       delete[] attr_inv_ptr;
  }
 
 // SI_conversion
 char * attr_ptr;
 if ( (attr_ptr = QarSIConv_product((*call_list)[0], inv_ptr)) != NULL) {
   QuSetTxtAttr("SI_conversion", attr_ptr, Curl_TS);
   QuSetTxtAttr("SI_conversion", attr_ptr, Curl_TS_MVA);
   QuSetTxtAttr("SI_conversion", attr_ptr, Div_TS);
   QuSetTxtAttr("SI_conversion", attr_ptr, Grad_TS);
   delete[] attr_ptr;
 }

 QuSetFrame((QdObject_var) Curl_TS, frame_V);
 QuSetTxtAttr("UNITS",curlunit,(QdObject_var) Curl_TS);
 if (iscurrent) {
   QuSetTxtAttr("LABLAXIS","J",(QdObject_var) Curl_TS);
 } else {
   QuSetTxtAttr("LABLAXIS","Curl V",(QdObject_var) Curl_TS);
 }   
 set_timetags((QdObject_var) Curl_TS, tt_out);

 QuSetFrame((QdObject_var) Curl_TS_MVA, frame_V);
 QuSetTxtAttr("UNITS",curlunit,(QdObject_var) Curl_TS_MVA);
 if (iscurrent) {
   QuSetTxtAttr("LABLAXIS","J",(QdObject_var) Curl_TS_MVA);
 } else {
   QuSetTxtAttr("LABLAXIS","Curl V",(QdObject_var) Curl_TS_MVA);
 }   
 set_timetags((QdObject_var) Curl_TS_MVA, tt_out);

 QuSetFrame((QdObject_var) Bary_TS, frame_V);
 set_timetags((QdObject_var) Bary_TS, tt_out);


 QuSetFrame((QdObject_var) Div_TS, "scalar");
 QuSetTxtAttr("UNITS",divunit,(QdObject_var) Div_TS);
 QuSetTxtAttr("LABLAXIS","Div V",(QdObject_var) Div_TS);
 set_timetags((QdObject_var) Div_TS, tt_out);

 QuSetFrame((QdObject_var) Grad_TS, frame_V);
 QuSetTxtAttr("UNITS",divunit,(QdObject_var) Grad_TS);
 QuSetTxtAttr("LABLAXIS","Grad V",(QdObject_var) Grad_TS);
 set_timetags((QdObject_var) Grad_TS, tt_out);

 QuSetFrame((QdObject_var) Velo_TS_MVA, frame_V);
 if(is_Bfield) QuSetTxtAttr("UNITS","km/s",(QdObject_var) Velo_TS_MVA);
 QuSetTxtAttr("LABLAXIS","J",(QdObject_var) Velo_TS_MVA);
 set_timetags((QdObject_var) Velo_TS_MVA, tt_out);

 sprintf(txtbuf,"%s^3",QuGetAttrText("UNITS", (*call_list)[6]));
 QuSetFrame((QdObject_var) Vol_TS, "scalar");
 QuSetTxtAttr("UNITS",txtbuf,(QdObject_var) Vol_TS);
 QuSetTxtAttr("LABLAXIS","TetraVolume",(QdObject_var) Vol_TS);
 set_timetags((QdObject_var) Vol_TS, tt_out);

 call_list->push_back((QdObject_var) Curl_TS);
 call_list->push_back((QdObject_var) Curl_TS_MVA);
 call_list->push_back((QdObject_var) Rmatx);
 call_list->push_back((QdObject_var) Bary_TS);
// call_list->push_back((QdObject_var) Velo_TS_MVA);
 call_list->push_back((QdObject_var) Grad_TS);
 call_list->push_back((QdObject_var) Div_TS);
 // call_list->push_back((QdObject_var) Vol_TS);

 return QPLUG_SUCCESS;
}
 


double maxVal(QdRMatrixSeq_var V)
{
  double vecsum, maxval = 0.0;

  long n = V->sequence_size();
  for (unsigned int i=0; i<n; i++){
    vecsum = (*V)[i][0]* (*V)[i][0] + (*V)[i][1]* (*V)[i][1] + (*V)[i][2]* (*V)[i][2];
    if (vecsum > maxval) maxval = vecsum;
  }
  return sqrt(maxval);
}




double sum_SCs(QdRMatrix_var V1, QdRMatrix_var V2, int compV1, int compV2)
{
  double sum = 0.0;

  for (int i=0; i<4; i++) {
    sum += (double)(*V1)[i][compV1] * (double)(*V2)[i][compV2];
  }

  return (sum);
}




double DivergenceVec(QdRMatrix_var V, QdRMatrix_var R)
  //
  // see Eq. 14.16 in Blue Book
  //
{
  return(sum_SCs(V,R,0,0) + sum_SCs(V,R,1,1) + sum_SCs(V,R, 2,2));
}

  



QdRMatrix_var CurlVec(QdRMatrix_var V, QdRMatrix_var R)
  //
  // see Eq. 14.17 in Blue Book
  // SEH, 2 May 2005; flipped
  //
{
  QdRMatrix_var Curl = new QdRMatrix(Dim(3), 0.0);

  (*Curl)[0] = (-sum_SCs(V,R,1,2) + sum_SCs(V,R,2,1));
  (*Curl)[1] = (-sum_SCs(V,R,2,0) + sum_SCs(V,R,0,2));
  (*Curl)[2] = (-sum_SCs(V,R,0,1) + sum_SCs(V,R,1,0));
  
  return(Curl);
}
 



QdRMatrix_var GradVec(QdRMatrix_var V, QdRMatrix_var R)
{
  QdRMatrix_var Grad = new QdRMatrix(Dim(3,3), 0.0);

  for (int i=0; i<3; i++) {
    for(int j=0; j<3; j++) {
      (*Grad)[i][j] = sum_SCs(V,R,i,j);
    }
  }
   
  return(Grad);
}




QdRMatrix_var GradScalar(QdRMatrix_var V, QdRMatrix_var R)  
{
  QdRMatrix_var Grad = new QdRMatrix(Dim(3), 0.0);
  QdRMatrix_var Magn = new QdRMatrix(Dim(4,1), 0.0);      // magnitude of vector from each SC
  QdRMatrix_var  Vec = new QdRMatrix(Dim(3), 0.0);
  int i,j;  

  for (i=0; i<4; i++) {
    for (j=0; j<3; j++) (*Vec)[j] = (*V)[i][j];
    (*Magn)[i][0] = vecsum(Vec);
  }

  for (i=0; i<3; i++) (*Grad)[i] = sum_SCs(Magn,R,0,i);
  return(Grad);
}




double vecsum(QdRMatrix_var V)
{
  double sum=0.0;
  
  for (int i=0; i<3; i++) sum += (*V)[i] * (*V)[i];

  return(sqrt(sum));
}

int main()
{
}
