#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include <algorithm>

#include "qplug_if.h"
#include "qdutil.h"
#include "qar.h"

using namespace QSAS;

extern "C" QplugReturnStatus PEACE2GSE(QplugArgList *);
extern "C" QplugReturnStatus CIS2GSE(QplugArgList *);
extern "C" QplugReturnStatus EFW2Pot(QplugArgList *);

int QrMJD(int year, int month, int day){
        
    /* Modified Julian Date from MJD(YEAR,MONTH,DAY) */
 
    /* This is NOT the same as Julian Date in Qanlib */
 
    int mjd;
    double wy, wm;
                   
    if (month > 2) {
        wy = (double) year;
        wm = (double) month + 1.0;
    }
    else{
        wy = (double) year - 1.0;
        wm = (double) month + 13.0;
    }      
  
    mjd = (int) (365.25*wy) + (int) (30.6001*wm) + day - 679019;

    return mjd;
 
} /* end QrMJD */


bool QrEpochToMJD_UT(double epoch, int * MJDdate, double * ut){

    int year, month, day, hour, minute, second, msec;
 
    if( epoch <= 0.0) return false;
  
    cdfEpoch(epoch).split_ymd(year, month, day, 
             hour, minute, second, msec);
                   
    *MJDdate = QrMJD(year, month, day);
  
    *ut = (double)( hour + ((double) minute)/ 60.0 
            + ((double) second)/ 3600.0 );
  
    return true;

} /* end QrEpochToMJD_UT */


QplugReturnStatus PEACE2GSE(QplugArgList * call_list)
{
    QdObject_var lat_temp=(*call_list)[0];
    QdObject_var lon_temp=(*call_list)[1];

    QdRScalarSeq_var lat_seq=QdRScalarSeq_var::narrow(lat_temp);
    QdRScalarSeq_var lon_seq=QdRScalarSeq_var::narrow(lon_temp);
    QdRScalar_var lat_const=QdRScalar_var::narrow(lat_temp);
    QdRScalar_var lon_const=QdRScalar_var::narrow(lon_temp);

    double conv2rad_lat=M_PI/180.;
    double conv2rad_lon=M_PI/180.;
    
    if(lat_seq.is_nil() && lat_const.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for Latitude");
        return QPLUG_FAILURE;
    }
    
    if(lon_seq.is_nil() && lon_const.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for Longitude");
        return QPLUG_FAILURE;
    }
    
    if(!lat_seq.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lat_seq);
        if(units.contains("rad", false)) conv2rad_lat=1.;
    }
    if(!lat_const.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lat_const);
        if(units.contains("rad", false)) conv2rad_lat=1.;
    }
    if(!lon_seq.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lon_seq);
        if(units.contains("rad", false)) conv2rad_lon=1.;
    }
    if(!lon_const.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lon_const);
        if(units.contains("rad", false)) conv2rad_lon=1.;
    }
    
    if(!lat_seq.is_nil() || !lon_seq.is_nil())
    {
        QdTimeSeq_var times;
        int n_elements;
        if(!lat_seq.is_nil() && !lon_seq.is_nil())
        {
            int qar_status=QarTestJoined(lat_seq, lon_seq);
            if(qar_status==QAR_ERR)
            {
                QplugAppendTextDisplay("Latitude and Longitude not joined.");
                return QPLUG_FAILURE;
            }
        }
        
        if(!lat_seq.is_nil())
        {
            times=get_timetags(lat_seq);
        }
        else
        {
            times=get_timetags(lon_seq);
        }
        
        n_elements=times->sequence_size();
        
        QdRMatrixSeq_var result=new QdRMatrixSeq(n_elements, Dim(3,3));
        
        double alpha_x, delta_x, T0, UT, M, Phi_h, delta_p;
        int MJD;
        double x_dot_h, x_dot_p, x_p_h, alpha;
        for(int i=0; i<n_elements; i++)
        {
            if(!lat_seq.is_nil())
                delta_x=lat_seq->elt(i)*conv2rad_lat;
            else delta_x=(double)(*lat_const)*conv2rad_lat;
            
            if(!lon_seq.is_nil())
                alpha_x=lon_seq->elt(i)*conv2rad_lon;
            else alpha_x=(double)(*lon_const)*conv2rad_lon;

            QrEpochToMJD_UT(cdfEpoch((*times)[i]), &MJD, &UT);
            
            
            T0=(MJD-51544.5)/36525.;
            M=(357.528+35999.05*T0+0.04107*UT)/180.*M_PI;
            
            Phi_h=280.460+36000.772*T0+0.04107*UT+(1.915-0.0048*T0)*sin(M)+0.02*sin(2.*M);
            Phi_h=Phi_h/180.*M_PI;
            
            delta_p=90.-23.439+0.013*T0;
            delta_p=delta_p/180.*M_PI;
            
            x_dot_h=cos(Phi_h)*cos(delta_x)*cos(alpha_x)+sin(Phi_h)*(sin(delta_p)*cos(delta_x)*sin(alpha_x)+cos(delta_p)*sin(delta_x));
            
            x_dot_p=sin(delta_p)*sin(delta_x)-cos(delta_p)*cos(delta_x)*sin(alpha_x);
            
            x_p_h=cos(Phi_h)*(cos(delta_p)*sin(delta_x)+sin(delta_p)*cos(delta_x)*sin(alpha_x))-sin(Phi_h)*cos(delta_x)*cos(alpha_x);
            
            alpha=sqrt(1.-x_dot_h*x_dot_h);
            
            result->elt(i)(0, 0)=alpha;
            result->elt(i)(0, 1)=0.;
            result->elt(i)(0, 2)=x_dot_h;
            result->elt(i)(1, 0)=-x_dot_h*x_p_h/alpha;
            result->elt(i)(1, 1)=x_dot_p/alpha;
            result->elt(i)(1, 2)=x_p_h;
            result->elt(i)(2, 0)=-x_dot_h*x_dot_p/alpha;
            result->elt(i)(2, 1)=-x_p_h/alpha;
            result->elt(i)(2, 2)=x_dot_p;
            
            
        }

        if(!lat_seq.is_nil())
            set_timetags(result, get_timetags(lat_seq));
        else set_timetags(result, get_timetags(lon_seq));
        call_list->push_back(result);
    }
    else
    {
        QdRMatrix_var result=new QdRMatrix(Dim(3,3));
        
        double lat((*lat_const)*conv2rad_lat), lon((*lon_const)*conv2rad_lon);
        double x_dot_h(cos(lat)*cos(lon)), x_dot_p(sin(lat)), x_p_h(cos(lat)*sin(lon));
        double alpha(sqrt(1.-x_dot_h*x_dot_h));
        
        (*result)(0, 0)=alpha;
        (*result)(0, 1)=0.;
        (*result)(0, 2)=x_dot_h;
        (*result)(1, 0)=-x_dot_h*x_p_h/alpha;
        (*result)(1, 1)=x_dot_p/alpha;
        (*result)(1, 2)=x_p_h;
        (*result)(2, 0)=-x_dot_h*x_dot_p/alpha;
        (*result)(2, 1)=-x_p_h/alpha;
        (*result)(2, 2)=x_dot_p;
        
        call_list->push_back(result);
    }
    
    return QPLUG_SUCCESS;
}

QplugReturnStatus CIS2GSE(QplugArgList * call_list)
{
    QdObject_var lat_temp=(*call_list)[0];
    QdObject_var lon_temp=(*call_list)[1];

    QdRScalarSeq_var lat_seq=QdRScalarSeq_var::narrow(lat_temp);
    QdRScalarSeq_var lon_seq=QdRScalarSeq_var::narrow(lon_temp);
    QdRScalar_var lat_const=QdRScalar_var::narrow(lat_temp);
    QdRScalar_var lon_const=QdRScalar_var::narrow(lon_temp);

    double conv2rad_lat=M_PI/180.;
    double conv2rad_lon=M_PI/180.;
    
    if(lat_seq.is_nil() && lat_const.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for Latitude");
        return QPLUG_FAILURE;
    }
    
    if(lon_seq.is_nil() && lon_const.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for Longitude");
        return QPLUG_FAILURE;
    }
    
    if(!lat_seq.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lat_seq);
        if(units.contains("rad", false)) conv2rad_lat=1.;
    }
    if(!lat_const.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lat_const);
        if(units.contains("rad", false)) conv2rad_lat=1.;
    }
    if(!lon_seq.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lon_seq);
        if(units.contains("rad", false)) conv2rad_lon=1.;
    }
    if(!lon_const.is_nil())
    {
        QString units=QuGetAttrText("UNITS", lon_const);
        if(units.contains("rad", false)) conv2rad_lon=1.;
    }
    
    if(!lat_seq.is_nil() || !lon_seq.is_nil())
    {
        QdTimeSeq_var times;
        int n_elements;
        if(!lat_seq.is_nil() && !lon_seq.is_nil())
        {
            int qar_status=QarTestJoined(lat_seq, lon_seq);
            if(qar_status==QAR_ERR)
            {
                QplugAppendTextDisplay("Latitude and Longitude not joined.");
                return QPLUG_FAILURE;
            }
        }
        
        if(!lat_seq.is_nil())
        {
            times=get_timetags(lat_seq);
        }
        else
        {
            times=get_timetags(lon_seq);
        }
        
        n_elements=times->sequence_size();
        
        QdRMatrixSeq_var result=new QdRMatrixSeq(n_elements, Dim(3,3));
        
        double alpha_x, delta_x, T0, UT, M, Phi_h, delta_p;
        int MJD;
        double x_dot_h, x_dot_p, x_p_h, alpha;
        for(int i=0; i<n_elements; i++)
        {
            if(!lat_seq.is_nil())
                delta_x=lat_seq->elt(i)*conv2rad_lat;
            else delta_x=(double)(*lat_const)*conv2rad_lat;
            
            if(!lon_seq.is_nil())
                alpha_x=lon_seq->elt(i)*conv2rad_lon;
            else alpha_x=(double)(*lon_const)*conv2rad_lon;

            double theta=alpha_x;
            double phi=M_PI/2.-delta_x;
            
            double RR[]={cos(theta)*sin(phi), sin(theta)*sin(phi), cos(phi)};
            double alpha=1.0/sqrt(RR[1]*RR[1] + RR[2]*RR[2]);
            
            result->elt(i)(0, 0)=alpha*(RR[0]*RR[0] + RR[2]*RR[2]); 
            result->elt(i)(0, 1)=-alpha*RR[0]*RR[1];
            result->elt(i)(0, 2)=-alpha*RR[0]*RR[2]; 
            result->elt(i)(1, 0)=0;
            result->elt(i)(1, 1)=alpha*RR[2];
            result->elt(i)(2, 2)=-alpha*RR[1];
            result->elt(i)(2, 0)=RR[0];
            result->elt(i)(2, 1)=RR[1];
            result->elt(i)(2, 2)=RR[2];
            
//             QrEpochToMJD_UT(cdfEpoch((*times)[i]), &MJD, &UT);
//             
//             
//             T0=(MJD-51544.5)/36525.;
//             M=(357.528+35999.05*T0+0.04107*UT)/180.*M_PI;
//             
//             Phi_h=280.460+36000.772*T0+0.04107*UT+(1.915-0.0048*T0)*sin(M)+0.02*sin(2.*M);
//             Phi_h=Phi_h/180.*M_PI;
//             
//             delta_p=90.-23.439+0.013*T0;
//             delta_p=delta_p/180.*M_PI;
//             
//             x_dot_h=cos(Phi_h)*cos(delta_x)*cos(alpha_x)+sin(Phi_h)*(sin(delta_p)*cos(delta_x)*sin(alpha_x)+cos(delta_p)*sin(delta_x));
//             
//             x_dot_p=sin(delta_p)*sin(delta_x)-cos(delta_p)*cos(delta_x)*sin(alpha_x);
//             
//             x_p_h=cos(Phi_h)*(cos(delta_p)*sin(delta_x)+sin(delta_p)*cos(delta_x)*sin(alpha_x))-sin(Phi_h)*cos(delta_x)*cos(alpha_x);
//             
//             alpha=sqrt(1.-x_dot_h*x_dot_h);
//             
//             result->elt(i)(0, 0)=alpha;
//             result->elt(i)(0, 1)=0.;
//             result->elt(i)(0, 2)=x_dot_h;
//             result->elt(i)(1, 0)=-x_dot_h*x_p_h/alpha;
//             result->elt(i)(1, 1)=x_dot_p/alpha;
//             result->elt(i)(1, 2)=x_p_h;
//             result->elt(i)(2, 0)=-x_dot_h*x_dot_p/alpha;
//             result->elt(i)(2, 1)=-x_p_h/alpha;
//             result->elt(i)(2, 2)=x_dot_p;
            

            //Temporary correction
//             for(int j=0; j<3; j++)
//                     result->elt(i)(j, 0)=-result->elt(i)(j, 0);
        }
        
        if(!lat_seq.is_nil())
            set_timetags(result, get_timetags(lat_seq));
        else set_timetags(result, get_timetags(lon_seq));
        call_list->push_back(result);
    }
    else
    {
        QdRMatrix_var result=new QdRMatrix(Dim(3,3));
        
        double lat((*lat_const)*conv2rad_lat), lon((*lon_const)*conv2rad_lon);
        double x_dot_h(cos(lat)*cos(lon)), x_dot_p(sin(lat)), x_p_h(cos(lat)*sin(lon));
        double alpha(sqrt(1.-x_dot_h*x_dot_h));
        
        (*result)(0, 0)=alpha;
        (*result)(0, 1)=0.;
        (*result)(0, 2)=x_dot_h;
        (*result)(1, 0)=-x_dot_h*x_p_h/alpha;
        (*result)(1, 1)=x_dot_p/alpha;
        (*result)(1, 2)=x_p_h;
        (*result)(2, 0)=-x_dot_h*x_dot_p/alpha;
        (*result)(2, 1)=-x_p_h/alpha;
        (*result)(2, 2)=x_dot_p;
        
        call_list->push_back(result);
    }
    
    return QPLUG_SUCCESS;
}

QplugReturnStatus EFW2Pot(QplugArgList * call_list)
{
    QdRScalar_var alpha=QdRScalar_var::narrow((*call_list)[1]);
    if(alpha.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for Alpha");
        return QPLUG_FAILURE;
    }
    
    QdRScalar_var beta=QdRScalar_var::narrow((*call_list)[2]);
    if(beta.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for Beta");
        return QPLUG_FAILURE;
    }
    
    QdNObject_var potential=QdNObject_var::narrow((*call_list)[0]);
    
    QdObject_var result=alpha*potential+beta; // TODO handle FILLVALs!
    
    result->copy_xrefs_from(potential);
    QuSetTxtAttr("FIELDNAM", "S/C potential", (QdObject_var)result);
    
    call_list->push_back(result);
    
    return QPLUG_SUCCESS;
}
