#include "triangulation.h"

// Computes the convex hull of a set of 3D or 4D points, using the QuickHull algorithm
// Barber, C. B., Dobkin, D. P., & Huhdanpaa, H. T. (1996). The Quickhull algorithm for convex hulls. ACM Trans. on Mathematical Software, 22(4),469--483.
// This is used for 2D (resp. 3D) delaunay triangulation. The z (resp. t) coordinate has to be a convex function of the others (z=x^2+y^2, t=x^2+y^2+z^2). The projections on the (x,y) (resp. (x, y, z)) hyperplane of the facets having normals "facing down" (z_normal<0 (resp t_normal<0)) are the triangles of the Delaunay triangulation
// See http://mathworld.wolfram.com/DelaunayTriangulation.html

// void plotHull(Facet3D*f)
// {
//    if(f->status==Facet3D::currentStatus) return;
//     
//    f->status=Facet3D::currentStatus;
//    //cout << "plothull" << endl;
//    cpgsci(100);
//    if(f->outerNormal.z<0.)
//    {
//     float plotx[3], ploty[3];
//     plotx[0]=f->p1->x;
//     ploty[0]=f->p1->y;
//     plotx[1]=f->p2->x;
//     ploty[1]=f->p2->y;
//     plotx[2]=f->p3->x;
//     ploty[2]=f->p3->y;
//     cpgpoly(3, plotx, ploty);
//    }
//    plotHull(f->neighbour1());
//    plotHull(f->neighbour2());
//    plotHull(f->neighbour3());
// }

// float plotHull(list<Point2D>& points, list<Element2D>& hull)
// {
//     float xmin(0.), xmax(0.), ymin(0.), ymax(0.), valmax(0.);
//     for(list<Point2D>::iterator iter=points.begin(); iter!=points.end(); iter++)
//     {
//         if(iter->val>valmax) valmax=iter->val;
//         if(iter->x>xmax) xmax=iter->x;
//         if(iter->y>ymax) ymax=iter->y;
//         if(iter->x<xmin) xmin=iter->x;
//         if(iter->y<ymin) ymin=iter->y;
//         cpgsci((int)floor(iter->val*100.));
//     }
//     
//     cpgeras();
//     cpgswin(xmin, xmax, ymin, ymax);
// //     cpgswin(-5., 5., -5., 5.);
// 
//     cpgsfs(1);
//     for(list<Point2D>::iterator iter=points.begin(); iter!=points.end(); iter++)
//     {
// 
//         cpgsci((int)floor(iter->val*100./valmax));
//         float x=iter->x;
//         float y=iter->y;
//         cpgpt(1, &x, &y, -4);
//     }
//     
//     cpgsfs(2);
//     float plotx[3], ploty[3];
//     for(list<Element2D>::iterator iter=hull.begin(); iter!=hull.end(); iter++)
//     {
//         cpgsci(100);
//         plotx[0]=iter->points[0][0];
//         ploty[0]=iter->points[0][1];
//         plotx[1]=iter->points[1][0];
//         ploty[1]=iter->points[1][1];
//         plotx[2]=iter->points[2][0];
//         ploty[2]=iter->points[2][1];
//         cpgpoly(3, plotx, ploty);
//     }
//     
//     return valmax;
// }

/**
* Converts linked 3D facets to the related Delaunay triangulation
* Dimension reduction. Only the facets of the lower part of the hull are relevant.
* @param f 1st 3D facet. The attached ones will be found recursively.
* @param hull Output delaunay triangulation. The elements are added in hull.
* @author Alban Rochel
*/
void convertToElementsList(Facet3D* f, list<Element2D>& hull)
{
    if(f==NULL) return;
    if(f->status!=Facet3D::currentStatus)
    {
        f->status=Facet3D::currentStatus;
        if(f->outerNormal.z<0.)
        {
            Element2D temp(f->p1, f->p2, f->p3);
            hull.push_front(temp);
        }
        convertToElementsList(f->neighbour1(), hull);
        convertToElementsList(f->neighbour2(), hull);
        convertToElementsList(f->neighbour3(), hull);
    }
}

/**
 * Collects the pointers to the facets and ridges to be deleted.
 * @warning Deletion is not performed
 * @param f 1st 3D facet. The attached ones will be found recursively.
 * @param facets List of pointers to facets to be deleted.
 * @param ridges List of pointers to ridges to be deleted.
 * @author Alban Rochel
 */
void cleanUp(Facet3D* f, list<Facet3D*>& facets, list<Ridge3D*>& ridges)
{
    if(f->status==Facet3D::currentStatus) return;
    
    f->status=Facet3D::currentStatus;
    facets.push_front(f);
    ridges.push_front(f->r1);
    ridges.push_front(f->r2);
    ridges.push_front(f->r3);
    
    cleanUp(f->neighbour1(), facets, ridges);
    cleanUp(f->neighbour2(), facets, ridges);
    cleanUp(f->neighbour3(), facets, ridges);
}

/**
 * Returns a pointer to a facet that has to be improved (non-empty visibility list).
 * The neighbours are visited randomly: this prevents the function to loop a long time in the same set of neighbours, and increases dramatically the performance. Each neighbour is visited once at most.
 * @param f 1st 3D facet. The attached ones will be found recursively.
 * @return An improvable 3D facet, or NULL if the facets define the convex hull of the set of points (end of algorithm).
 * @author Alban Rochel
 */
Facet3D* getImprovableFacet(Facet3D* f)
{    
    if(f->status==Facet3D::currentStatus)
    {
        return NULL;
    }
    
    f->status=Facet3D::currentStatus;
    
    if(!f->halfspace.empty())
    {
        return f;
    }
    
    Facet3D* tempRes;
    
    switch(rand()%3)
    {
        case 0:    tempRes=getImprovableFacet(f->neighbour1());
                    if(tempRes==NULL)
                    {
                        tempRes=getImprovableFacet(f->neighbour2());
                        if(tempRes==NULL)
                        {
                            tempRes=getImprovableFacet(f->neighbour3());
                        }
                    }
                    break;
                    
        case 1:    tempRes=getImprovableFacet(f->neighbour2());
                    if(tempRes==NULL)
                    {
                        tempRes=getImprovableFacet(f->neighbour3());
                        if(tempRes==NULL)
                        {
                            tempRes=getImprovableFacet(f->neighbour1());
                        }
                    }
                    break;
                    
        default:    tempRes=getImprovableFacet(f->neighbour3());
                    if(tempRes==NULL)
                    {
                        tempRes=getImprovableFacet(f->neighbour1());
                        if(tempRes==NULL)
                        {
                            tempRes=getImprovableFacet(f->neighbour2());
                        }
                    }
                    break;
    }
    

    
    return tempRes;
}

/**
 * Returns a pointer to a ridge.
 * Creates the ridge if not exisiting. Deletes a ridge when getRidge was called twice on this ridge (1 ridge = 2 and only 2 neighbours). First call with an empty ridge list. The subsequent calls will create or delete ridges in the list. If used properly, after the last call, the ridge list is empty (pairs of ridges were created).
 * @param Ridges The current list of ridges.
 * @param p1 First point of the ridge (the order is not significant)
 * @param p1 Second point of the ridge (the order is not significant)
 * @return Pointer to a ridge
 * @author Alban Rochel
 */
Ridge3D* getRidge(list<Ridge3D*>& Ridges, Point3D* p1, Point3D* p2)
{
    Ridge3D* tempRidge=new Ridge3D(p1, p2);
    Ridge3D* found=NULL;
    for(list<Ridge3D*>::iterator iter=Ridges.begin(); iter!=Ridges.end() && found==NULL; iter++)
    {
        if(*tempRidge==**iter) found=*iter;
    }
    if(found==NULL)
    {
        Ridges.push_front(tempRidge);
    }
    else
    {
        Ridges.remove(found);
        delete(tempRidge);
        tempRidge=found;
    }
    return tempRidge;

}

/**
 * Being given a facet and a point visible from this facet, computes the neighbouring facets that "see" the point (and, thus, that will be removed), as well as the set of ridges on which to build the new facets: those facets will link the ridges to the point.
 * @param f An improvable facet (non-empty visibility list)
 * @param ref The reference point, visible from the facet (ideally, the farthest one)
 * @param link After calling removeVisibleFacets, this points to a safe facet, that will not be removed.
 * @param orphanRidges Ridges on the border of the hole to fill, having only one neighbour. Needs to be sorted and having doubles removed after call.
 * @param halfspace List of the points that were visible from the set of removed facets. Needs to be sorted and having doubles removed after call.
 * @param facetsToDelete List of pointers to delete.
 * @param ridgesToDelete List of pointers to delete.
 * @author Alban Rochel
 */
void removeVisibleFacets(Facet3D* f, Point3D* ref, Facet3D* link, list<Ridge3D*>& orphanRidges, list<Point3D*>& halfspace, list<Facet3D*>& facetsToDelete, list<Ridge3D*>& ridgesToDelete)
{
    if(f==NULL)
    {
        return;
    }
    
    if(f->status==Facet3D::currentStatus) // already processed
    {
        return;
    }

    if(f->distance(ref)>0.) // Facet visible from ref point
    {
        f->status=Facet3D::currentStatus;

        halfspace.splice(halfspace.begin(), f->halfspace);
        
        Facet3D* f1=f->neighbour1();
        Facet3D* f2=f->neighbour2();
        Facet3D* f3=f->neighbour3();
        
        // delete every link to this facet
        if(f->neighbour1()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r1);
        }
        else
        {
            if(f->r1->f1==f)
            {
                f->r1->f1=NULL;
                if(f->r1->f2==NULL) ridgesToDelete.push_front(f->r1);
            }
            else
            {
                if(f->r1->f2==f)
                {
                    f->r1->f2=NULL;
                    if(f->r1->f1==NULL)  ridgesToDelete.push_front(f->r1);
                }
                else cout << "Ouch" << endl;
            }
        }
        
        if(f->neighbour2()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r2);
        }
        else
        {
            if(f->r2->f1==f)
            {
                f->r2->f1=NULL;
                if(f->r2->f2==NULL)  ridgesToDelete.push_front(f->r2);
            }
            else
            {
                if(f->r2->f2==f)
                {
                    f->r2->f2=NULL;
                    if(f->r2->f1==NULL)  ridgesToDelete.push_front(f->r2);
                }
                else cout << "Ouch" << endl;
            }
        }
        
        if(f->neighbour3()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r3);
        }
        else
        {
            if(f->r3->f1==f)
            {
                f->r3->f1=NULL;
                if(f->r3->f2==NULL)  ridgesToDelete.push_front(f->r3);
            }
            else
            {
                if(f->r3->f2==f)
                {
                    f->r3->f2=NULL;
                    if(f->r3->f1==NULL)  ridgesToDelete.push_front(f->r3);
                }
                else cout << "Ouch" << endl;
            }
        }
        
        facetsToDelete.push_front(f);

        removeVisibleFacets(f1, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        removeVisibleFacets(f2, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        removeVisibleFacets(f3, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
    }
    else
    {
        link=f; // safe facet, won't be deleted
        if(f->neighbour1()==NULL)
        {
            orphanRidges.push_front(f->r1);
        }
        if(f->neighbour2()==NULL)
        {
            orphanRidges.push_front(f->r2);
        }
        if(f->neighbour3()==NULL)
        {
            orphanRidges.push_front(f->r3);
        }
    }
}

/**
 * Creates the first facet of the set.
 * To approximate properly the hull with one facet, creating a facet with distant points.
 * @pre "points" must be sorted so that the first and last points are distant.
 * @param points The set of points.
 * @return Pointer to a facet.
 * @author Alban Rochel
 */
Facet3D* buildFirstFacet(list<Point3D*>& points)
{
    // Building the first triangle
    
    Point3D *p1=points.front();
    Point3D *p2=points.back();
    points.pop_front();
    points.pop_back();
    Point3D *p3;
    Point3D *p4;
    
    Ridge3D *r12, *r13, *r14, *r23, *r24, *r34;
    
    r12=new Ridge3D(p1, p2);
    
    // Look for the point maximizing the triangle area

    double current(0.), max(-1.), d12((*p2-*p1).norm()), d13, d23;
    
    
    for(list<Point3D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        d13=(**iter-*p1).norm();
        d23=(**iter-*p2).norm();
        current=(d12+d13+d23)*(-d12+d13+d23)*(d12-d13+d23)*(d12+d13-d23); // 1/4*sqrt normally, but we don't care
        if(max<0. || current>max)
        {
            max=current;
            p3=*iter;
        }
    }
   
    r13=new Ridge3D(p1, p3);
    r23=new Ridge3D(p2, p3);
    
    Facet3D* f123=new Facet3D(r12, r13, r23);
    
    // Look for the point that is the farthest to the ridge
    max=-1;
    Point3D normal=(*p2-*p1).crossProd(*p3-*p1);
    
    for(list<Point3D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        current=fabs(normal.dotProd(**iter-f123->centre));
        if(max<0. || current>max)
        {
            max=current;
            p4=*iter;
        }
    }
    
    r14=new Ridge3D(p1, p4);
    r24=new Ridge3D(p2, p4);
    r34=new Ridge3D(p3, p4);
    
    Facet3D* f124=new Facet3D(r12, r14, r24);
    Facet3D* f134=new Facet3D(r13, r14, r34);
    Facet3D* f234=new Facet3D(r23, r24, r34);
    
    // Link the facets altogether
    r12->f1=f123;   r12->f2=f124;
    r13->f1=f123;   r13->f2=f134;
    r14->f1=f124;   r14->f2=f134;
    r23->f1=f123;   r23->f2=f234;
    r24->f1=f124;   r24->f2=f234;
    r34->f1=f134;   r34->f2=f234;
    
    
    // Compute outer normals
    f123->setOuterNormal(p4);
    f124->setOuterNormal(p3);
    f134->setOuterNormal(p2);
    f234->setOuterNormal(p1);

    // Build associated lists
    
    for(list<Point3D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        if(*iter!=p3 && *iter!=p4)
        {
            if(f123->outerNormal.dotProd(**iter-f123->centre) > 0.)
            {
                f123->halfspace.push_front(*iter);
            }
            if(f124->outerNormal.dotProd(**iter-f124->centre) > 0.)
            {
                f124->halfspace.push_front(*iter);
            }
            if(f134->outerNormal.dotProd(**iter-f134->centre) > 0.)
            {
                f134->halfspace.push_front(*iter);
            }
            if(f234->outerNormal.dotProd(**iter-f234->centre) > 0.)
            {
                f234->halfspace.push_front(*iter);
            }
        }
    }

    return f123;
}

/**
 * Computes the Delaunay triangulation of the (x,y) projection of a set of 3D points, using the QuickHull algorithm.
 * @param points The set of 3D points
 * @param hull The output list of Element2D (Delaunay trangulation)
 * @pre There must be at least 4 points in the input set of points.
 * @author Alban Rochel
 */
void quickHull(list<Point3D*>& points, list<Element2D>& hull)
{
    // Building first facet
    Facet3D* firstFacet=buildFirstFacet(points);

    // Getting a facet with a non-empty visibility list
    Facet3D* currentFacet=getImprovableFacet(firstFacet);
    Facet3D::currentStatus++;

    Point3D* np; // new point
    double max, current;
    Point3D centre;
    Facet3D* tempFacet;
    while(currentFacet!=NULL) //while some facets are improvable
    {
        centre=currentFacet->centre;
        max=-1;

        // The point that will be used in the following is the farthest visible point
        for(list<Point3D*>::iterator iter=currentFacet->halfspace.begin(); iter!=currentFacet->halfspace.end(); iter++)
        {
            current=currentFacet->distance(*iter);
            if(max<current)
            {
                max=current;
                np=*iter;
            }
        }

        Facet3D* startFacet=currentFacet;
        // We will delete the facets that see np
        // List of points visible by these deleted facets (not to lose them)
        list<Point3D*> halfspace;
        // Ridges onto which to build new facets
        list<Ridge3D*> orphanRidges;
        // Stuff to delete
        list<Facet3D*> facetsToDelete;
        list<Ridge3D*> ridgesToDelete;
        removeVisibleFacets(startFacet, np, currentFacet, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        Facet3D::currentStatus++;
        
        // Deleting what's to be deleted
        facetsToDelete.sort();
        facetsToDelete.unique();
        while(!facetsToDelete.empty())
        {
            delete facetsToDelete.front();
            facetsToDelete.pop_front();
        }
        ridgesToDelete.sort();
        ridgesToDelete.unique();
        while(!ridgesToDelete.empty())
        {
            delete ridgesToDelete.front();
            ridgesToDelete.pop_front();
        }
        // Cleaning the remaining lists
        halfspace.sort();
        halfspace.unique();
        orphanRidges.sort();
        orphanRidges.unique();
        
        // Now we will create new facets, and thus new ridges
        list<Facet3D*> newFacets;
        list<Ridge3D*> newRidges;
        
        // For all ridge that lost one neighbour (border of the gap)
        for(list<Ridge3D*>::iterator iter=orphanRidges.begin(); iter!=orphanRidges.end(); iter++)
        {
            // Creation of 2 ridges linking the current one to the reference point
            Point3D* p1=(*iter)->p1;
            Point3D* p2=(*iter)->p2;
            
            Ridge3D* r1n=getRidge(newRidges, p1, np);
            Ridge3D* r2n=getRidge(newRidges, p2, np);
            
            // Creation of the facet defined by the 3 ridges
            Facet3D* fn =new Facet3D(*iter, r1n, r2n);
            
            currentFacet=fn;
            
            // Linking the neighbours altogether
            if(r1n->f1==NULL)
            {
                r1n->f1=fn;
            }
            else
            {
                r1n->f2=fn;
            }
            
            if(r2n->f1==NULL)
            {
                r2n->f1=fn;
            }
            else
            {
                r2n->f2=fn;
            }
            
            if((*iter)->f1==NULL)
            {
                (*iter)->f1=fn;
            }
            else
            {
                if((*iter)->f2==NULL)
                {
                    (*iter)->f2=fn;
                }
                else
                {
                    cout << "Double orphan!" << endl;
                }
            }
            
            // Computation of the new facet outer normal.
            // Computation unrolled for the sake of performance: (*(fn->p2)-*(fn->p1)).crossProd(*(fn->p3)-*(fn->p1));
            fn->outerNormal=Point3D((fn->p2->y-fn->p1->y)*(fn->p3->z-fn->p1->z)-(fn->p2->z-fn->p1->z)*(fn->p3->y-fn->p1->y), (fn->p2->z-fn->p1->z)*(fn->p3->x-fn->p1->x)-(fn->p2->x-fn->p1->x)*(fn->p3->z-fn->p1->z), (fn->p2->x-fn->p1->x)*(fn->p3->y-fn->p1->y)-(fn->p2->y-fn->p1->y)*(fn->p3->x-fn->p1->x));
            if(fn->outerNormal.dotProd(*np-centre)<=0.)
            {
                fn->outerNormal.opposite();
            }
            
            // Creation of the visibility list.
            for(list<Point3D*>::iterator currentPoint=halfspace.begin(); currentPoint!=halfspace.end(); currentPoint++)
            {
                if(*currentPoint!=np && fn->distance(*currentPoint)>0.)
                    fn->halfspace.push_front(*currentPoint);
            }
            newFacets.push_front(fn);
            
        }
        // Security check... Normally, never displays anything.
        for(list<Facet3D*>::iterator f1=newFacets.begin(); f1!=newFacets.end(); f1++)
        {
            if((*f1)->neighbour2()==NULL)
            {
                cout << "unmatched 2" << endl;
            }
            if((*f1)->neighbour3()==NULL)
            {
                cout << "unmatched 3" << endl;
            }
            
            if((*f1)->neighbour3()==NULL)
            {
                cout << "unmatched 1" << endl;
            }
            if((*f1)->neighbour1()==(*f1)->neighbour2() || (*f1)->neighbour1()==(*f1)->neighbour3())
            {
                cout << "Twice the same neighbour 1" << endl;
            }
            if((*f1)->neighbour2()==(*f1)->neighbour3())
            {
                cout << "Twice the same neighbour 2" << endl;
            }
        }

        // Find a new facet to continue
        tempFacet=getImprovableFacet(currentFacet);
        Facet3D::currentStatus++;

        if(tempFacet!=NULL)
        {
            currentFacet=tempFacet;
        }
        else // No improvable facet? Great, we have the hull!
        {
            //update triangularisation
            convertToElementsList(currentFacet, hull);
            Facet3D::currentStatus++;

            // Cleaning up
            cleanUp(currentFacet, facetsToDelete, ridgesToDelete);
            facetsToDelete.sort();
            facetsToDelete.unique();
            ridgesToDelete.sort();
            ridgesToDelete.unique();
            
            while(!facetsToDelete.empty())
            {
                delete facetsToDelete.front();
                facetsToDelete.pop_front();
            }
            while(!ridgesToDelete.empty())
            {
                delete ridgesToDelete.front();
                ridgesToDelete.pop_front();
            }

            currentFacet=tempFacet;
        }
    }
}

// //////////////////////////////////////////
// ////////// 3D tools //////////////////////
// //////////////////////////////////////////

void plotHull(list<Point3D>& points, list<Element3D>& hull)
{
    ofstream output("quickhull.m", ios::out | ios::trunc);
    if(!output)
    {
        cout << "Impossible to create quickhull.m" << endl;
        return;
    }
    
    for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        output << "vtk_plot3(["<< iter->x << "], [" << iter->y << "], [" << iter->z << "], 'MarkerColor', [" << iter->val << ", " << iter->val << ", 0])" << endl;
    }
    
    for(list<Element3D>::iterator iter=hull.begin(); iter!=hull.end(); iter++)
    {
        output << "vtk_line3([" << iter->points[0][0] << ", " << iter->points[1][0] << ", " << iter->points[2][0] << ", " << iter->points[0][0] << ", " << iter->points[3][0] << ", " << iter->points[2][0] << "]', [" << iter->points[0][1] << ", " << iter->points[1][1] << ", " << iter->points[2][1] << ", " << iter->points[0][1] << ", " << iter->points[3][1] << ", " << iter->points[2][1] << "]', [" << iter->points[0][2] << ", " << iter->points[1][2] << ", " << iter->points[2][2] << ", " << iter->points[0][2] << ", " << iter->points[3][2] << ", " << iter->points[2][2] << "]')" << endl;
            
        output << "vtk_line3([" << iter->points[1][0] << ", " << iter->points[3][0] << ", " << iter->points[2][0] << "]', [" << iter->points[1][1] << ", " << iter->points[3][1] << ", " << iter->points[2][1] << "]', [" << iter->points[1][2] << ", " << iter->points[3][2] << ", " << iter->points[2][2] << "]')" << endl;
    }
    
    output.close();
}

void convertToElementsList(Facet4D* f, list<Element3D>& hull)
{
    if(f==NULL) return;
    
    if(f->status!=Facet4D::currentStatus)
    {
        f->status=Facet4D::currentStatus;
        if(f->outerNormal.t<0.)
        {
            Element3D* temp=new Element3D(f->p1, f->p2, f->p3, f->p4);
            hull.push_front(*temp);
            delete(temp);
        }
        
        convertToElementsList(f->neighbour1(), hull);
        convertToElementsList(f->neighbour2(), hull);
        convertToElementsList(f->neighbour3(), hull);
        convertToElementsList(f->neighbour4(), hull);
    }
}
void cleanUp(Facet4D* f, list<Facet4D*>& facets, list<Ridge4D*>& ridges)
{
    if(f->status==Facet4D::currentStatus) return;
    
    f->status=Facet4D::currentStatus;
    facets.push_front(f);
    ridges.push_front(f->r1);
    ridges.push_front(f->r2);
    ridges.push_front(f->r3);
    ridges.push_front(f->r4);
    
    cleanUp(f->neighbour1(), facets, ridges);
    cleanUp(f->neighbour2(), facets, ridges);
    cleanUp(f->neighbour3(), facets, ridges);
    cleanUp(f->neighbour4(), facets, ridges);
}

Facet4D* getImprovableFacet(Facet4D* f)
{    
    if(f->status==Facet4D::currentStatus)
    {
        return NULL;
    }
    
    f->status=Facet4D::currentStatus;
    
    if(!f->halfspace.empty())
    {
        return f;
    }
    
    Facet4D* tempRes;
    
    switch(rand()%4)
    {
        case 0:    tempRes=getImprovableFacet(f->neighbour1());
        if(tempRes==NULL)
        {
            tempRes=getImprovableFacet(f->neighbour2());
            if(tempRes==NULL)
            {
                tempRes=getImprovableFacet(f->neighbour3());
                if(tempRes==NULL)
                {
                    tempRes=getImprovableFacet(f->neighbour4());
                }
            }
        }
        break;
                    
        case 1:    tempRes=getImprovableFacet(f->neighbour2());
        if(tempRes==NULL)
        {
            tempRes=getImprovableFacet(f->neighbour3());
            if(tempRes==NULL)
            {
                tempRes=getImprovableFacet(f->neighbour4());
                if(tempRes==NULL)
                {
                    tempRes=getImprovableFacet(f->neighbour1());
                }
            }
        }
        break;
                    
        case 2:    tempRes=getImprovableFacet(f->neighbour3());
        if(tempRes==NULL)
        {
            tempRes=getImprovableFacet(f->neighbour4());
            if(tempRes==NULL)
            {
                tempRes=getImprovableFacet(f->neighbour1());
                if(tempRes==NULL)
                {
                    tempRes=getImprovableFacet(f->neighbour2());
                }
            }
        }
        break;
        
        default:    tempRes=getImprovableFacet(f->neighbour4());
        if(tempRes==NULL)
        {
            tempRes=getImprovableFacet(f->neighbour1());
            if(tempRes==NULL)
            {
                tempRes=getImprovableFacet(f->neighbour2());
                if(tempRes==NULL)
                {
                    tempRes=getImprovableFacet(f->neighbour3());
                }
            }
        }
        break;
    }
    

    
    return tempRes;
}


Ridge4D* getRidge(list<Ridge4D*>& Ridges, Point4D* p1, Point4D* p2, Point4D* p3)
{
    Ridge4D* tempRidge=new Ridge4D(p1, p2, p3);
    Ridge4D* found=NULL;
    for(list<Ridge4D*>::iterator iter=Ridges.begin(); iter!=Ridges.end() && found==NULL; iter++)
    {
        if(*tempRidge==**iter) found=*iter;
    }
    if(found==NULL)
    {
        Ridges.push_front(tempRidge);
    }
    else
    {
        Ridges.remove(found);
        delete(tempRidge);
        tempRidge=found;
    }
    return tempRidge;

}

void removeVisibleFacets(Facet4D* f, Point4D* ref, Facet4D* link, list<Ridge4D*>& orphanRidges, list<Point4D*>& halfspace, list<Facet4D*>& facetsToDelete, list<Ridge4D*>& ridgesToDelete)
{
    if(f==NULL)
    {
        return;
    }
    
    if(f->status==Facet4D::currentStatus) // already processed
    {
        return;
    }

    if(f->distance(ref)>0.) // Facet visible from ref point
    {
        f->status=Facet4D::currentStatus;

        //f->halfspace.sort();
        //halfspace.merge(f->halfspace);
        halfspace.splice(halfspace.begin(), f->halfspace);
        
        Facet4D* f1=f->neighbour1();
        Facet4D* f2=f->neighbour2();
        Facet4D* f3=f->neighbour3();
        Facet4D* f4=f->neighbour4();
        
        // delete every link to this facet
        if(f->neighbour1()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r1);
        }
        else
        {
            if(f->r1->f1==f)
            {
                f->r1->f1=NULL;
                if(f->r1->f2==NULL) ridgesToDelete.push_front(f->r1);//delete(f->r1);
            }
            else
            {
                if(f->r1->f2==f)
                {
                    f->r1->f2=NULL;
                    if(f->r1->f1==NULL)  ridgesToDelete.push_front(f->r1);//delete(f->r1);
                }
                else cout << "Ouch" << endl;
            }
        }
        
        if(f->neighbour2()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r2);
        }
        else
        {
            if(f->r2->f1==f)
            {
                f->r2->f1=NULL;
                if(f->r2->f2==NULL)  ridgesToDelete.push_front(f->r2);//delete(f->r2);
            }
            else
            {
                if(f->r2->f2==f)
                {
                    f->r2->f2=NULL;
                    if(f->r2->f1==NULL)  ridgesToDelete.push_front(f->r2);//delete(f->r2);
                }
                else cout << "Ouch" << endl;
            }
        }
        
        if(f->neighbour3()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r3);
        }
        else
        {
            if(f->r3->f1==f)
            {
                f->r3->f1=NULL;
                if(f->r3->f2==NULL)  ridgesToDelete.push_front(f->r3);//delete(f->r3);
            }
            else
            {
                if(f->r3->f2==f)
                {
                    f->r3->f2=NULL;
                    if(f->r3->f1==NULL)  ridgesToDelete.push_front(f->r3);//delete(f->r3);
                }
                else cout << "Ouch" << endl;
            }
        }
        
        if(f->neighbour4()==NULL) // next facet deleted : delete common ridge.
        {
            delete(f->r4);
        }
        else
        {
            if(f->r4->f1==f)
            {
                f->r4->f1=NULL;
                if(f->r4->f2==NULL)  ridgesToDelete.push_front(f->r4);//delete(f->r3);
            }
            else
            {
                if(f->r4->f2==f)
                {
                    f->r4->f2=NULL;
                    if(f->r4->f1==NULL)  ridgesToDelete.push_front(f->r4);//delete(f->r3);
                }
                else cout << "Ouch" << endl;
            }
        }
        // All links down
        //cout << "deleting " << f << endl;
        //delete(f);
        facetsToDelete.push_front(f);
        //f=NULL;
        removeVisibleFacets(f1, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        removeVisibleFacets(f2, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        removeVisibleFacets(f3, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        removeVisibleFacets(f4, ref, link, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
    }
    else
    {
        //f->status=2;
        link=f; // safe facet, won't be deleted
        if(f->neighbour1()==NULL)
        {
            orphanRidges.push_front(f->r1);
        }
        if(f->neighbour2()==NULL)
        {
            orphanRidges.push_front(f->r2);
        }
        if(f->neighbour3()==NULL)
        {
            orphanRidges.push_front(f->r3);
        }
        if(f->neighbour4()==NULL)
        {
            orphanRidges.push_front(f->r4);
        }
    }
}

Facet4D* buildFirstFacet(list<Point4D*>& points)
{
//     points.sort();
    
    // Building the first triangle
    
    Point4D* p1=points.front();
    Point4D* p2=points.back();
    points.pop_front();
    points.pop_back();
    Point4D* p3;
    Point4D* p4;
    Point4D* p5;
    
    Ridge4D *r123, *r124, *r125, *r134, *r135, *r145, *r234, *r235, *r245, *r345;
    
    // Look for the point maximizing the triangle area

    double current(0.), max(-1.), d12((*p2-*p1).norm()), d13, d23;
    
    
    for(list<Point4D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        d13=(**iter-*p1).norm();
        d23=(**iter-*p2).norm();
        current=(d12+d13+d23)*(-d12+d13+d23)*(d12-d13+d23)*(d12+d13-d23); // 1/4*sqrt normally, but we don't care
        if(max<0. || current>max)
        {
            max=current;
            p3=*iter;
        }
            
    }
    
//     points.remove(p3);
    
    r123=new Ridge4D(p1, p2, p3);
    if(r123==NULL) cout << "new r123 failed" << endl;
    
    // Look for the point maximizing the tetrahedron volume
    max=-1;
    d13=((*p3-*p1).norm2);
    d23=((*p3-*p2).norm2);
    d12=d12*d12;

    double d14, d24, d34;
    
    for(list<Point4D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        d14=(**iter-*p1).norm2;// d14=d14*d14;
        d24=(**iter-*p2).norm2;// d24=d24*d24;
        d34=(**iter-*p3).norm2;// d34=d34*d34;
        current=2.*(-d23*d34*d24-d12*d12*d34-d12*d34*d34-d13*d24*d24-d14*d23*d23+d12*d23*d34-d24*d13*d13-d23*d14*d14-d13*d34*d14+d13*d12*d34+d13*d24*d14+d14*d12*d34+d14*d23*d13-d12*d24*d14+d12*d24*d13+d14*d12*d23-d12*d23*d13+d12*d34*d24+d13*d34*d24+d13*d24*d23+d14*d23*d34+d14*d24*d23);
        // Thank you Maple

        if(max<0. || current>max)
        {
            max=current;
            p4=*iter;
        }
    }
//     points.remove(p4);
    r124=new Ridge4D(p1, p2, p4);
    r134=new Ridge4D(p1, p3, p4);
    r234=new Ridge4D(p2, p3, p4);
    
    Facet4D* f1234=new Facet4D(r123, r124, r134, r234);
    
    // Look for the farthest point
    
    max=-1;
    Point4D normal=(*p2-*p1).crossProd(*p3-*p1, *p4-*p1);
    for(list<Point4D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        current=fabs(normal.dotProd(**iter-f1234->centre));
        if(max<0. || current>max)
        {
            max=current;
            p5=*iter;
        }
    }
//     points.remove(p5);
    r125=new Ridge4D(p1, p2, p5);
    r135=new Ridge4D(p1, p3, p5);
    r145=new Ridge4D(p1, p4, p5);
    r235=new Ridge4D(p2, p3, p5);
    r245=new Ridge4D(p2, p4, p5);
    r345=new Ridge4D(p3, p4, p5);
    
    Facet4D* f1235=new Facet4D(r123, r125, r135, r235);
    Facet4D* f1245=new Facet4D(r124, r125, r145, r245);
    Facet4D* f1345=new Facet4D(r134, r135, r145, r345);
    Facet4D* f2345=new Facet4D(r234, r235, r245, r345);
    
    // Link the facets altogether
    
    r123->f1=f1234; r123->f2=f1235;
    r124->f1=f1234; r124->f2=f1245;
    r125->f1=f1235; r125->f2=f1245;
    r134->f1=f1345; r134->f2=f1234;
    r135->f1=f1345; r135->f2=f1235;
    r145->f1=f1245; r145->f2=f1345;
    r234->f1=f1234; r234->f2=f2345;
    r235->f1=f1235; r235->f2=f2345;
    r245->f1=f1245; r245->f2=f2345;
    r345->f1=f1345; r345->f2=f2345;

    f1234->setOuterNormal(p5);
    f1235->setOuterNormal(p4);
    f1245->setOuterNormal(p3);
    f1345->setOuterNormal(p2);
    f2345->setOuterNormal(p1);

    // Build associated lists
    
    for(list<Point4D*>::iterator iter=points.begin(); iter!=points.end(); iter++)
    {
        if(*iter!=p3 && *iter!=p4 && *iter!=p5)
        {
            if(f1234->outerNormal.dotProd(**iter-f1234->centre) > 0.)
            {
                f1234->halfspace.push_front(*iter);
            }
            if(f1235->outerNormal.dotProd(**iter-f1235->centre) > 0.)
            {
                f1235->halfspace.push_front(*iter);
            }
            if(f1245->outerNormal.dotProd(**iter-f1245->centre) > 0.)
            {
                f1245->halfspace.push_front(*iter);
            }
            if(f1345->outerNormal.dotProd(**iter-f1345->centre) > 0.)
            {
                f1345->halfspace.push_front(*iter);
            }
            if(f2345->outerNormal.dotProd(**iter-f2345->centre) > 0.)
            {
                f2345->halfspace.push_front(*iter);
            }
        }
    }

    return f1234;
}

bool sortCriterionRidges(Ridge4D* r1, Ridge4D* r2)
{
    return (*r1)<(*r2);
}

bool uniqueCriterionRidges(Ridge4D* r1, Ridge4D* r2)
{
    return (*r1)==(*r2);
}

void quickHull(list<Point4D*>& points, list<Element3D>& hull)
{
    Facet4D* firstFacet=buildFirstFacet(points);

    Facet4D* currentFacet=getImprovableFacet(firstFacet);
    Facet4D::currentStatus++;

    Point4D* np; // new point
    double max, current;
    Point4D centre;
    Facet4D* tempFacet;
    while(currentFacet!=NULL)
    {
        centre=currentFacet->centre;
        max=-1;

        for(list<Point4D*>::iterator iter=currentFacet->halfspace.begin(); iter!=currentFacet->halfspace.end(); iter++)
        {
            current=currentFacet->distance(*iter);
            if(max<current)
            {
                max=current;
                np=*iter;
            }
        }

        
        Facet4D* startFacet=currentFacet;
        list<Point4D*> halfspace;
        list<Ridge4D*> orphanRidges;
        
        list<Facet4D*> facetsToDelete;
        list<Ridge4D*> ridgesToDelete;
        removeVisibleFacets(startFacet, np, currentFacet, orphanRidges, halfspace, facetsToDelete, ridgesToDelete);
        Facet4D::currentStatus++;
        
        facetsToDelete.sort();
        facetsToDelete.unique();
        while(!facetsToDelete.empty())
        {
            delete facetsToDelete.front();
            facetsToDelete.pop_front();
        }
        ridgesToDelete.sort();
        ridgesToDelete.unique();
        while(!ridgesToDelete.empty())
        {
            delete ridgesToDelete.front();
            ridgesToDelete.pop_front();
        }
        halfspace.sort();
        halfspace.unique();
        
        orphanRidges.sort();
        orphanRidges.unique();
        
        list<Facet4D*> newFacets;
        list<Ridge4D*> newRidges;
        
        for(list<Ridge4D*>::iterator iter=orphanRidges.begin(); iter!=orphanRidges.end(); iter++)
        {
            Point4D* p1=(*iter)->p1;
            Point4D* p2=(*iter)->p2;
            Point4D* p3=(*iter)->p3;
           
            Ridge4D* r12n=getRidge(newRidges, p1, p2, np);
            Ridge4D* r13n=getRidge(newRidges, p1, p3, np);
            Ridge4D* r23n=getRidge(newRidges, p2, p3, np);
            
            Facet4D* fn =new Facet4D(*iter, r12n, r13n, r23n);
            
            currentFacet=fn;
            
            if(r12n->f1==NULL)
            {
                r12n->f1=fn;
            }
            else
            {
                r12n->f2=fn;
            }
            
            if(r13n->f1==NULL)
            {
                r13n->f1=fn;
            }
            else
            {
                r13n->f2=fn;
            }
            
            if(r23n->f1==NULL)
            {
                r23n->f1=fn;
            }
            else
            {
                r23n->f2=fn;
            }
            
            if((*iter)->f1==NULL)
            {
                (*iter)->f1=fn;
            }
            else
            {
                if((*iter)->f2==NULL)
                {
                    (*iter)->f2=fn;
                }
                else
                {
                    cout << "Double orphan!" << endl;
                }
            }
            
            fn->outerNormal=(*(fn->p2)-*(fn->p1)).crossProd(*(fn->p3)-*(fn->p1), *(fn->p4)-*(fn->p1));
            if(fn->outerNormal.dotProd(*np-centre)<=0.)
            {
                fn->outerNormal.opposite();
            }
            
            for(list<Point4D*>::iterator currentPoint=halfspace.begin(); currentPoint!=halfspace.end(); currentPoint++)
            {
                if(*currentPoint!=np && fn->distance(*currentPoint)>0.)
                    fn->halfspace.push_front(*currentPoint);
            }
            newFacets.push_front(fn);
        }

//         for(list<Facet4D*>::iterator f1=newFacets.begin(); f1!=newFacets.end(); f1++)
//         {
//             if((*f1)->neighbour1()==NULL)
//             {
//                 cout << "unmatched 1" << endl;
//                 exit(0);
//             }
//             
//             if((*f1)->neighbour2()==NULL)
//             {
//                 cout << "unmatched 2" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour3()==NULL)
//             {
//                 cout << "unmatched 3" << endl;
//                 exit(0);
//             }
//             
//             if((*f1)->neighbour4()==NULL)
//             {
//                 cout << "unmatched 4" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour1()==(*f1)->neighbour2())
//             {
//                 cout << "Twice the same neighbour 1/2" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour1()==(*f1)->neighbour3())
//             {
//                 cout << "Twice the same neighbour 1/3" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour1()==(*f1)->neighbour4())
//             {
//                 cout << "Twice the same neighbour 1/4" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour2()==(*f1)->neighbour3())
//             {
//                 cout << "Twice the same neighbour 2/3" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour2()==(*f1)->neighbour4())
//             {
//                 cout << "Twice the same neighbour 2/4" << endl;
//                 exit(0);
//             }
//             if((*f1)->neighbour3()==(*f1)->neighbour4())
//             {
//                 cout << "Twice the same neighbour 3/4" << endl;
//                 exit(0);
//             }
//         }

        tempFacet=getImprovableFacet(currentFacet);
        Facet4D::currentStatus++;

        if(tempFacet!=NULL)
        {
            currentFacet=tempFacet;

        }
        else
        {
            convertToElementsList(currentFacet, hull);
            Facet4D::currentStatus++;

            cleanUp(currentFacet, facetsToDelete, ridgesToDelete);
            facetsToDelete.sort();
            facetsToDelete.unique();
            ridgesToDelete.sort();
            ridgesToDelete.unique();
            
            while(!facetsToDelete.empty())
            {
                delete facetsToDelete.front();
                facetsToDelete.pop_front();
            }
            while(!ridgesToDelete.empty())
            {
                delete ridgesToDelete.front();
                ridgesToDelete.pop_front();
            }

            currentFacet=tempFacet;
        }
    }
}
