/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    VectorTSDecomp
* 
* MODULE:	QSAS Decomposition of Vector Time Series
*
* LANGUAGE:	C++
*
* FUNCTION:
*
* PURPOSE: Decompose a vector time series into a Scalar Time Series of
*                    one of its components (x,y,z,r,theta,phi)
* 
* DESCRIPTION:
*
* ARGUMENTS:
*    IN:
*     VECTOR_TIME_SERIES    ts_in                 input vector time series 
*                                                  data object
*     INTEGER               componentdo           flag for which component   
*    OUT:
*        SCALAR_TIME_SERIES ts_out
* GLOBALS:
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
*
* FUNCTIONS_USED:	 QpAxisLabelMakeStem - from QPUI to make label stem
*
* ALGORITHM:   pretty obvious, most of the work is in the attributes
*
* HISTORY:		Steve Schwartz, July 1996
*   Amended to use Qfui interface instead of Qai - A R Hare, 12 Sep 96.
*   SJS  19 May 97 - amended to use Qplug instead of Qplug 
*   SJS  30 June 97 - fixed some memory leaks
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "Qdos.h"
#include "qdutil.h"
#include "VectorTSDecomp.h"

using namespace QSAS;

QplugReturnStatus  VectorTSDecomp  ( QplugArgList *call_list )

{

QdObject_var  ts_in;
QdObject_var  ts_out;
QdRScalar_var  componentdo;
int             icomponent;

char *string_ptr;
char fieldnam[150];                  // attributes
char siconversion[150];
char units[150];
char lablaxis[150];
char stem[150];
double hs,hsp;

int    ndata_in;                                  /* number of data points  */
                                                  /* array new'ed below  */

//
//	...first, unpack the input data objects
//

//	..first item is input Time Series

ts_in = (*call_list)[0];

if ( !QuIsThreeVector(ts_in) ) {	
   QplugAppendTextDisplay (
"VectorTSDecomp>> ERROR - input not a Vector Time Series!\n");
	return QPLUG_FAILURE; 
}        

ndata_in = ts_in->sequence_size();


//	2nd item integer flag for component (order same as PUI):
//           0 = x
//           1 = y
//           2 = z
//           3 = r(mag)
//           4 = Lat(deg)
//           5 = theta(deg)
//           6 = phi(deg)
//           7 = Lat(rad)
//           8 = theta(rad)
//           9 = phi(rad)


componentdo = QdRScalar_var::narrow((*call_list)[1]);

 if ( componentdo.is_nil() ) { 
    QplugAppendTextDisplay (
        "VectorTSDecomp>> ERROR - argument not an integer flag!\n");
    return QPLUG_FAILURE; 
}  // end if

 icomponent=(int)(*componentdo);

 strcpy(stem,QpAxisLabelMakeStem(ts_in) );            // Make stem for label

 string_ptr = QuGetAttrText("FIELDNAM", ts_in);
 strcpy(fieldnam,string_ptr);
       
       
switch(icomponent){
     case(0):            // x component
        string_ptr = QuGetAttrText("UNITS", ts_in);
        strcpy(units,string_ptr);
             
        string_ptr = QuGetAttrText("SI_conversion", ts_in);
        if( strcmp(string_ptr, "") != 0){
          strcpy(siconversion, string_ptr);
        } 
        else {                                    // set SI_conversion
          strcpy(siconversion,"1>(unknown)");     //  empty if fail
        }      
        
        strcpy(lablaxis,stem);
        strcat(lablaxis,"_x");
        strcat(fieldnam,"_x");
        ts_out = QuGetVectorComponent(ts_in, "x");
        break;
     case(1):            // y component
        string_ptr = QuGetAttrText("UNITS", ts_in);
        strcpy(units,string_ptr);

        string_ptr = QuGetAttrText("SI_conversion", ts_in);
        if( strcmp(string_ptr, "") != 0){
          strcpy(siconversion, string_ptr);
        } 
        else {                                    // set SI_conversion
          strcpy(siconversion,"1>(unknown)");     //  empty if fail
        }      
        
        strcpy(lablaxis,stem);
        strcat(lablaxis,"_y");
        strcat(fieldnam,"_y");
        ts_out = QuGetVectorComponent(ts_in, "y");
        break;
     case(2):            // z component
        string_ptr = QuGetAttrText("UNITS", ts_in);
        strcpy(units,string_ptr);

        string_ptr = QuGetAttrText("SI_conversion", ts_in);
        if( strcmp(string_ptr, "") != 0){
          strcpy(siconversion, string_ptr);
        } 
        else {                                    // set SI_conversion
          strcpy(siconversion,"1>(unknown)");     //  empty if fail
        }      

        strcpy(lablaxis,stem);
        strcat(lablaxis,"_z");
        strcat(fieldnam,"_z");
        ts_out = QuGetVectorComponent(ts_in, "z");
        break;
     case(3):            // r component
        string_ptr = QuGetAttrText("UNITS", ts_in);
        strcpy(units,string_ptr);

        string_ptr = QuGetAttrText("SI_conversion", ts_in);
        if( strcmp(string_ptr, "") != 0){
          strcpy(siconversion, string_ptr);
        } 
        else {                                    // set SI_conversion
          strcpy(siconversion,"1>(unknown)");     //  empty if fail
        }      

        strcpy(lablaxis,"|");
        strcat(lablaxis,stem);
        strcat(lablaxis,"|");
        strcat(fieldnam,"_magnitude");
        ts_out = QuGetVectorComponent(ts_in, "r(mag)");
        break;
     case(4):            // Lat(deg) component
        strcpy(units,"deg"); 
        strcpy(siconversion,"1>degree");
        strcpy(lablaxis,"Lat");
        strcat(lablaxis,"_");
        strcat(lablaxis,stem);
        strcat(fieldnam,"_Lat(deg)");
        ts_out = QuGetVectorComponent(ts_in, "Lat(deg)");
        break;
     case(5):            // theta(deg) component
        strcpy(units,"deg"); 
        strcpy(siconversion,"1>degree");
        strcpy(lablaxis,"theta");
        strcat(lablaxis,"_");
        strcat(lablaxis,stem);
        strcat(fieldnam,"_theta(deg)");
        ts_out = QuGetVectorComponent(ts_in, "theta(deg)");
        break;
     case(6):            // phi(deg) component
        strcpy(units,"deg"); 
        strcpy(siconversion,"1>degree");
        strcpy(lablaxis,"phi");
        strcat(lablaxis,"_");
        strcat(lablaxis,stem);
        strcat(fieldnam,"_phi(deg)");
        ts_out = QuGetVectorComponent(ts_in, "phi(deg)");
        break;
     case(7):            // Lat(rad) component
        strcpy(units,"rad"); 
        strcpy(siconversion,"1>rad");
        strcpy(lablaxis,"Lat");
        strcat(lablaxis,"_");
        strcat(lablaxis,stem);
        strcat(fieldnam,"_Lat(rad)");
        ts_out = QuGetVectorComponent(ts_in, "Lat(rad)");
        break;
     case(8):            // theta(rad) component
        strcpy(units,"rad"); 
        strcpy(siconversion,"1>rad");
        strcpy(lablaxis,"theta");
        strcat(lablaxis,"_");
        strcat(lablaxis,stem);
        strcat(fieldnam,"_theta(rad)");
        ts_out = QuGetVectorComponent(ts_in, "theta(rad)");
        break;
     case(9):            // phi(rad) component
        strcpy(units,"rad"); 
        strcpy(siconversion,"1>rad");
        strcpy(lablaxis,"phi");
        strcat(lablaxis,"_");
        strcat(lablaxis,stem);
        strcat(fieldnam,"_phi(deg)");
        ts_out = QuGetVectorComponent(ts_in, "phi(rad)");
        break;
     default:                     // Shouldn't get here!   
        QplugAppendTextDisplay 
           ("VectorTSDecompAVG>> Unknown component requested; Failure\n");
        return QPLUG_FAILURE;
        break;
}                                 // end switch(icomp)


// Create the output Time Series and fill it

 QdTimeSeq_var tt = get_timetags(ts_in);
 QdTimeSeq_var tt_out = (QdTimeSeq_var) Factory::clone(tt);
 
 set_timetags(ts_out, tt_out);

// load all the attributes, and the half spacing/sample period:

 QuSetTxtAttr("FIELDNAM", fieldnam, ts_out);
 QuSetFrame(ts_out, "scalar");
 QuSetTxtAttr("SI_conversion", siconversion, ts_out);
 QuSetTxtAttr("UNITS", units, ts_out);
 QuSetTxtAttr("LABLAXIS", lablaxis, ts_out);

//	..add  the output series to the list

  call_list->push_back(ts_out);     

// Clean up:


QplugAppendTextDisplay ("VectorTSDecomp>> data object created - exiting successfully\n");

return QPLUG_SUCCESS;

}
