/*************************************************************************
* PROJECT:      Cluster UK CDHF
* COMPONENT:    Plug In library
* MODULE:	XY2Polar.cc
* LANGUAGE:	C++
* FUNCTION:     QSAS XY2Polar Plug In
* PURPOSE:      Calculate radius and phase of two scalar series.  
* ARGUMENTS:
*    IN:    Scalar series X.  
*           Scalar series Y.
*   OUT:    Radius = SQRT(X^2+Y^2) Scalar series
*           Phase  = ATAN2(Y,X) Scalar series
* RETURN:   QplugReturnStatus
* FUNCTIONS_USED: 
* HISTORY: created  15/3/2001 Markus Fraenz 
*          corrected atan2(xx,yy) to atan2(yy,xx) AJA 18/08/2005
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

using namespace QSAS;

extern "C" QplugReturnStatus  XYtoPolar( QplugArgList *);

QplugReturnStatus  XYtoPolar( QplugArgList *call_list )
{
// define some C variables
int  qar_status;
unsigned long	Ndata;
char str_out[150];

// Display initial message in QpluGui window 
QplugAppendTextDisplay("XY2Polar Plugin 1.0 11 Apr 2001\n");

// ...first, unpack the input data objects
QdObject_var X_in = (* call_list)[0];  
QdObject_var Y_in = (* call_list)[1];

//check whether these are scalar series, if not return.
QdRScalarSeq_var xx = QdRScalarSeq_var::narrow(X_in);
if(xx.is_nil() ){
   QplugAppendTextDisplay ("XY2Polar>> ERROR - X sequence type not handled\n");
   return QPLUG_FAILURE;  
}
QdRScalarSeq_var yy = QdRScalarSeq_var::narrow(Y_in);
if(yy.is_nil() ){
   QplugAppendTextDisplay ("XY2Polar>> ERROR - Y sequence type not handled\n");
   return QPLUG_FAILURE;  
}

Ndata = xx->sequence_size();

//compare sequence sizes, and if time series test if joined

// check whether input conformal
qar_status =  QarTestConformal(X_in, Y_in);
if(qar_status == QAR_ERR)  
  {
    QplugAppendTextDisplay ("Arithmetic>> input sequences not joined properly\n");
     return QPLUG_FAILURE; 
  }


// check whether both have same units
qar_status = QarAreUnitsSame(X_in, Y_in);
if(qar_status == QAR_ERR) {
  QplugAppendTextDisplay (
     "Arithmetic>> This Operation requires inputs in same units (double click object and correct SI_conversion)\n");
  return QPLUG_FAILURE; 
}
else if(qar_status == QAR_WARN){
  QplugAppendTextDisplay (
    "Arithmetic>> Warning: units (SI_conversion) unknown for at least one input\n");
  QplugAppendTextDisplay ("Proceed with caution!\n");
}
  
//now tell user that plugin is actually executing
QplugAppendTextDisplay ("XY2Polar>> Calculating.\n");

//create output data objects
QdRScalarSeq_var out_rad_series;
QdRScalarSeq_var out_phas_series;

//calculate radius and phase
try{
  out_rad_series = QdRScalarSeq_var::narrow(sqrt(xx*xx + yy*yy));
  out_phas_series = QdRScalarSeq_var::narrow(atan2(yy,xx));
}
catch(Exception &e){
   e.print_msg();
   return QPLUG_FAILURE;   
}
 
 // copy xrefs before setting time as time tags are also Xrefs.
out_rad_series->copy_xrefs_from(X_in);

//get timetags from X or Y and add to output objects
QdTimeSeq_var tags = get_timetags(X_in);
if (tags.is_nil()) tags = get_timetags(Y_in);
if (!tags.is_nil()) {
   set_timetags(out_rad_series, tags);
   set_timetags(out_phas_series, tags);
}

//set minimal attributes by copying from input objects
  QuSetFrame((QdObject_var) out_rad_series, "scalar");
  sprintf(str_out,"r(%s)", QuGetAttrText("FIELDNAM", X_in));
  QuSetTxtAttr("LABLAXIS", str_out, (QdObject_var) out_rad_series);
  QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_rad_series);

  QuSetTxtAttr("SI_conversion", "1>rad", (QdObject_var) out_phas_series);
  sprintf(str_out,"Phase(%s)", QuGetAttrText("FIELDNAM", X_in));
  QuSetTxtAttr("LABLAXIS", str_out, (QdObject_var) out_phas_series);
  QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_phas_series);
  QuSetFrame((QdObject_var) out_phas_series, "scalar");
  QuSetTxtAttr("UNITS", "rad", (QdObject_var) out_phas_series);

//add output objects to  QSAS working list
  call_list->push_back( (QdObject_var) out_rad_series);    
  call_list->push_back( (QdObject_var) out_phas_series);     


//tell user that Plug In finished
sprintf(str_out, "XY2Polar>> %i data points\n", Ndata);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("XY2Polar>> finished...\n");


return QPLUG_SUCCESS;
}



