/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#include "qie_caa.h"
#include "cdf.h"
//#include "qie.h"

   
/**********************************************************************
************************ START OF CDF DATA IMPORT FUNCTIONS ***********
***********************************************************************/

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
*
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QieGetCSDSgenCDF
*
* PURPOSE: Get data for whole CDF file. Metadata are retrieved separately.
*
* DESCRIPTION: Opens named file and reads each variable in turn.
*              r-variables are coerced into z-variables on the fly.
*
* GLOBALS:
*
* RETURN:
*    TYPE:   int
*    VALUES: QMW_OK         Executed successfully.
*            CDF_OPEN_ERR   Unable to open named file.
*            CDF_N_VAR_ERR  Unable to get number of variables.
*
* FUNCTIONS_USED:
*
*
* ALGORITHM:
*
*
***********************************************************************/

bool QieData::queryCDFfile(const char * fullName){
   
  CDFstatus status;
  CDFid id;
  long n_vars;
 
  inFile.name = (char *) malloc(sizeof(char) *(strlen(fullName) + 1) );
  strcpy(inFile.name, fullName); // used only for setting Parents attribute for QBU browser
  
   // open named cdf file 
   status = CDFlib(OPEN_, CDF_, fullName, &id, NULL_);
     if (status != CDF_OK){
     	QieMsg ("ERROR for '", fullName, "' Cannot open file\n");
	return false;
     }
     
   // set to z mode to convert r to z variables 
    CDFlib(SELECT_, CDF_zMODE_, zMODEon2, NULL_);

   status = CDFlib(GET_, CDF_NUMzVARS_, &n_vars,
                         zVARs_MAXREC_, &maxCDFnumRecs,
                   NULL_);
     if (status != CDF_OK) {
        QieMsg ("ERROR for '", fullName, "' Cannot count variables or records\n");
        return false;
     }
     
     // count vars and initialise variable objects
 		    
     if( !getCDFvariables(n_vars) ){
        QieMsg ("ERROR for '", fullName, "' Cannot get information on Variables\n");
        return false;
     }
     
     // Get Global attributes
    
     getCDFglobals();
     
     // Read  Attribute Helper file if present
    
     getAttributeHelper(fullName);
       
     ensureDependZero();
     
     return true;
     
}

// --------------------------------------------------------------------------- 


bool QieData::getCDFvariables(int n_vars){

   for(int n=0; n<n_vars; n++){
      QieVars.push_back(getCDFVar(n));  
   }
   
 return true;
   
} 
// --------------------------------------------------------------------------- 

QieVariable QieData::getCDFVar(int nvar){
  
  char name[CDF_VAR_NAME_LEN+1];
  size_t length;
  int size_of_datatype;
  long dim_sizes[CDF_MAX_DIMS];
  long dim_varys[CDF_MAX_DIMS];
  long data_type;
  long recVary;
  long nDims;
  long max_rec_num;
  
  QieVariable nextVar;
  nextVar.name = " ";
  nextVar.col_start = -1; // unused for cdf
  nextVar.col_width = -1; // unused for cdf
//	  nextVar.ftpackets.timeformat = NOT_A_TIME; // SJS re FT - safe initialisation
  nextVar.recVary = false;
  nextVar.datatype = string("CHAR");
  nextVar.n_posn = -1; // unused for cdf
  nextVar.cachedFrom = -1; // used to keep rec number for data range read already
  nextVar.cachedTo = -1;   // used to keep rec number for data range read already

  // select the variable and its V attributes
  if ( CDFlib(SELECT_, zVAR_, nvar,
                       zENTRY_, nvar,
                       NULL_)!= CDF_OK ) {
	cout << "Failed to select variable number " << nvar << endl;
	return nextVar;
  }
  

  if ( CDFlib(GET_, zVAR_NAME_, &name[0], NULL_) != CDF_OK) {
     cout << "Failed to get name of variable number " << nvar << endl;
     return nextVar;
  }
  nextVar.name = &name[0];

 
  if( CDFlib(GET_, zVAR_NUMDIMS_, &nDims, NULL_) != CDF_OK) {
     cout << "Failed to get number of dimensions of variable '" << nextVar.name.c_str() << "'" << endl;
     return nextVar;
  }

  if (nDims > 0 ){

    if( CDFlib(GET_, zVAR_DIMSIZES_, &dim_sizes[0], 
                     zVAR_DIMVARYS_, &dim_varys[0],
                     NULL_) != CDF_OK) {
      cout << "Failed to get dimension sizes of variable '" << nextVar.name.c_str() << "'" << endl;
      return nextVar;
    }
    nextVar.objDim.clear();
    for(int k=0; k<nDims; k++){
       if(dim_varys[k] == NOVARY) continue;
       int dimSize = dim_sizes[k];
       nextVar.objDim.push_back(dimSize);
    }

  }

  if( CDFlib(GET_, zVAR_MAXREC_, &max_rec_num, NULL_) != CDF_OK) {
     cout << "Failed to get number of records for variable '" << nextVar.name.c_str() << "'" << endl;
     return nextVar;
  }
  if (max_rec_num < 0)   nextVar.nRecs = 0;  // Virtual variable, no records written
  else                   nextVar.nRecs = max_rec_num+1;

   if( CDFlib(GET_, zVAR_RECVARY_, &recVary, NULL_)!= CDF_OK) {
    cout << " Failed to find record variance for variable '" << nextVar.name.c_str() << "'" << endl;
    return nextVar;
  }
  if (recVary ==  VARY) nextVar.recVary = true;
  else nextVar.recVary = false;

  if (CDFlib(GET_, zVAR_DATATYPE_, &data_type,
                  NULL_) != CDF_OK){
     cout << "Failed to get data type of variable '" << nextVar.name.c_str() << "'" << endl;
     return nextVar;
  }
  if( data_type == CDF_EPOCH || data_type == CDF_EPOCH16){
     if(time_var_num == -1) time_var_num = nvar; // keep as default only used in case of no DEPEND_0
     nextVar.datatype = string("ISO_TIME");
  }
  
  //  create dummy variable until data read in
  nextVar.Vobj = new QdString("No Data Read from CDF for this variable");
      
  // add V attributes now that variable exists
  addVarAttrs(nextVar, nvar); // navr is cdf var num which may differ from index of QieVar vector
  
  return nextVar;
}
// ----------------------------------------------------------------------------------------

void QieData::addVarAttrs(QieVariable &V, int nvar){


long i;
long max_num_attrs;
long data_type;
long num_elems;
long attr_scope;
char attribute_name[ CDF_ATTR_NAME_LEN +1 ];

  // VARIABLE and zENTRY_ already selected
  
  if(CDFlib( GET_, CDF_NUMATTRS_, &max_num_attrs, NULL_ ) != CDF_OK ){
    cout << "Failed to get number of attributes for variable '" << V.name.c_str() << "'" << endl;
    return;
  }


  // count entries as we go, zENTRY_ already selected by calling routine 

  // read attribute entries: if we fail, try next in list 

  for( i=0 ; i < max_num_attrs ; i++ )
  {
    
     if( CDFlib( SELECT_, ATTR_, i, NULL_) != CDF_OK ) continue;

     if( CDFlib( GET_, ATTR_SCOPE_,  &attr_scope, NULL_) != CDF_OK ) continue;
    
     if(attr_scope != VARIABLE_SCOPE) continue; // skip globals here

     if( CDFlib( CONFIRM_, zENTRY_EXISTENCE_, (long) nvar, NULL_) != CDF_OK ) continue;
 
     if( CDFlib( GET_, ATTR_NAME_, &(attribute_name[0]), NULL_ ) != CDF_OK ) continue;

     if( CDFlib( GET_, zENTRY_DATATYPE_, &data_type,
                       zENTRY_NUMELEMS_, &num_elems,
                       NULL_ ) != CDF_OK ) continue;

    // allocate entry space

     void *data = QdaAttrEntryMalloc( num_elems, data_type );

     if(  CDFlib( GET_,  zENTRY_DATA_, data, NULL_ ) != CDF_OK ) continue;

     // Note NULL terminator for char is already handled by  QiAttrEntryMalloc for char types

     // convert to QDOS object
     QdObject_var attrObj = QdaCDFattrBufferToObj(data, data_type, num_elems);

     V.Vobj->change_xref(attribute_name, attrObj); // use change_xref as it is safe against duplicate attr in file
     
     // keep on lists as well as these will be used to convert pointers to vars
     V.attrList.push_back(attrObj);
     V.attrNameList.push_back(std::string(attribute_name));

  } // end for loop over attributes 

  return;

}

// -------------------------------------------------------------

void *QdaAttrEntryMalloc( long num_elems, long data_type )
{
void *p;
size_t size;

if( num_elems < 1 ) num_elems = 1;   /* some defensive programming ... */

switch (data_type) {
  case CDF_BYTE: case CDF_INT1: case CDF_UINT1:
  {
    size = num_elems;
    break;
  }
  case CDF_INT2: case CDF_UINT2:
  {
    size = 2 * num_elems;
    break;
  }
  case CDF_INT4: case CDF_UINT4: case CDF_REAL4: case CDF_FLOAT:
  {
    size = 4 * num_elems;
    break;
  }

  case CDF_REAL8:  case CDF_EPOCH: case CDF_DOUBLE:
  {
    size = 8 * num_elems;
    break;
  }

  case CDF_EPOCH16: 
  {
    size = 16 * num_elems;
    break;
  }

  case CDF_CHAR: case CDF_UCHAR:
  {
    size = num_elems + 1;
    break;
  }
  default: size = 8 * num_elems;
}

p = (void *)malloc( size );

if( data_type == CDF_CHAR || data_type == CDF_UCHAR )
  ((char *)p)[num_elems] = '\0' ;

return p;

}   

// ------------------------------------------------------------

QdObject_var QdaCDFattrBufferToObj(void *data, int data_type, int num_elems)
{
    QdObject_var a_obj;
  
    // For V Attrs or G attr entries 
    // switch on data type
    // either string, ISO_TIME or cast to double
    
    DataBuffer_var data_buffer;
    if( data_type == CDF_CHAR || data_type == CDF_UCHAR){
      
       // cdf uses num_elems for string length but we also NULL terminate
        
       char *p = ( char*) data;
         
       int strLen = (int)strlen(p);
       if(strLen > num_elems) strLen = num_elems; // for safety only
       while ( strLen > 0 && p[strLen-1] == ' ' ) strLen--; // remove trailing spaces
       
       string sdata(p, strLen);
       QdString_var strObj = new QdString(sdata);
       a_obj = strObj;

      
    }
    else if( data_type == CDF_EPOCH ){
      
      vector<int> dims;
      int n_length = (int) num_elems;    
      if( n_length == 1){
         // normal case of single time
         double *dp = (double *)data;
         QdTime_var tv = new QdTime(cdfEpoch( *dp));
	 
         a_obj = tv;  
      }
      else if( n_length == 2) {
         // special case if time interval

         double *dp = (double *)data;
	 
	 QdTimeInterval_var tiv = new QdTimeInterval(QdTime(cdfEpoch(  *( static_cast<double*>(dp)  ))), 
	                                             QdTime(cdfEpoch(  *( static_cast<double*>(dp) + 1 ))));
   
         a_obj = tiv;
      }
      else {
        // QSAS/QTRAN/CAA cannot handle arrays of times, only sequences
      }
      
    }
    else if( data_type == CDF_EPOCH16 ){
      
      // epoch 16 uses seconds in field 1 and picoseconds in field 2
      double *dp = (double *)data;
      QdTime_var tv = new QdTime(cdfEpoch16(*( static_cast<double*>(dp)) ));
      double picosec = *( static_cast<double*>(dp) + 1 );
  
      *tv += Seconds(picosec*1.0e-12);
     
      a_obj = tv;

    }
    else{
      // cast all other data types to double matrix

      NumericVDataBuffer *ndb_p = new NumericVDataBuffer();
      
      vector<int> dims;
      int n_length = (int) num_elems;    
      if( n_length > 1) dims.push_back( n_length );
      ndb_p->set_dims( dims );
      ndb_p->set_majority( DBUFFER_ROW_MAJOR );

      // for numeric types use element by element static casting
      //   and let ndb_p->push_back(double) do the conversion to double

      switch( data_type )
      {
        case CDF_BYTE:
        case CDF_INT1:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<char*>(data) + i ) );
          break;
        case CDF_UINT1:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<unsigned char*>(data) + i ) );
          break;
        case CDF_INT2:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<short*>(data) + i ) );
          break;
        case CDF_UINT2:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<unsigned short*>(data) + i ) );
          break;
        case CDF_INT4:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<int*>(data) + i ) );
          break;
        case CDF_UINT4:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<unsigned int*>(data) + i ) );
          break;
        case CDF_REAL4:
        case CDF_FLOAT:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<float*>(data) + i ) );
          break;
        case CDF_REAL8:
        case CDF_DOUBLE:
          for(int i=0; i<n_length; i++)
	    ndb_p->push_back( *( static_cast<double*>(data) + i ) );
          break;
      }

      data_buffer=ndb_p;
      a_obj = Factory::clone(data_buffer);


  } // end else on string vs double

   return a_obj;
}

// --------------------------------------------------------------------

void QieData::getCDFglobals(){

  long num_attrs;
  long attr_scope;

  /* loop through all attributes and get G attrs */

  if( CDFlib( GET_, CDF_NUMATTRS_, &num_attrs, NULL_) != CDF_OK ){
    cout << "Failed to get number of attributes " << endl;
    return;
  }
  
  for(int i=0 ; i < (int) num_attrs ; i++ )
  {
    if( CDFlib( SELECT_, ATTR_, i, NULL_) != CDF_OK ) continue;

    if( CDFlib( GET_, ATTR_SCOPE_,  &attr_scope, NULL_) != CDF_OK ) continue;

    if( attr_scope != GLOBAL_SCOPE ) continue;  // skip V attrs

    QdObject_var nextGattr = getGlobalAttr();
    
    if ( !nextGattr.is_nil() ) QieGlobals.push_back(nextGattr);
       

  } // end loop over num_attrs



  return;

}

// --------------------------------------------------

QdObject_var QieData::getGlobalAttr()
{
  QdObject_var Gobj;
  
  char attribute_name[ CDF_ATTR_NAME_LEN +1 ];
  long num_entries;
  long data_type;
  long num_elems;
  void *data;
  
  // selection of next G attr by number already done
  
  if(  CDFlib( GET_,
             ATTR_NAME_, &(attribute_name[0]),
             ATTR_MAXgENTRY_, &num_entries,
             NULL_ )!= CDF_OK ){
     cout << "Failed to get name for Global attribute " << endl;
     return Gobj;
  }
  
   // actually what comes back is the maximum gEntry number, but we want
   // to use the number of gEntries _including_ any gEntries which
   // have no data (ie ones which "don't exist")
   // This is why we do not use ATTR_NUMgENTRIES_
   // therefore increment ..

  num_entries++;

  if(num_entries <= 0) return Gobj;
  
  long data_type_0 =0;
  long num_elems_0 =0;
  vector <QdObject_var> SeqObj;
  
  for(int i = 0; i < num_entries; i++ ){
    if( CDFlib( CONFIRM_, gENTRY_EXISTENCE_, i, NULL_) != CDF_OK ) continue;
    
    if( CDFlib( SELECT_, gENTRY_, i,
                GET_, gENTRY_DATATYPE_, &data_type,
                      gENTRY_NUMELEMS_, &num_elems,
                      NULL_ ) != CDF_OK ){
       cout << "Failed to get data type/number of elements for Global attribute " << endl;
       return Gobj;
    }
    if( i>0 ) {
       if( data_type != data_type_0) continue; // skip inhomogeneous data entries, QSAS can't handle
       if( num_elems != num_elems_0) continue; // skip inhomogeneous data entries, QSAS can't handle
    }
    else {
       data_type_0 = data_type;
       num_elems_0 = num_elems;
    }
   
    
    data = QdaAttrEntryMalloc( num_elems, data_type );

    if( CDFlib( GET_, gENTRY_DATA_, data, NULL_ ) != CDF_OK) {
      cout << "Failed to get data for Global attribute " << endl;
      return Gobj;
    }
     
    // turn void * buffer into object
    QdObject_var nextGobj = QdaCDFattrBufferToObj(data, data_type, num_elems);
    SeqObj.push_back(nextGobj);
    
  } /* end loop over entries */
  
  // Handle single and multiple entries separately
  // empty and inhomogeneous entries have been dropped, so use vector size()
  if( SeqObj.size() == 1){
       Gobj = SeqObj[0];
  }
  else{
     switch (data_type) {
        case CDF_BYTE: 
        case CDF_INT1: 
        case CDF_UINT1:
        case CDF_REAL8:  
	case CDF_DOUBLE:
        { 
	   if(num_elems == 1){
	      QdRScalarSeq_var ss = new QdRScalarSeq(SeqObj.size());
              for (int k=0; k<SeqObj.size(); k++){
	         ss[k] = *(QdRScalar_var::narrow(SeqObj[k]));
              }
	      Gobj = ss;
           }
	   else{
	      QdRMatrixSeq_var ms = new QdRMatrixSeq(SeqObj.size(), Dim(num_elems));
              for (int k=0; k<SeqObj.size(); k++){
	         ms[k] = *(QdRMatrix_var::narrow(SeqObj[k]));
              }
	      Gobj = ms;
	   }
            break;
        }
        case CDF_EPOCH: 
        {
           if(num_elems == 1){
	      QdTimeSeq_var ts = new QdTimeSeq(SeqObj.size());
              for (int k=0; k<SeqObj.size(); k++){
	         ts[k] = *(QdTime_var::narrow(SeqObj[k]));
              }
	      Gobj = ts;
           }
	   else{
	      // qdos cannot handle sequence of time arrays or sequence of time intervals
	      return Gobj; 
	   }
            break;
        }
        case CDF_EPOCH16: {
	
           if(num_elems == 1){
	      QdTimeSeq_var ts = new QdTimeSeq(SeqObj.size());
              for (int k=0; k<SeqObj.size(); k++){
	         ts[k] = *(QdTime_var::narrow(SeqObj[k]));
              }
	      Gobj = ts;
           }
	   else{
	      // qdos cannot handle sequence of time arrays or sequence of time intervals
	      return Gobj; 
	   }
	   break;
	}
        case CDF_CHAR: case CDF_UCHAR:
        {
		QdStringSeq_var strs = new QdStringSeq(SeqObj.size());
		for (int k=0; k<SeqObj.size(); k++){
			strs[k] = *(QdString_var::narrow(SeqObj[k]));
		}
		Gobj = strs;
		break;
        }
        default: 
	   cout << " Unknown data type for Global Attribute: " << attribute_name << endl;
           return Gobj;
     } // end switch
  
  } // end if on sequence size
  
  // attach name for G list viewing
  QdString_var nameXref =  new QdString(attribute_name);
  Gobj->change_xref("name", nameXref);

  return Gobj;

} 

 // -------------------------------------------------------------------------
 void QieData::getCDFvarDataList(vector <int> getList, int recStart, int recEnd){

    for (int i=0; i<getList.size(); i++){

       getCDFvarData(getList[i], recStart, recEnd);
       
    }
   
    // now all dependent vars and attrs have been read we can fix Themis epoch
  
    fixThemisEpoch(getList);
    // The variables needed are read into their Vobj pointers for the records requested

     for(int i=0; i<getList.size(); i++){
	attachXrefs(&(QieVars[getList[i]]));
     }
	
  
 }
 
// --------------------------------------------------------------------------- 

void QieData::getCDFvarData(int nvar, int recStart, int recEnd){

  long dim_indices[CDF_MAX_DIMS];
  long dim_counts[CDF_MAX_DIMS];
  long dim_varys[CDF_MAX_DIMS];
  long data_type;
  long num_elems=1;
  int strLength=1;
  size_t size_of_datatype =1;
  int nRecsToGet;
  
  if(nvar >= QieVars.size()) return;
  
  QieVariable * V = &(QieVars[nvar]);
  
  if(recEnd == -1) recEnd = V->nRecs-1; // cope with request for xref to virtual object

 // if( V->cachedFrom == recStart && V->cachedTo == recEnd) return; // variable already read for this range
  //else {
    // V->cachedFrom = recStart;
     //V->cachedTo = recEnd;
  //}
  
  // select the variable 
  if ( CDFlib(SELECT_, zVAR_, (long) nvar,
                       NULL_)!= CDF_OK ) {
	cout << "Failed to select variable number " << nvar << " for variable named '" << V->name.c_str() << "'" << endl;
	return;
  }

  if (CDFlib(GET_, zVAR_DATATYPE_, &data_type,
                  zVAR_DIMVARYS_, &(dim_varys[0]),
                  NULL_) != CDF_OK){
     cout << "Failed to get data type or dimension varys of variable '" << V->name.c_str() << "'" << endl;
     return;
  }
  
    if( CDFlib(GET_, zVAR_NUMELEMS_, &num_elems, NULL_) != CDF_OK){
    cout << " Failed to find number of elements for variable '" << V->name.c_str() << "'" << endl;
    return;
  }

  // convert to CAA text used on QSAS side
  switch(data_type){
    case CDF_CHAR:
    case CDF_UCHAR:
    {
      V->datatype = "CHAR";
      size_of_datatype = sizeof(char);
      strLength = num_elems;
      break;
    }
    case CDF_UINT1:
    case CDF_INT1:
    case CDF_BYTE:
    {
      V->datatype = "CHAR";
      size_of_datatype = sizeof(char);
      break;
    }
    case CDF_INT2:
    case CDF_UINT2:
    {
      V->datatype = "INT";
      size_of_datatype = 2;
      break;
    }
    case CDF_INT4:
    case CDF_UINT4:
    {
      V->datatype = "INT";
      size_of_datatype = 4;
      break;
    }
    case CDF_REAL4:
    case CDF_FLOAT:
    {
      V->datatype = "DOUBLE";
      size_of_datatype = 4;
      break;
    }
    case CDF_REAL8:
    case CDF_DOUBLE:
    {
      V->datatype = "DOUBLE";
      size_of_datatype = 8;
      break;
    }
    case CDF_EPOCH:
    {
      V->datatype = "ISO_TIME";
      size_of_datatype = 8;
      break;
    }
    case CDF_EPOCH16:
    {
      V->datatype = "ISO_TIME";
      size_of_datatype = 16;
      break;
    }
    default:
    {
      cout << " Unrecognised data type for variable '" << V->name.c_str() << "'" << endl;
      return;
    }
  }
  

  // allocate memory and dimension counts
  void *dataBlock;
  size_t length;
  if(V->recVary == true) length = (size_t) num_elems * (recEnd - recStart + 1) ;
  else  length = (size_t) num_elems;
  
  for (int n=0; n < V->objDim.size(); n++){

    dim_indices[n] = 0;

    if(dim_varys[n] == VARY){
      length = length * V->objDim[n];
      dim_counts[n] = V->objDim[n];
    }
    else{
      dim_counts[n] = 1;
    }
  }

  if (length <= 0) return;
  
  if (V->recVary == true) {
      nRecsToGet = recEnd-recStart+1;
      if( CDFlib (SELECT_, zVAR_RECNUMBER_, (long)recStart, 
                       zVAR_RECCOUNT_,  (long)nRecsToGet, 
		       zVAR_DIMCOUNTS_, dim_counts,
                       zVAR_DIMINDICES_, dim_indices,
		       NULL_) != CDF_OK) {
        cout << " Cannot set start record and number of records to get" << endl;
        return;
      }
  }
  else{
      nRecsToGet = 1;
      if( CDFlib (SELECT_, zVAR_RECNUMBER_, (long) 0, 
                       zVAR_RECCOUNT_,  (long) 1, 
		       zVAR_DIMCOUNTS_, dim_counts,
                       zVAR_DIMINDICES_, dim_indices,
		       NULL_) != CDF_OK) {
        cout << " Cannot set start record and number of records to get" << endl;
        return;
      }
   
  }
  
  void *data = (void *) calloc (length, size_of_datatype);

  // get data array in same ordering as cdf file 

  if( CDFlib (GET_, zVAR_HYPERDATA_, data, NULL_) < CDF_WARN)  {
     cout << "failed to get CDF data for variable " << V->name.c_str() << "'" << endl;
     free (data);
     return;
  }
  
  // having filled void pointer unpack into qdos object
  
  convertCDFbufferToObj(nvar, nRecsToGet, data_type, strLength, data);
     
  free (data);
  
  return;
}

  // ----------------------------------------------------------------------------
 
void QieData::convertCDFbufferToObj(int vnum, int seq_size, long data_type, int strLength, void *data){

  if(seq_size == 1) return convertCDFbufferToSingleObj(vnum, data_type, strLength, data);
  
  if(vnum >= QieVars.size()) return;
  
  QieVariable * V = &(QieVars[vnum]);
    
  // test for Data Object is TimeTags object

  if( data_type == CDF_EPOCH ) {
  
     if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) ) { 

        // Time sequence
        V->ts = new QdTimeSeq();
        double *ep = (double *)data;
  	for (int i=0; i<seq_size; i++){
           V->ts->push_back( cdfEpoch(ep[i]) );
        }
   
        // This is a time sequence object so broaden to the QdObject
        V->Vobj = V->ts;
     }
     else if(V->objDim.size() == 1 && V->objDim[0] == 2){
         // assume to be a time interval NRV object (get first interval only)
         if(seq_size > 1) cout << " QDOS cannot handle sequence of time intervals, only first interval returned" << endl;
	 double *ep = (double *)data;
	 V->ti = new QdTimeInterval(); 

	 Time *iv_start = new QDOS::Time(cdfEpoch(ep[0]));
	 Time *iv_end = new QDOS::Time(cdfEpoch(ep[1]));
	 V->ti->set(*iv_start, *iv_end );
	    
         // This is a time interval object so broaden to the QdObject
         V->Vobj = V->ti;
     }
     else{ 
        cout << " QDOS cannot handle arrays of times " << endl;
	return;
     }
  }

  else if( data_type == CDF_EPOCH16 ) { 
  
     // epoch 16 uses seconds in field 1 and picoseconds in field 2
     if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) ) { 

        // Time sequence
        V->ts = new QdTimeSeq();
        double *ep = (double *)data;
  	for (int i=0; i<seq_size; i+=2){
	   Time t = cdfEpoch16(ep[i]);
	   t += Seconds(ep[i+1] * 1.e-12);
           V->ts->push_back( t );
        }
   
        // This is a time sequence object so broaden to the QdObject
        V->Vobj = V->ts;
     }
     else if(V->objDim.size() == 1 && V->objDim[0] == 2){
         // assume to be a time interval NRV object (get first interval only)
         if(seq_size > 1) cout << " QDOS cannot handle sequence of time intervals, only first interval returned" << endl;
	 double *ep = (double *)data;
	 V->ti = new QdTimeInterval(); 

	 Time *iv_start = new QDOS::Time(cdfEpoch(ep[0]));
	 *iv_start += Seconds(ep[1] * 1.e-12);
	 Time *iv_end = new QDOS::Time(cdfEpoch(ep[2]));
	 *iv_end += Seconds(ep[3] * 1.e-12);
	 
	 V->ti->set(*iv_start, *iv_end );
	    
         // This is a time interval object so broaden to the QdObject
         V->Vobj = V->ti;
     }
     else{ 
        cout << " QDOS cannot handle arrays of times " << endl;
	return;
     }
  
  
  }

  else{ 
  
     DataBuffer_var data_buffer; 
     // data object
  
     // Sequence of objects

     if(data_type == CDF_CHAR || data_type == CDF_UCHAR) {
    
        StringVDataBuffer *sdb_p = new StringVDataBuffer();
      
        // if an array set up matrix with dimensions
   
        int n_elems=1; 
        for(int i=0; i<V->objDim.size(); i++)
        {
           n_elems *= V->objDim[i];
        }
    
        sdb_p->set_dims( V->objDim );
        sdb_p->set_majority( DBUFFER_ROW_MAJOR );
   
        const char* p = (const char*)data;
    
        for(int r=0; r <seq_size; r++ )
        {
           const char* next = p + r*strLength*n_elems;
           for(int i=0; i<n_elems; i++)
           { 
	      // count forward till we hit NULLS to get real length
              int actualLen=1;
	      for (int kk=1; kk<strLength; kk++) {
	         if( *(next+i*strLength+kk) != '\0') actualLen++;
	      }
	      // strip trailing spaces 
	      while (*(next+i*strLength+actualLen-1) == ' ')  actualLen--;
	      
              string xstr(next+i*strLength, actualLen);
              sdb_p->push_back(xstr);
           }
        }
        data_buffer=sdb_p;
        V->Vobj = Factory::clone(data_buffer);
     }
     else{
        // cast all other data types to double 

        // create new matrix sequence object 
     
        NumericVDataBuffer *ndb_p = new NumericVDataBuffer();
      
        int n_length = 1;
        for(int j=0; j<V->objDim.size(); j++){
            n_length *= V->objDim[j];
        }

        // if an array set up matrix with dimensions
    
        ndb_p->set_dims( V->objDim );
        ndb_p->set_majority( DBUFFER_ROW_MAJOR );
    
        // if not it will default to a Scalar sequence
    
        int next = 0;
    
        for(int r=0; r<seq_size; r++ )
        {

            // for numeric types use element by element static casting
            //   and let ndb_p->push_back(double) do the conversion to double
      
           switch( data_type )
           {
             case CDF_BYTE:
             case CDF_INT1:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((char *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_UINT1:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((unsigned char *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_INT2:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((short *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_UINT2:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((unsigned short *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_INT4:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((int *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_UINT4:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((unsigned int *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_REAL4:
             case CDF_FLOAT:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *( ((float *)data+next) ) );
	           next++;
	        }
                break;
             case CDF_REAL8:
             case CDF_DOUBLE:
                for(int i=0; i<n_length; i++){
                   ndb_p->push_back( *((double *)data+next)  );
	           next++;
	        }
                break;
	     default:
	        cout << "converting CDF buffer, unknown data type" << endl;
	        break;
          }
       }  // loop over records

       data_buffer=ndb_p;
       V->Vobj = Factory::clone(data_buffer);

       
    } // end else on string vs double
  
  } // end else on data or TT
       

}

// ----------------------------------------------------------------------------

void QieData::fixThemisEpoch(vector <int> getList){

  double * newData;
  QdString_var virt, comp0, comp1;
  QdTime_var baseT;
  QdRScalarSeq_var offset;

  for(int j=0; j<getList.size(); j++){
     int i = getList[j];
     if(QieVars[i].datatype != "ISO_TIME") continue;
     QieVariable *epochV = &(QieVars[i]);

     /* look for Themis virtual flag */
     for(int m=0; m< epochV->attrNameList.size(); m++){
        if( epochV->attrNameList[m] == "VIRTUAL" )   {
	   virt = QdString_var::narrow(epochV->attrList[m]);
	}
        if( epochV->attrNameList[m] == "COMPONENT_0")  {
	   comp0 = QdString_var::narrow(epochV->attrList[m]);
	    if( !comp0.is_nil() ){
	       // get variables for the two parts 
	       int comp0_num = findVarNumByName('V', comp0->c_str());
	       baseT = QdTime_var::narrow(QieVars[comp0_num].Vobj);
	   }
	}
        if( epochV->attrNameList[m] == "COMPONENT_1" ) {
	   comp1 = QdString_var::narrow(epochV->attrList[m]);  

	   if( !comp1.is_nil() ) {
	      int comp1_num = findVarNumByName('V', comp1->c_str());

	      offset = QdRScalarSeq_var::narrow(QieVars[comp1_num].Vobj);

	   }
	}
     }

     if( !virt.is_nil() &&  !baseT.is_nil() && !offset.is_nil()  ){

         // determine Themis time tags */
         if (*virt == "TRUE" ){
	    QdTimeSeq_var real_tt = new QdTimeSeq( offset->sequence_size() );
	    for (int recN=0; recN< offset->sequence_size(); recN++){
	        real_tt[recN] = *baseT + Seconds( offset[recN]); 
	    }
	 
	    epochV->Vobj = real_tt;	    
	    epochV->ts = real_tt;

	 }
      }
     
  }
  
}


// ----------------------------------------------------------------------------
 
void QieData::convertCDFbufferToSingleObj(int vnum, long data_type, int strLength, void *data){

  
  if(vnum >= QieVars.size()) return;
  
  QieVariable * V = &(QieVars[vnum]);
  
  // test for Data Object is TimeTags object

  if( data_type == CDF_EPOCH ) {
  
     if(V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] ==1) ) { 

        // Time sequence

        double *ep = (double *)data;
        QdTime_var tobj = new QdTime( cdfEpoch(ep[0]) );
   
        // This is a time sequence object so broaden to the QdObject
        V->Vobj = tobj;
     }
     else if(V->objDim.size() == 1 && V->objDim[0] == 2){
         // assume to be a time interval NRV object (get first interval only)

	 double *ep = (double *)data;
	 QdTimeInterval_var tiobj = new QdTimeInterval(); 

	 Time *iv_start = new QDOS::Time(cdfEpoch(ep[0]));
	 Time *iv_end = new QDOS::Time(cdfEpoch(ep[1]));
	 tiobj->set(*iv_start, *iv_end );
	    
         // This is a time interval object so broaden to the QdObject
         V->Vobj = tiobj;
     }
     else{ 
        cout << " QDOS cannot handle arrays of times " << endl;
	return;
     }
  }

  else if( data_type == CDF_EPOCH16 ) {
  
      // epoch 16 uses seconds in field 1 and picoseconds in field 2
      double *dp = (double *)data;
      QdTime_var tobj = new QdTime(cdfEpoch16(*( static_cast<double*>(dp)) ));
      double picosec = *( static_cast<double*>(dp) + 1 );
  
      *tobj += Seconds(picosec*1.0e-12);
     
      V->Vobj = tobj;
  
  }

  else{ 
  
     DataBuffer_var data_buffer; 
     // data object
  
     // Sequence of objects

     if(data_type == CDF_CHAR || data_type == CDF_UCHAR) {
    
        StringVDataBuffer *sdb_p = new StringVDataBuffer();
      
        // if an array set up matrix with dimensions
   
        int n_elems=1; 
        for(int i=0; i<V->objDim.size(); i++)
        {
           n_elems *= V->objDim[i];
        }
    
        sdb_p->set_dims( V->objDim );
        sdb_p->set_majority( DBUFFER_ROW_MAJOR );
   
        const char* p = (const char*)data;
    
        for(int i=0; i<n_elems; i++)
        {
           int actualLen=1;
           for (int kk=1; kk<strLength; kk++) {
	          if( *(p+i*strLength+kk) != '\0') actualLen++;
	       }   
	   
	   // strip trailing spaces 
	   while (*(p+i*strLength+actualLen-1) == ' ' && actualLen > 0)  actualLen--;
           string xstr(p+i*strLength, actualLen);
           sdb_p->push_back(xstr);
        }
        
        data_buffer=sdb_p;
        V->Vobj = Factory::clone(data_buffer);
     }
     else{
        // cast all other data types to double 

        // create new matrix sequence object 
     
        NumericVDataBuffer *ndb_p = new NumericVDataBuffer();
      
        int n_length = 1;
        for(int j=0; j<V->objDim.size(); j++){
            n_length *= V->objDim[j];
        }

        // if an array set up matrix with dimensions
    
        ndb_p->set_dims( V->objDim );
        ndb_p->set_majority( DBUFFER_ROW_MAJOR );
    
        // if not it will default to a Scalar sequence
    
        int next = 0;

            // for numeric types use element by element static casting
            //   and let ndb_p->push_back(double) do the conversion to double
      
        switch( data_type )
        {
          case CDF_BYTE:
          case CDF_INT1:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((char *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_UINT1:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((unsigned char *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_INT2:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((short *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_UINT2:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((unsigned short *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_INT4:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((int *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_UINT4:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((unsigned int *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_REAL4:
          case CDF_FLOAT:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *( ((float *)data+next) ) );
	        next++;
	     }
             break;
          case CDF_REAL8:
          case CDF_DOUBLE:
             for(int i=0; i<n_length; i++){
                ndb_p->push_back( *((double *)data+next)  );
	        next++;
	     }
             break;
	  default:
	     cout << "converting CDF buffer, unknown data type" << endl;
	     break;
       }

       data_buffer=ndb_p;
       V->Vobj = Factory::clone(data_buffer);

       
    } // end else on string vs double
  
  } // end else on data or TT
       

}




