// qpui_trace_options.cc
//
// Plot Interface Implementation - Trace Options 
//
// Janet Barnes
//
// July 2002
//
//  Tony Allen
//  Jan 2008

#include "qpui.h"
#include "TimeUtils.h"
#include "qdutil.h"

#include <qcolordialog.h>
#include <qtooltip.h>

extern QPalette QPUI_normal_palette;



// QpTraceTab Class
// ----------------

#if QT_VERSION < 0x040000
QpTraceTab::QpTraceTab(QTabWidget *tabWidget,  QString name, QpTraceItem * _trace):
    QpTab(tabWidget, name, QP_TRACE_TAB)
{
	// Constructor for the Trace Options Tab
   
	// keep track of item which owns this tab
	I_Trace = _trace;
	
	QVBoxLayout *trace_top_layout = new QVBoxLayout( this);
	trace_top_layout->setSpacing(4);

	QGroupBox *trace_line_groupbox=new QGroupBox(1, Qt::Vertical, this, "trace_line_groupbox");

	trace_line_groupbox->setTitle("Line");

	QWidget* trace_widget = new QWidget(trace_line_groupbox, "trace_widget");

	QGridLayout *trace_layout = new QGridLayout(trace_widget, 9, 5);
	trace_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	trace_layout->setMargin(QSAS_profiles->current()->GUImargins);

	QLabel* style_label = new QLabel("Line style: ",trace_widget, "style_label");
	trace_layout->addWidget(style_label,1,0,Qt::AlignRight);
	_line_style_fld=new QComboBox(false, trace_widget, "line_style");
	_line_style_fld->insertItem("None", QP_NOLINE);
	_line_style_fld->insertItem("Solid", QP_SOLID);
	_line_style_fld->insertItem("Dashed", QP_DASHED);
	_line_style_fld->insertItem("Dot-dash", QP_DOTDASH);
	_line_style_fld->insertItem("Dotted", QP_DOTTED);
	_line_style_fld->insertItem("Dash-dot-dot-dot", QP_DASHDOTDOTDOT);
	QFontMetrics line_style_fm(_line_style_fld->font());
	int line_style_w=line_style_fm.width("Dash-dot-dot-dot");
	int line_style_h=line_style_fm.height();
	_line_style_fld->setMinimumSize(line_style_w,line_style_h);
	QToolTip::add( _line_style_fld, "Select line style");
	trace_layout->addWidget(_line_style_fld,1,1,Qt::AlignLeft);
	connect( _line_style_fld, SIGNAL(activated(int)), this, SLOT(setLineStyle(int)) );
	_line_style_fld->setCurrentItem(QP_SOLID);

	QLabel* line_width_label=new QLabel("Line width: ", trace_widget,"line_width_label");
	trace_layout->addWidget(line_width_label,1,2,Qt::AlignRight);
	_line_width_fld = new FloatSpinBox(0.1,20.,1,0.1, trace_widget, "line_width");
	_line_width_fld->setValue(QSAS_profiles->current()->GUImargins);
	_line_width_fld->setFixedWidth(100);
	trace_layout->addWidget(_line_width_fld,1,3,Qt::AlignLeft);

	connect(_line_width_fld, SIGNAL(valueChanged(double)), this, SLOT(setLineWidth(double)));

	QLabel* connect_label = new QLabel("Connect style: ",trace_widget, "connect_label");
	trace_layout->addWidget(connect_label,3,0,Qt::AlignRight);
	_connect_style_fld = new QComboBox(false,trace_widget, "connect_style");
	_connect_style_fld->insertItem("Straight",QPUI_STRAIGHT);
	_connect_style_fld->insertItem("Histogram",QPUI_HIST);
	_connect_style_fld->insertItem("None",QPUI_NOCONNECT);
	QFontMetrics connect_style_fm(_connect_style_fld->font());
	int connect_style_w=connect_style_fm.width("Histogram");
	int connect_style_h=connect_style_fm.height();
	_connect_style_fld->setMinimumSize(connect_style_w,connect_style_h);
	QToolTip::add( _connect_style_fld, "Select style for connecting points");
	trace_layout->addWidget(_connect_style_fld,3,1,Qt::AlignLeft);
	connect( _connect_style_fld, SIGNAL(activated(int)), this, SLOT(setConnectStyle(int)) );

	QLabel* line_colour_label=new QLabel("Line colour: ",trace_widget, "line_colour_label");
	trace_layout->addWidget(line_colour_label,3,2,Qt::AlignRight);
	_line_colour_fld = new QPushButton("Colour...",trace_widget,  "colour");
	_line_colour_fld->setFixedWidth(100);
	_line_colour_fld->setFixedHeight(30);
	QFont f = QFont(qApp->font());
	f.setBold(TRUE);
	_line_colour_fld->setFont(f);
	trace_layout->addWidget(_line_colour_fld,3,3,Qt::AlignLeft);

	connect(_line_colour_fld, SIGNAL(clicked()), this, SLOT(setLineColour()));

	trace_layout->addRowSpacing(0,0);
	trace_layout->addRowSpacing(1,_line_style_fld->height());
	trace_layout->addRowSpacing(2,0);
	trace_layout->addRowSpacing(3,_line_colour_fld->height());
	trace_layout->addRowSpacing(4,0);

	trace_layout->setRowStretch(0,0);
	trace_layout->setRowStretch(1,1);
	trace_layout->setRowStretch(2,0);
	trace_layout->setRowStretch(3,1);
	trace_layout->setRowStretch(4,0);

	trace_layout->setColStretch(0,0);
	trace_layout->setColStretch(1,0);
	trace_layout->setColStretch(2,0);
	trace_layout->setColStretch(3,0);
	trace_layout->setColStretch(4,1);

	trace_top_layout->addWidget(trace_line_groupbox);

	QGroupBox *trace_trace_symbols_groupbox=new QGroupBox(1, Qt::Vertical, this, "trace_trace_symbols_groupbox");

	trace_trace_symbols_groupbox->setTitle("Symbols");

	QWidget* trace_symbols_widget=new QWidget(trace_trace_symbols_groupbox, "trace_symbols_widget");

	QGridLayout *trace_symbols_layout=new QGridLayout(trace_symbols_widget, 8,5);
	trace_symbols_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	trace_symbols_layout->setMargin(QSAS_profiles->current()->GUImargins);

	QLabel* shape_label=new QLabel("Shape: ",trace_symbols_widget, "shape_label");
	trace_symbols_layout->addWidget(shape_label,1,0,Qt::AlignRight);
	_symbol_shape_fld=new QComboBox(false,trace_symbols_widget, "symbol_shape");
	_symbol_shape_fld->insertItem("None",0);
	_symbol_shape_fld->insertItem("Point",1);
	_symbol_shape_fld->insertItem("Plus sign",2);
	_symbol_shape_fld->insertItem("Asterisk",3);
	_symbol_shape_fld->insertItem("Circle",4);
	_symbol_shape_fld->insertItem("Cross",5);
	_symbol_shape_fld->insertItem("Square",6);
	_symbol_shape_fld->insertItem("Triangle",7);
	_symbol_shape_fld->insertItem("Circled Cross",8);
	_symbol_shape_fld->insertItem("Circled Point",9);
	_symbol_shape_fld->insertItem("Lens Flare",10);
	_symbol_shape_fld->insertItem("Diamond",11);
	_symbol_shape_fld->insertItem("Star",12);
// 	_symbol_shape_fld->insertItem("Filled Triangle",QPUI_FILLEDTRIANGLE);
	_symbol_shape_fld->insertItem("Wide cross",13);
// 	_symbol_shape_fld->insertItem("Filled Asterisk",QPUI_FILLEDASTERISK);
	_symbol_shape_fld->insertItem("Filled Square",14);
	_symbol_shape_fld->insertItem("Filled Circle",15);
	_symbol_shape_fld->insertItem("Filled Star",16);
	QFontMetrics symbol_shape_fm(_symbol_shape_fld->font());
	int symbol_shape_w=symbol_shape_fm.width("Filled Triangle");
	int symbol_shape_h=symbol_shape_fm.height();
	_symbol_shape_fld->setMinimumSize(symbol_shape_w,symbol_shape_h);
	QToolTip::add( _symbol_shape_fld, "Select symbol shape");
	trace_symbols_layout->addWidget(_symbol_shape_fld,1,1,Qt::AlignLeft);
	connect( _symbol_shape_fld, SIGNAL(activated(int)), this, SLOT(setSymbolShape(int)) );

	QLabel* size_label = new QLabel("Size: ", trace_symbols_widget,"size_label");
	trace_symbols_layout->addWidget(size_label,1,2,Qt::AlignRight);
	_symbol_size_fld = new FloatSpinBox(0.1,20.,1,0.5,trace_symbols_widget, "symbol_size");
	_symbol_size_fld->setFixedWidth(100);
	trace_symbols_layout->addWidget(_symbol_size_fld,1,3,Qt::AlignLeft);
	_symbol_size_fld->setValue(1.0);
	
	connect(_symbol_size_fld,SIGNAL(valueChanged(double)), this, SLOT(setSymbolSize(double)));

	QLabel* symbols_colour_label=new QLabel("Colour: ",trace_symbols_widget, "symbols_colour_label");
	trace_symbols_layout->addWidget(symbols_colour_label,3,2,Qt::AlignRight);
	_symbol_colour_fld=new QPushButton("Colour...",trace_symbols_widget, "colour");
	_symbol_colour_fld->setFixedWidth(100);
	_symbol_colour_fld->setFixedHeight(30);
	_symbol_colour_fld->setFont(f); // declared bold above
	trace_symbols_layout->addWidget(_symbol_colour_fld,3,3,Qt::AlignLeft);

	connect(_symbol_colour_fld, SIGNAL(clicked()), this, SLOT(setSymbolColour()));

	trace_symbols_layout->addRowSpacing(0,0);
	trace_symbols_layout->addRowSpacing(1,_symbol_shape_fld->height());
	trace_symbols_layout->addRowSpacing(2,0);
	trace_symbols_layout->addRowSpacing(3,_symbol_colour_fld->height());
	trace_symbols_layout->addRowSpacing(4,0);

	trace_symbols_layout->setRowStretch(0,0);
	trace_symbols_layout->setRowStretch(1,1);
	trace_symbols_layout->setRowStretch(2,0);
	trace_symbols_layout->setRowStretch(3,1);
	trace_symbols_layout->setRowStretch(4,0);

	trace_symbols_layout->setColStretch(0,0);
	trace_symbols_layout->setColStretch(1,0);
	trace_symbols_layout->setColStretch(2,0);
	trace_symbols_layout->setColStretch(3,0);
	trace_symbols_layout->setColStretch(4,1);

	trace_top_layout->addWidget(trace_trace_symbols_groupbox);

	trace_top_layout->addSpacing(4);
	trace_top_layout->addStretch(10);

	trace_widget->adjustSize();
	trace_symbols_widget->adjustSize();

	return;
}
#else
QpTraceTab::QpTraceTab(QTabWidget *tabWidget,  QString name, QpTraceItem * _trace):
		QpTab(tabWidget, name, QP_TRACE_TAB)
{
	// Constructor for the Trace Options Tab
   
	// keep track of item which owns this tab
	I_Trace = _trace;
	
	QVBoxLayout *trace_top_layout = new QVBoxLayout( this);
	trace_top_layout->setSpacing(4);

	QGroupBox *trace_line_groupbox=new QGroupBox(this);

	trace_line_groupbox->setTitle("Line");

// 	QWidget* trace_widget = new QWidget(trace_line_groupbox, "trace_widget");

	QGridLayout *trace_layout = new QGridLayout;
	trace_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	trace_layout->setMargin(QSAS_profiles->current()->GUImargins);

	QLabel* style_label = new QLabel("Line style: ");
	trace_layout->addWidget(style_label,1,0,Qt::AlignRight);
	_line_style_fld=new QComboBox(false);
	_line_style_fld->insertItem("None", QP_NOLINE);
	_line_style_fld->insertItem("Solid", QP_SOLID);
	_line_style_fld->insertItem("Dashed", QP_DASHED);
	_line_style_fld->insertItem("Dot-dash", QP_DOTDASH);
	_line_style_fld->insertItem("Dotted", QP_DOTTED);
	_line_style_fld->insertItem("Dash-dot-dot-dot", QP_DASHDOTDOTDOT);
	QFontMetrics line_style_fm(_line_style_fld->font());
	int line_style_w=line_style_fm.width("Dash-dot-dot-dot");
	int line_style_h=line_style_fm.height();
	_line_style_fld->setMinimumSize(line_style_w,line_style_h);
	QToolTip::add( _line_style_fld, "Select line style");
	trace_layout->addWidget(_line_style_fld,1,1,Qt::AlignLeft);
	connect( _line_style_fld, SIGNAL(activated(int)), this, SLOT(setLineStyle(int)) );
	_line_style_fld->setCurrentItem(QP_SOLID);

	QLabel* line_width_label=new QLabel("Line width: ");
	trace_layout->addWidget(line_width_label,1,2,Qt::AlignRight);
	_line_width_fld = new FloatSpinBox(0.1,20.,1,0.1, trace_line_groupbox, "line_width");
	_line_width_fld->setValue(QSAS_profiles->current()->LineWidth);
	_line_width_fld->setFixedWidth(100);
	trace_layout->addWidget(_line_width_fld,1,3,Qt::AlignLeft);

	connect(_line_width_fld, SIGNAL(valueChanged(double)), this, SLOT(setLineWidth(double)));

	QLabel* connect_label = new QLabel("Connect style: ");
	trace_layout->addWidget(connect_label,3,0,Qt::AlignRight);
	_connect_style_fld = new QComboBox(false);
	_connect_style_fld->insertItem("Straight",QPUI_STRAIGHT);
	_connect_style_fld->insertItem("Histogram",QPUI_HIST);
	_connect_style_fld->insertItem("None",QPUI_NOCONNECT);
	QFontMetrics connect_style_fm(_connect_style_fld->font());
	int connect_style_w=connect_style_fm.width("Histogram");
	int connect_style_h=connect_style_fm.height();
	_connect_style_fld->setMinimumSize(connect_style_w,connect_style_h);
	QToolTip::add( _connect_style_fld, "Select style for connecting points");
	trace_layout->addWidget(_connect_style_fld,3,1,Qt::AlignLeft);
	connect( _connect_style_fld, SIGNAL(activated(int)), this, SLOT(setConnectStyle(int)) );

	QLabel* line_colour_label=new QLabel("Line colour: ");
	trace_layout->addWidget(line_colour_label,3,2,Qt::AlignRight);
	_line_colour_fld = new QPushButton("Colour...");
	_line_colour_fld->setFixedWidth(100);
	_line_colour_fld->setFixedHeight(30);
	QFont f = QFont(qApp->font());
	f.setBold(TRUE);
	_line_colour_fld->setFont(f);
	trace_layout->addWidget(_line_colour_fld,3,3,Qt::AlignLeft);

	connect(_line_colour_fld, SIGNAL(clicked()), this, SLOT(setLineColour()));

// 	trace_layout->addRowSpacing(0,0);
// 	trace_layout->addRowSpacing(1,_line_style_fld->height());
// 	trace_layout->addRowSpacing(2,0);
// 	trace_layout->addRowSpacing(3,_line_colour_fld->height());
// 	trace_layout->addRowSpacing(4,0);

	trace_layout->setRowStretch(0,0);
	trace_layout->setRowStretch(1,1);
	trace_layout->setRowStretch(2,0);
	trace_layout->setRowStretch(3,1);
	trace_layout->setRowStretch(4,0);

	trace_layout->setColStretch(0,0);
	trace_layout->setColStretch(1,0);
	trace_layout->setColStretch(2,0);
	trace_layout->setColStretch(3,0);
	trace_layout->setColStretch(4,1);

	trace_line_groupbox->setLayout(trace_layout);
	
	trace_top_layout->addWidget(trace_line_groupbox);

	QGroupBox *trace_trace_symbols_groupbox=new QGroupBox(this);

	trace_trace_symbols_groupbox->setTitle("Symbols");

// 	QWidget* trace_symbols_widget=new QWidget(trace_trace_symbols_groupbox, "trace_symbols_widget");

	QGridLayout *trace_symbols_layout=new QGridLayout;
	trace_symbols_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	trace_symbols_layout->setMargin(QSAS_profiles->current()->GUImargins);

	QLabel* shape_label=new QLabel("Shape: ");
	trace_symbols_layout->addWidget(shape_label,1,0,Qt::AlignRight);
	_symbol_shape_fld=new QComboBox(false);
	_symbol_shape_fld->insertItem("None",0);
	_symbol_shape_fld->insertItem("Point",1);
	_symbol_shape_fld->insertItem("Plus sign",2);
	_symbol_shape_fld->insertItem("Asterisk",3);
	_symbol_shape_fld->insertItem("Circle",4);
	_symbol_shape_fld->insertItem("Cross",5);
	_symbol_shape_fld->insertItem("Square",6);
	_symbol_shape_fld->insertItem("Triangle",7);
	_symbol_shape_fld->insertItem("Circled Cross",8);
	_symbol_shape_fld->insertItem("Circled Point",9);
	_symbol_shape_fld->insertItem("Lens Flare",10);
	_symbol_shape_fld->insertItem("Diamond",11);
	_symbol_shape_fld->insertItem("Star",12);
// 	_symbol_shape_fld->insertItem("Filled Triangle",QPUI_FILLEDTRIANGLE);
	_symbol_shape_fld->insertItem("Wide cross",13);
// 	_symbol_shape_fld->insertItem("Filled Asterisk",QPUI_FILLEDASTERISK);
	_symbol_shape_fld->insertItem("Filled Square",14);
	_symbol_shape_fld->insertItem("Filled Circle",15);
	_symbol_shape_fld->insertItem("Filled Star",16);
	QFontMetrics symbol_shape_fm(_symbol_shape_fld->font());
	int symbol_shape_w=symbol_shape_fm.width("Filled Triangle");
	int symbol_shape_h=symbol_shape_fm.height();
	_symbol_shape_fld->setMinimumSize(symbol_shape_w,symbol_shape_h);
	QToolTip::add( _symbol_shape_fld, "Select symbol shape");
	trace_symbols_layout->addWidget(_symbol_shape_fld,1,1,Qt::AlignLeft);
	connect( _symbol_shape_fld, SIGNAL(activated(int)), this, SLOT(setSymbolShape(int)) );

	QLabel* size_label = new QLabel("Size: ");
	trace_symbols_layout->addWidget(size_label,1,2,Qt::AlignRight);
	_symbol_size_fld = new FloatSpinBox(0.1,20.,1,0.5,trace_trace_symbols_groupbox, "symbol_size");
	_symbol_size_fld->setFixedWidth(100);
	trace_symbols_layout->addWidget(_symbol_size_fld,1,3,Qt::AlignLeft);
	_symbol_size_fld->setValue(1.0);
	
	connect(_symbol_size_fld,SIGNAL(valueChanged(double)), this, SLOT(setSymbolSize(double)));

	QLabel* symbols_colour_label=new QLabel("Colour: ");
	trace_symbols_layout->addWidget(symbols_colour_label,3,2,Qt::AlignRight);
	_symbol_colour_fld=new QPushButton("Colour...");
	_symbol_colour_fld->setFixedWidth(100);
	_symbol_colour_fld->setFixedHeight(30);
	_symbol_colour_fld->setFont(f); // declared bold above
	trace_symbols_layout->addWidget(_symbol_colour_fld,3,3,Qt::AlignLeft);

	connect(_symbol_colour_fld, SIGNAL(clicked()), this, SLOT(setSymbolColour()));

// 	trace_symbols_layout->addRowSpacing(0,0);
// 	trace_symbols_layout->addRowSpacing(1,_symbol_shape_fld->height());
// 	trace_symbols_layout->addRowSpacing(2,0);
// 	trace_symbols_layout->addRowSpacing(3,_symbol_colour_fld->height());
// 	trace_symbols_layout->addRowSpacing(4,0);

	trace_symbols_layout->setRowStretch(0,0);
	trace_symbols_layout->setRowStretch(1,1);
	trace_symbols_layout->setRowStretch(2,0);
	trace_symbols_layout->setRowStretch(3,1);
	trace_symbols_layout->setRowStretch(4,0);

	trace_symbols_layout->setColStretch(0,0);
	trace_symbols_layout->setColStretch(1,0);
	trace_symbols_layout->setColStretch(2,0);
	trace_symbols_layout->setColStretch(3,0);
	trace_symbols_layout->setColStretch(4,1);

	trace_trace_symbols_groupbox->setLayout(trace_symbols_layout);
	
	trace_top_layout->addWidget(trace_trace_symbols_groupbox);

	trace_top_layout->addSpacing(4);
	trace_top_layout->addStretch(10);

// 	trace_widget->adjustSize();
// 	trace_symbols_widget->adjustSize();

	return;
}
#endif

// Trace Options
// ------------

void QpTraceTab::setLineStyle(int choice)
{
   // only tell list view to make changes to selected trace items if this
   // is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
   if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_LINE_STYLE, choice);
}

void QpTraceTab::setLineWidth(double choice)
{
   // only tell list view to make changes to selected trace items if this
   // is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
   if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_LINE_WIDTH, 0, (float)choice);
}

void QpTraceTab::setConnectStyle(int choice)
{
   // only tell list view to make changes to selected trace items if this
   // is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
   if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_CONNECT_STYLE, choice);
}

void QpTraceTab::setLineColour()
{
	bool changed = false;
	// ask the user for colour choice
	QColor c = QuGetColour(I_Trace->_line_colour, &changed, this);
	if(changed)
	{
		I_Trace->userLineColour = true;
		setLineColour(c);
	}

	// only tell list view to make changes to selected trace items if this
	// is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
	if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_LINE_COLOUR, 0, 0., "", false, c);
}

void QpTraceTab::setLineColour(QColor c)
{
	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QColorGroup::ButtonText, c);
	line_palette.setColor(QPalette::Inactive, QColorGroup::ButtonText, c);
	this->_line_colour_fld->setPalette(line_palette);
	I_Trace->_line_colour = c;
}

void QpTraceTab::setSymbolShape(int choice)
{
   // only tell list view to make changes to selected trace items if this
   // is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
   if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_SYMBOL_SHAPE, choice);
}

void QpTraceTab::setSymbolSize(double choice)
{
   // only tell list view to make changes to selected trace items if this
   // is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
   if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_SYMBOL_SIZE, 0, choice);
}

void QpTraceTab::setSymbolColour()
{
   bool changed = false;
   // ask the user for colour choice
   QColor c = QuGetColour(I_Trace->_symbol_colour, &changed, this);
   if(changed)
   {
		I_Trace->userSymbolColour = true;
        setSymbolColour(c);
   }

   // only tell list view to make changes to selected trace items if this
   // is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
   
   if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_SYMBOL_COLOUR, 0, 0., "", false, c);
}

void QpTraceTab::setSymbolColour(QColor c)
{
  
	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QColorGroup::ButtonText, c);
	line_palette.setColor(QPalette::Inactive, QColorGroup::ButtonText, c);
	this->_symbol_colour_fld->setPalette(line_palette);
	I_Trace->_symbol_colour=c;
 
}

// QpDataTab class
//------------------

#if QT_VERSION < 0x040000
QpDataTab::QpDataTab(QTabWidget *tabWidget,  QString name, QpPlotItem * _plot, QpItemType plotType):
    QpTab(tabWidget, name, QP_DATA_TAB)
{
	// Constructor for the Data Tab for the Trace Options
  
	// keep track of item which owns this tab
	I_Plot = _plot;
	I_PlotType = plotType;
	
	old_BaseName = QString("");
	old_Comp = QString("");
	old_BaseXName = QString("");
	old_XComp = QString("");
	
	QString PlotTitle = "Plot Object";
	if(plotType == QP_SPECTROGRAM) PlotTitle = "Spectrogram Data";
	else if(plotType == QP_TRACE) PlotTitle = "Trace Data";
	else if(plotType == QP_CONST_XT && I_Plot->frameType() == QP_XY) PlotTitle = "Const  X  Data";
	else if(plotType == QP_CONST_XT && I_Plot->frameType() == QP_TS) PlotTitle = "Const  T  Data";
	else if(plotType == QP_CONST_Y) PlotTitle = "Const  Y  Data";
	
	QVBoxLayout *data_top_layout = new QVBoxLayout(this);
	data_top_layout->setSpacing(4);

	QGroupBox *data_plot_groupbox = new QGroupBox(1, Qt::Vertical, this, "data_plot_groupbox");
	data_plot_groupbox->setTitle(PlotTitle);
	

	QWidget* data_plot_widget=new QWidget(data_plot_groupbox, "data_plot_widget");

	QGridLayout *data_plot_layout=new QGridLayout( data_plot_widget, 9, 4);
	data_plot_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	data_plot_layout->setMargin(QSAS_profiles->current()->GUImargins);

	if(plotType == QP_TRACE || plotType == QP_CONST_XT || plotType == QP_CONST_Y)
	{
		_y_label_fld=new QLabel("Y: ",data_plot_widget,"y_label");
		data_plot_layout->addWidget(_y_label_fld,1,0,Qt::AlignRight);
		_y_data_fld=new QguWLinput(data_plot_widget,"y_data");
		_y_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget(_y_data_fld,1,1,1,2,Qt::AlignLeft);
		_y_data_fld->acceptScalarObject=true;
		_y_data_fld->acceptTimeSubset=true;
		_y_data_fld->acceptTimeSlice=true;
		_y_data_fld->reduce3DtoScalar = true;

		connect(_y_data_fld, SIGNAL(validated_OK()), this, SLOT(setYData()));
		connect(_y_data_fld, SIGNAL(validated_WARN()), this, SLOT(setYTextOnly()));

		_x_label_fld = new QLabel("X: ",data_plot_widget, "x_label");
		data_plot_layout->addWidget(_x_label_fld,3,0,Qt::AlignRight);
		_x_data_fld = new QguWLinput(data_plot_widget,"x_data");
		_x_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget(_x_data_fld,3,3,1,2,Qt::AlignLeft);
		_x_data_fld->acceptScalarObject=true;
		_x_data_fld->acceptTimeSubset=true;
		_x_data_fld->acceptTimeSlice=true;
		connect(_x_data_fld, SIGNAL(validated_OK()), this, SLOT(setXData()));
		
		timeAsSec_toggle = new QCheckBox("Use time as sec from T1", data_plot_widget, "timeAsSec_toggle");
		data_plot_layout->addMultiCellWidget(timeAsSec_toggle,5,5,1,2,Qt::AlignLeft);
		timeAsSec_toggle->setChecked(false);
		connect(timeAsSec_toggle, SIGNAL(toggled(bool)), this, SLOT(useTimeAsSec(bool)) );
		if(I_Plot->frameType() == QP_TS) timeAsSec_toggle->hide();


		data_plot_layout->addRowSpacing(0,0);
		data_plot_layout->addRowSpacing(1,_y_data_fld->height());
		data_plot_layout->addRowSpacing(2,0);
		data_plot_layout->addRowSpacing(3,_x_data_fld->height());
		data_plot_layout->addRowSpacing(4,0);

		data_plot_layout->setRowStretch(0,0);
		data_plot_layout->setRowStretch(1,1);
		data_plot_layout->setRowStretch(2,0);
		data_plot_layout->setRowStretch(3,1);
		data_plot_layout->setRowStretch(4,0);

		data_plot_layout->setColStretch(0,0);
		data_plot_layout->setColStretch(1,0);
		data_plot_layout->setColStretch(2,0);
		data_plot_layout->setColStretch(3,0);
		data_plot_layout->setColStretch(4,10);
		
		data_top_layout->addWidget(data_plot_groupbox);
		
		QGroupBox *data_legend_groupbox=new QGroupBox(1, Qt::Vertical, this, "data_legend_groupbox");

		data_legend_groupbox->setTitle("Legend");

		QWidget* data_legend_widget=new QWidget(data_legend_groupbox, "data_legend_widget");

		QGridLayout *data_legend_layout=new QGridLayout(data_legend_widget,5,6);
		data_legend_layout->setSpacing(QSAS_profiles->current()->GUImargins);
		data_legend_layout->setMargin(QSAS_profiles->current()->GUImargins);

		_show_legend_fld = new QCheckBox("Include in legend", data_legend_widget, "show_legend");
		_show_legend_fld->setChecked(true);
		data_legend_layout->addWidget(_show_legend_fld,1,0,Qt::AlignCenter);
		connect(_show_legend_fld, SIGNAL(toggled(bool)), this, SLOT(setShowLegend()) );

		QLabel *legend_label_fld=new QLabel("Text: ",data_legend_widget,"legend_label");
		data_legend_layout->addWidget(legend_label_fld,1,1,Qt::AlignRight);
		_legend_text_fld=new QLineEdit(data_legend_widget,"legend_text");
		_legend_text_fld->setMinimumWidth(300);
		data_legend_layout->addMultiCellWidget(_legend_text_fld,1,1,2,4,Qt::AlignLeft);
		connect( _legend_text_fld, SIGNAL(textChanged(const QString &)) ,this, SLOT(setLegendText(const QString &)) );
	
		setSlotColour(false);
		_legend_text_fld->setPalette(GUIpalette);
		_auto_legend_fld = new QCheckBox("Auto-legend", data_legend_widget, "show_legend");
		_auto_legend_fld->setChecked(true);
		data_legend_layout->addWidget(_auto_legend_fld,1,5,Qt::AlignCenter);
		connect(_auto_legend_fld, SIGNAL(toggled(bool)), this, SLOT(setAutoLegend(bool)) );

		data_legend_layout->addRowSpacing(0,0);
		data_legend_layout->addRowSpacing(1,_legend_text_fld->height());
		data_legend_layout->addRowSpacing(2,0);

		data_legend_layout->setRowStretch(0,0);
		data_legend_layout->setRowStretch(1,1);
		data_legend_layout->setRowStretch(2,0);


		data_top_layout->addWidget(data_legend_groupbox);

		data_type_groupbox = new QGroupBox(1, Qt::Vertical, this, "data_type_groupbox");
		setDataTypeTitle();
		

		QWidget* data_type_widget=new QWidget(data_type_groupbox, "data_type_widget");

		QGridLayout *data_type_layout=new QGridLayout(data_type_widget,1,4);
		data_type_layout->setSpacing(QSAS_profiles->current()->GUImargins*2.5);
		data_type_layout->setMargin(QSAS_profiles->current()->GUImargins*2.5);

		QPushButton *_type_btn = new QPushButton("Change Plot Type...", data_type_widget,"type_btn");
		_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_type_btn,0,0,Qt::AlignLeft);
		connect(_type_btn, SIGNAL(pressed()), this, SLOT(launchTypeSelector()) );

		QPushButton *_frame_type_btn = new QPushButton("Change Frame Type...", data_type_widget, "frame_type_btn");
		_frame_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_frame_type_btn,0,2,Qt::AlignLeft);
		connect(_frame_type_btn, SIGNAL(pressed()), this->I_Plot->I_Panel->I_Frame->xtAxisTab, SLOT(launchFrameTypeSelector()) );
		QToolTip::add( _frame_type_btn, "Sets whether ALL plot items in frame will be TS or XY");
		
		
		data_type_layout->addRowSpacing(0,_type_btn->height());		

		data_type_layout->setColStretch(0,0);
		data_type_layout->setColStretch(1,1);
		data_type_layout->setColStretch(2,0);
		data_type_layout->setColStretch(3,5);

		data_top_layout->addWidget(data_type_groupbox);
	

	}
	else if(plotType == QP_SPECTROGRAM)
	{
	
		QLabel* z_label_fld  =new QLabel("Z: ", data_plot_widget, "z_label");
		data_plot_layout->addWidget(z_label_fld, 1, 0, Qt::AlignRight);
		_z_data_fld = new QguWLinput(data_plot_widget, "z_data");
		_z_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget( _z_data_fld, 1, 1, 1, 2, Qt::AlignLeft);
		_z_data_fld->acceptObjectPlus=true;
		_z_data_fld->acceptTimeSubset=true;
		_z_data_fld->acceptTimeSlice=true;
		_z_data_fld->reduce3Dto1 = true;
		
		connect(_z_data_fld, SIGNAL(validated_OK()), this, SLOT(setZData())); 
		connect(_z_data_fld, SIGNAL(validated_WARN()), this, SLOT(setZTextOnly()));

		_XYtoggle = new QCheckBox("Swap default X and Y axes", data_plot_widget, "XYtoggle");
		data_plot_layout->addMultiCellWidget(_XYtoggle, 7, 7, 2, 2, Qt::AlignLeft);
		_XYtoggle->setChecked(false);
		if(I_Plot->frameType() == QP_TS) _XYtoggle->setEnabled(false);
		connect(_XYtoggle, SIGNAL(clicked()), this, SLOT(setXYswap()));
		
		QLabel* y_start_label_fld=new QLabel("Y: ", data_plot_widget, "y_start_label");
		QFont it_font = QFont(qApp->font());
		it_font.setItalic(true);
		y_start_label_fld->setFont(it_font);
		data_plot_layout->addWidget(y_start_label_fld, 3, 0, Qt::AlignRight);
		_y_start_data_fld = new QguWLinput(data_plot_widget, "y_start_data");
		_y_start_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget( _y_start_data_fld, 3, 3, 1, 2, Qt::AlignLeft);
		_y_start_data_fld->acceptObjectPlus=true;
		_y_start_data_fld->acceptTimeSubset=true;
		_y_start_data_fld->acceptTimeSlice=true;

		connect(_y_start_data_fld, SIGNAL(validated_OK()), this, SLOT(setYStartData()));

		QLabel* y_end_label_fld = new QLabel("Y End: ", data_plot_widget, "y_end_label");
		y_end_label_fld->setFont(it_font);
		data_plot_layout->addWidget(y_end_label_fld, 4, 1, Qt::AlignRight);
		_y_end_data_fld = new QguWLinput(data_plot_widget, "y_end_data");
		_y_end_data_fld->setMinimumWidth(300);
		data_plot_layout->addMultiCellWidget(_y_end_data_fld, 4, 4, 2, 2, Qt::AlignRight);
		_y_end_data_fld->acceptObjectPlus=true;
		_y_end_data_fld->acceptTimeSubset=true;
		_y_end_data_fld->acceptTimeSlice=true;

		connect(_y_end_data_fld, SIGNAL(validated_OK()), this, SLOT(setYEndData()));

		_x_start_label_fld = new QLabel("X: ", data_plot_widget, "x_start_label");
		_x_start_label_fld->setFont(it_font);
		data_plot_layout->addWidget(_x_start_label_fld, 5, 0, Qt::AlignRight);
		_x_start_data_fld  =new QguWLinput(data_plot_widget, "x_start_data");
		_x_start_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget( _x_start_data_fld, 5, 5, 1, 2, Qt::AlignLeft);
		_x_start_data_fld->acceptObjectPlus=true;
		_x_start_data_fld->acceptTimeSubset=true;
		_x_start_data_fld->acceptTimeSlice=true;
		if(I_Plot->frameType() == QP_TS) 
		{
			_x_start_label_fld->hide();
			_x_start_data_fld->hide();
		}
	
		connect(_x_start_data_fld, SIGNAL(validated_OK()), this, SLOT(setXStartData()));

		_x_end_label_fld = new QLabel("X End: ", data_plot_widget, "x_end_label");
		_x_end_label_fld->setFont(it_font);
		data_plot_layout->addWidget(_x_end_label_fld, 6, 1, Qt::AlignRight);
		_x_end_data_fld = new QguWLinput(data_plot_widget, "x_end_data");
		_x_end_data_fld->setMinimumWidth(300);
		data_plot_layout->addMultiCellWidget( _x_end_data_fld, 6, 6, 2, 2, Qt::AlignRight);
		_x_end_data_fld->acceptObjectPlus=true;
		_x_end_data_fld->acceptTimeSubset=true;
		_x_end_data_fld->acceptTimeSlice=true;
		if(I_Plot->frameType() == QP_TS) {
			_x_end_label_fld->hide();
			_x_end_data_fld->hide();
		}

		connect(_x_end_data_fld, SIGNAL(validated_OK()), this, SLOT(setXEndData()));

		timeAsSec_toggle = new QCheckBox("Use time as sec from T1", data_plot_widget, "timeAsSec_toggle");
		data_plot_layout->addMultiCellWidget(timeAsSec_toggle, 7, 7, 1, 2, Qt::AlignLeft);
		timeAsSec_toggle->setChecked(false);
		connect(timeAsSec_toggle, SIGNAL(toggled(bool)), this, SLOT(useTimeAsSec(bool)) );
		if(I_Plot->frameType() == QP_TS) timeAsSec_toggle->hide();

		gap_toggle = new QCheckBox("Fill t gaps", data_plot_widget, "gap_toggle");
		data_plot_layout->addMultiCellWidget(gap_toggle, 7, 7, 1, 1, Qt::AlignLeft);
		gap_toggle->setChecked(false);
		if(I_Plot->frameType() == QP_XY) gap_toggle->hide();

		connect(gap_toggle, SIGNAL(clicked()), this, SLOT(gap_toggle_changed()));
		
		data_plot_layout->addRowSpacing(0,0);
		data_plot_layout->addRowSpacing(1,_z_data_fld->height());
		data_plot_layout->addRowSpacing(2,0);
		data_plot_layout->addRowSpacing(3,_y_start_data_fld->height());
		data_plot_layout->addRowSpacing(4,0);
		data_plot_layout->addRowSpacing(5,_x_start_data_fld->height());
		data_plot_layout->addRowSpacing(6,0);

		data_plot_layout->setRowStretch(0,0);
		data_plot_layout->setRowStretch(1,1);
		data_plot_layout->setRowStretch(2,0);
		data_plot_layout->setRowStretch(3,1);
		data_plot_layout->setRowStretch(4,0);
		data_plot_layout->setRowStretch(5,1);


		data_top_layout->addWidget(data_plot_groupbox);

		data_type_groupbox=new QGroupBox(1, Qt::Vertical, this, "data_type_groupbox");
		setDataTypeTitle();

		QWidget* data_type_widget=new QWidget(data_type_groupbox, "data_type_widget");

		QGridLayout *data_type_layout=new QGridLayout(data_type_widget,1,4);
		data_type_layout->setSpacing(QSAS_profiles->current()->GUImargins*2.5);
		data_type_layout->setMargin(QSAS_profiles->current()->GUImargins*2.5);

		QPushButton *_type_btn = new QPushButton("Change Plot Type...", data_type_widget, "type_btn");
		_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_type_btn,0,0,Qt::AlignLeft);
		connect(_type_btn, SIGNAL(pressed()), this, SLOT(launchTypeSelector()) );
	
		QPushButton *_frame_type_btn = new QPushButton("Change Frame Type...", data_type_widget, "frame_type_btn");
		_frame_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_frame_type_btn,0,2,Qt::AlignLeft);
		connect(_frame_type_btn, SIGNAL(pressed()), this->I_Plot->I_Panel->I_Frame->xtAxisTab, SLOT(launchFrameTypeSelector()) );
		QToolTip::add( _frame_type_btn, "Sets whether ALL plot items in frame will be TS or XY");

		data_type_layout->addRowSpacing(0,_type_btn->height());

		data_type_layout->setColStretch(0,0);
		data_type_layout->setColStretch(1,1);
		data_type_layout->setColStretch(2,0);
		data_type_layout->setColStretch(3,5);
		data_plot_layout->setColStretch(4,10);

		data_top_layout->addWidget(data_type_groupbox);

	}
	

	data_top_layout->addSpacing(4);
	data_top_layout->addStretch(10);

	return;
}
#else
QpDataTab::QpDataTab(QTabWidget *tabWidget,  QString name, QpPlotItem * _plot, QpItemType plotType):
		QpTab(tabWidget, name, QP_DATA_TAB)
{
	// Constructor for the Data Tab for the Trace Options
  
	// keep track of item which owns this tab
	I_Plot = _plot;
	I_PlotType = plotType;
	
	old_BaseName = QString("");
	old_Comp = QString("");
	old_BaseXName = QString("");
	old_XComp = QString("");
	
	QString PlotTitle = "Plot Object";
	if(plotType == QP_SPECTROGRAM) PlotTitle = "Spectrogram Data";
	else if(plotType == QP_TRACE) PlotTitle = "Trace Data";
	else if(plotType == QP_CONST_XT && I_Plot->frameType() == QP_XY) PlotTitle = "Const  X  Data";
	else if(plotType == QP_CONST_XT && I_Plot->frameType() == QP_TS) PlotTitle = "Const  T  Data";
	else if(plotType == QP_CONST_Y) PlotTitle = "Const  Y  Data";
	
	QVBoxLayout *data_top_layout = new QVBoxLayout(this);
	data_top_layout->setSpacing(4);

	QGroupBox *data_plot_groupbox = new QGroupBox(this);
	data_plot_groupbox->setTitle(PlotTitle);
	

// 	QWidget* data_plot_widget=new QWidget(data_plot_groupbox, "data_plot_widget");

	QGridLayout *data_plot_layout=new QGridLayout;
	data_plot_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	data_plot_layout->setMargin(QSAS_profiles->current()->GUImargins);

	if(plotType == QP_TRACE || plotType == QP_CONST_XT || plotType == QP_CONST_Y)
	{
		_y_label_fld=new QLabel("Y: ");
		data_plot_layout->addWidget(_y_label_fld,1,0,Qt::AlignRight);
		_y_data_fld=new QguWLinput(data_plot_groupbox,"y_data");
		_y_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget(_y_data_fld,1,1,1,2,Qt::AlignLeft);
		_y_data_fld->acceptScalarObject=true;
		_y_data_fld->acceptTimeSubset=true;
		_y_data_fld->acceptTimeSlice=true;
		_y_data_fld->reduce3DtoScalar = true;

		connect(_y_data_fld, SIGNAL(validated_OK()), this, SLOT(setYData()));
		connect(_y_data_fld, SIGNAL(validated_WARN()), this, SLOT(setYTextOnly()));

		_x_label_fld = new QLabel("X: ");
		data_plot_layout->addWidget(_x_label_fld,3,0,Qt::AlignRight);
		_x_data_fld = new QguWLinput(data_plot_groupbox,"x_data");
		_x_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget(_x_data_fld,3,3,1,2,Qt::AlignLeft);
		_x_data_fld->acceptScalarObject=true;
		_x_data_fld->acceptTimeSubset=true;
		_x_data_fld->acceptTimeSlice=true;
		connect(_x_data_fld, SIGNAL(validated_OK()), this, SLOT(setXData()));
		
		timeAsSec_toggle = new QCheckBox("Use time as sec from T1");
		data_plot_layout->addMultiCellWidget(timeAsSec_toggle,5,5,1,2,Qt::AlignLeft);
		timeAsSec_toggle->setChecked(false);
		connect(timeAsSec_toggle, SIGNAL(toggled(bool)), this, SLOT(useTimeAsSec(bool)) );
		if(I_Plot->frameType() == QP_TS) timeAsSec_toggle->hide();


// 		data_plot_layout->addRowSpacing(0,0);
// 		data_plot_layout->addRowSpacing(1,_y_data_fld->height());
// 		data_plot_layout->addRowSpacing(2,0);
// 		data_plot_layout->addRowSpacing(3,_x_data_fld->height());
// 		data_plot_layout->addRowSpacing(4,0);

		data_plot_layout->setRowStretch(0,0);
		data_plot_layout->setRowStretch(1,1);
		data_plot_layout->setRowStretch(2,0);
		data_plot_layout->setRowStretch(3,1);
		data_plot_layout->setRowStretch(4,0);

		data_plot_layout->setColStretch(0,0);
		data_plot_layout->setColStretch(1,0);
		data_plot_layout->setColStretch(2,0);
		data_plot_layout->setColStretch(3,0);
		data_plot_layout->setColStretch(4,10);
		
		data_plot_groupbox->setLayout(data_plot_layout);
		
		data_top_layout->addWidget(data_plot_groupbox);
		
		QGroupBox *data_legend_groupbox=new QGroupBox(this);

		data_legend_groupbox->setTitle("Legend");

// 		QWidget* data_legend_widget=new QWidget(data_legend_groupbox, "data_legend_widget");

		QGridLayout *data_legend_layout=new QGridLayout;
		data_legend_layout->setSpacing(QSAS_profiles->current()->GUImargins);
		data_legend_layout->setMargin(QSAS_profiles->current()->GUImargins);

		_show_legend_fld = new QCheckBox("Include in legend");
		_show_legend_fld->setChecked(true);
		data_legend_layout->addWidget(_show_legend_fld,1,0,Qt::AlignCenter);
		connect(_show_legend_fld, SIGNAL(toggled(bool)), this, SLOT(setShowLegend()) );

		QLabel *legend_label_fld=new QLabel("Text: ");
		data_legend_layout->addWidget(legend_label_fld,1,1,Qt::AlignRight);
		_legend_text_fld=new QLineEdit;
		_legend_text_fld->setMinimumWidth(300);
		data_legend_layout->addMultiCellWidget(_legend_text_fld,1,1,2,4,Qt::AlignLeft);
		connect( _legend_text_fld, SIGNAL(textChanged(const QString &)) ,this, SLOT(setLegendText(const QString &)) );
	
		setSlotColour(false);
		_legend_text_fld->setPalette(GUIpalette);
		_auto_legend_fld = new QCheckBox("Auto-legend");
		_auto_legend_fld->setChecked(true);
		data_legend_layout->addWidget(_auto_legend_fld,1,5,Qt::AlignCenter);
		connect(_auto_legend_fld, SIGNAL(toggled(bool)), this, SLOT(setAutoLegend(bool)) );

// 		data_legend_layout->addRowSpacing(0,0);
// 		data_legend_layout->addRowSpacing(1,_legend_text_fld->height());
// 		data_legend_layout->addRowSpacing(2,0);

		data_legend_layout->setRowStretch(0,0);
		data_legend_layout->setRowStretch(1,1);
		data_legend_layout->setRowStretch(2,0);

		data_legend_groupbox->setLayout(data_legend_layout);

		data_top_layout->addWidget(data_legend_groupbox);

		data_type_groupbox = new QGroupBox(this);
		setDataTypeTitle();
		

// 		QWidget* data_type_widget=new QWidget(data_type_groupbox, "data_type_widget");

		QGridLayout *data_type_layout=new QGridLayout;
		data_type_layout->setSpacing(QSAS_profiles->current()->GUImargins*2.5);
		data_type_layout->setMargin(QSAS_profiles->current()->GUImargins*2.5);

		QPushButton *_type_btn = new QPushButton("Change Plot Type...");
		_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_type_btn,0,0,Qt::AlignLeft);
		connect(_type_btn, SIGNAL(pressed()), this, SLOT(launchTypeSelector()) );

		QPushButton *_frame_type_btn = new QPushButton("Change Frame Type...");
		_frame_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_frame_type_btn,0,2,Qt::AlignLeft);
		connect(_frame_type_btn, SIGNAL(pressed()), this->I_Plot->I_Panel->I_Frame->xtAxisTab, SLOT(launchFrameTypeSelector()) );
		QToolTip::add( _frame_type_btn, "Sets whether ALL plot items in frame will be TS or XY");
		
		
// 		data_type_layout->addRowSpacing(0,_type_btn->height());		

		data_type_layout->setColStretch(0,0);
		data_type_layout->setColStretch(1,1);
		data_type_layout->setColStretch(2,0);
		data_type_layout->setColStretch(3,5);

		data_type_groupbox->setLayout(data_type_layout);
		data_top_layout->addWidget(data_type_groupbox);
	

	}
	else if(plotType == QP_SPECTROGRAM)
	{
	
		QLabel* z_label_fld  =new QLabel("Z: ");
		data_plot_layout->addWidget(z_label_fld, 1, 0, Qt::AlignRight);
		_z_data_fld = new QguWLinput(data_plot_groupbox, "z_data");
		_z_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget( _z_data_fld, 1, 1, 1, 2, Qt::AlignLeft);
		_z_data_fld->acceptObjectPlus=true;
		_z_data_fld->acceptTimeSubset=true;
		_z_data_fld->acceptTimeSlice=true;
		_z_data_fld->reduce3Dto1 = true;
		
		connect(_z_data_fld, SIGNAL(validated_OK()), this, SLOT(setZData())); 
		connect(_z_data_fld, SIGNAL(validated_WARN()), this, SLOT(setZTextOnly()));

		_XYtoggle = new QCheckBox("Swap default X and Y axes");
		data_plot_layout->addMultiCellWidget(_XYtoggle, 7, 7, 2, 2, Qt::AlignLeft);
		_XYtoggle->setChecked(false);
		if(I_Plot->frameType() == QP_TS) _XYtoggle->setEnabled(false);
		connect(_XYtoggle, SIGNAL(clicked()), this, SLOT(setXYswap()));
		
		QLabel* y_start_label_fld=new QLabel("Y: ");
		QFont it_font = QFont(qApp->font());
		it_font.setItalic(true);
		y_start_label_fld->setFont(it_font);
		data_plot_layout->addWidget(y_start_label_fld, 3, 0, Qt::AlignRight);
		_y_start_data_fld = new QguWLinput(data_plot_groupbox, "y_start_data");
		_y_start_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget( _y_start_data_fld, 3, 3, 1, 2, Qt::AlignLeft);
		_y_start_data_fld->acceptObjectPlus=true;
		_y_start_data_fld->acceptTimeSubset=true;
		_y_start_data_fld->acceptTimeSlice=true;

		connect(_y_start_data_fld, SIGNAL(validated_OK()), this, SLOT(setYStartData()));

		QLabel* y_end_label_fld = new QLabel("Y End: ");
		y_end_label_fld->setFont(it_font);
		data_plot_layout->addWidget(y_end_label_fld, 4, 1, Qt::AlignRight);
		_y_end_data_fld = new QguWLinput(data_plot_groupbox, "y_end_data");
		_y_end_data_fld->setMinimumWidth(300);
		data_plot_layout->addMultiCellWidget(_y_end_data_fld, 4, 4, 2, 2, Qt::AlignRight);
		_y_end_data_fld->acceptObjectPlus=true;
		_y_end_data_fld->acceptTimeSubset=true;
		_y_end_data_fld->acceptTimeSlice=true;

		connect(_y_end_data_fld, SIGNAL(validated_OK()), this, SLOT(setYEndData()));

		_x_start_label_fld = new QLabel("X: ");
		_x_start_label_fld->setFont(it_font);
		data_plot_layout->addWidget(_x_start_label_fld, 5, 0, Qt::AlignRight);
		_x_start_data_fld  =new QguWLinput(data_plot_groupbox, "x_start_data");
		_x_start_data_fld->setMinimumWidth(500);
		data_plot_layout->addMultiCellWidget( _x_start_data_fld, 5, 5, 1, 2, Qt::AlignLeft);
		_x_start_data_fld->acceptObjectPlus=true;
		_x_start_data_fld->acceptTimeSubset=true;
		_x_start_data_fld->acceptTimeSlice=true;
		if(I_Plot->frameType() == QP_TS) 
		{
			_x_start_label_fld->hide();
			_x_start_data_fld->hide();
		}
	
		connect(_x_start_data_fld, SIGNAL(validated_OK()), this, SLOT(setXStartData()));

		_x_end_label_fld = new QLabel("X End: ");
		_x_end_label_fld->setFont(it_font);
		data_plot_layout->addWidget(_x_end_label_fld, 6, 1, Qt::AlignRight);
		_x_end_data_fld = new QguWLinput(data_plot_groupbox, "x_end_data");
		_x_end_data_fld->setMinimumWidth(300);
		data_plot_layout->addMultiCellWidget( _x_end_data_fld, 6, 6, 2, 2, Qt::AlignRight);
		_x_end_data_fld->acceptObjectPlus=true;
		_x_end_data_fld->acceptTimeSubset=true;
		_x_end_data_fld->acceptTimeSlice=true;
		if(I_Plot->frameType() == QP_TS) {
			_x_end_label_fld->hide();
			_x_end_data_fld->hide();
		}

		connect(_x_end_data_fld, SIGNAL(validated_OK()), this, SLOT(setXEndData()));

		timeAsSec_toggle = new QCheckBox("Use time as sec from T1");
		data_plot_layout->addMultiCellWidget(timeAsSec_toggle, 7, 7, 1, 2, Qt::AlignLeft);
		timeAsSec_toggle->setChecked(false);
		connect(timeAsSec_toggle, SIGNAL(toggled(bool)), this, SLOT(useTimeAsSec(bool)) );
		if(I_Plot->frameType() == QP_TS) timeAsSec_toggle->hide();

		gap_toggle = new QCheckBox("Fill t gaps");
		data_plot_layout->addMultiCellWidget(gap_toggle, 7, 7, 1, 1, Qt::AlignLeft);
		gap_toggle->setChecked(false);
		if(I_Plot->frameType() == QP_XY) gap_toggle->hide();

// 		connect(gap_toggle, SIGNAL(clicked()), this, SLOT(setZData()));
		connect(gap_toggle, SIGNAL(clicked()), this, SLOT(gap_toggle_changed()));
		
// 		data_plot_layout->addRowSpacing(0,0);
// 		data_plot_layout->addRowSpacing(1,_z_data_fld->height());
// 		data_plot_layout->addRowSpacing(2,0);
// 		data_plot_layout->addRowSpacing(3,_y_start_data_fld->height());
// 		data_plot_layout->addRowSpacing(4,0);
// 		data_plot_layout->addRowSpacing(5,_x_start_data_fld->height());
// 		data_plot_layout->addRowSpacing(6,0);

		data_plot_layout->setRowStretch(0,0);
		data_plot_layout->setRowStretch(1,1);
		data_plot_layout->setRowStretch(2,0);
		data_plot_layout->setRowStretch(3,1);
		data_plot_layout->setRowStretch(4,0);
		data_plot_layout->setRowStretch(5,1);

		data_plot_groupbox->setLayout(data_plot_layout);

		data_top_layout->addWidget(data_plot_groupbox);

		data_type_groupbox=new QGroupBox(this);
		setDataTypeTitle();

// 		QWidget* data_type_widget=new QWidget(data_type_groupbox, "data_type_widget");

		QGridLayout *data_type_layout=new QGridLayout;
		data_type_layout->setSpacing(QSAS_profiles->current()->GUImargins*2.5);
		data_type_layout->setMargin(QSAS_profiles->current()->GUImargins*2.5);

		QPushButton *_type_btn = new QPushButton("Change Plot Type...");
		_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_type_btn,0,0,Qt::AlignLeft);
		connect(_type_btn, SIGNAL(pressed()), this, SLOT(launchTypeSelector()) );
	
		QPushButton *_frame_type_btn = new QPushButton("Change Frame Type...");
		_frame_type_btn->setMinimumWidth(100);
		data_type_layout->addWidget(_frame_type_btn,0,2,Qt::AlignLeft);
		connect(_frame_type_btn, SIGNAL(pressed()), this->I_Plot->I_Panel->I_Frame->xtAxisTab, SLOT(launchFrameTypeSelector()) );
		QToolTip::add( _frame_type_btn, "Sets whether ALL plot items in frame will be TS or XY");

// 		data_type_layout->addRowSpacing(0,_type_btn->height());

		data_type_layout->setColStretch(0,0);
		data_type_layout->setColStretch(1,1);
		data_type_layout->setColStretch(2,0);
		data_type_layout->setColStretch(3,5);
		data_plot_layout->setColStretch(4,10);
		
		data_type_groupbox->setLayout(data_type_layout);

		data_top_layout->addWidget(data_type_groupbox);

	}
	

	data_top_layout->addSpacing(4);
	data_top_layout->addStretch(10);

	return;
}
#endif

//---------------------------------------------

void QpDataTab::setDataTypeTitle()
{

	string thisType;
	if (I_PlotType == QP_TRACE)				thisType = "    Plot Type:   Trace            ";
	else if (I_PlotType == QP_CONST_XT && I_Plot->frameType() == QP_TS)		thisType = "    Plot Type:   Const T Trace  ";
	else if (I_PlotType == QP_CONST_XT && I_Plot->frameType() == QP_XY)		thisType = "    Plot Type:   Const X Trace  ";
	else if (I_PlotType == QP_CONST_Y)		thisType = "    Plot Type:   Const Y Trace  ";
	else if (I_PlotType == QP_SPECTROGRAM)	thisType = "    Plot Type:   Spectrogram     ";
	else;
		
	if(I_Plot->frameType() == QP_TS) 
	{
		thisType += "                    Frame Type:  Time Series";
	}
	else 
	{
		thisType += "                    Frame Type:  X-Y";
	}
	data_type_groupbox->setTitle(thisType.c_str());

}
// --------------------------------------------

void QpDataTab::useTimeAsSec(bool setOn)
{
	if( QuSlotDebug() ) cout << " QpDataTab::useTimeAsSec  " << endl;
	
	QpTraceItem *I_Trace = dynamic_cast<QpTraceItem*>(I_Plot);
	QpSpectrumItem *I_Spectrum = dynamic_cast<QpSpectrumItem*>(I_Plot);
	if(I_Trace)
	{
		if(setOn) {
			_x_label_fld->setText("T1: ");
			_x_data_fld->acceptObjectPlus=false;
			_x_data_fld->acceptTimeSubset=false;
			_x_data_fld->acceptTimeSlice=false;
			_x_data_fld->acceptTime=true;
			_x_data_fld->acceptTimeObject=true;
			_x_data_fld->reValidate();
			if( !(_y_data_fld->text().isEmpty()) )
			{
				_x_data_fld->setText(baseObjectText(_y_data_fld->text().latin1()));
				_x_data_fld->reValidate();
			}

		}
		else
		{
			_x_label_fld->setText("X: ");
			_x_data_fld->acceptObjectPlus=true;
			_x_data_fld->acceptTimeSubset=true;
			_x_data_fld->acceptTimeSlice=true;
			_x_data_fld->acceptTime=false;
			_x_data_fld->acceptTimeObject=false;
			_x_data_fld->setText("");
			_x_data_fld->reValidate();
		}
		
		// only tell list view to make changes to selected trace items if this
		// is the tab being user modified (avoids infinite loop from signals as hidden tabs change)

		if(this->isVisible() ) 
		{
			// propogate to selected items if not hard linked
			if(I_Trace->linkedPlots.size() == 0) 
			{
				I_Plot->I_ListView->setPlotOption(USE_TIME_AS_SEC, setOn);
			}
		}
	
		// pass on to linked items	
		if(!(I_Trace->I_ListView->isLink)){
			I_Trace->I_ListView->isLink = true; // stops infinite recursion;
			list <QpPlotItem*>::iterator it;
			for( it=I_Trace->linkedPlots.begin(); it != I_Trace->linkedPlots.end(); it++)
			{
				if( *it == I_Trace) continue;
				if((*it)->DataTab->timeAsSec_toggle->isChecked() != setOn) (*it)->DataTab->timeAsSec_toggle->toggle();
			}
			I_Trace->I_ListView->isLink = false;
		}
		I_Trace->updateAxes();
		
	}
	if(I_Spectrum)
	{
		if(setOn) {
			_x_start_label_fld->setText("T1: ");
			_x_start_data_fld->acceptObjectPlus=false;
			_x_start_data_fld->acceptTimeSubset=false;
			_x_start_data_fld->acceptTimeSlice=false;
			_x_start_data_fld->acceptTime=true;
			_x_start_data_fld->acceptTimeObject=true;
			_x_end_label_fld->hide();
			_x_end_data_fld->hide();
			_x_start_data_fld->reValidate();
			if( !(_z_data_fld->text().isEmpty()) )
			{
				_x_start_data_fld->setText(baseObjectText(_z_data_fld->text().latin1()));
				_x_start_data_fld->reValidate();
			}
		}
		else
		{
			_x_start_label_fld->setText("X: ");
			_x_start_data_fld->acceptObjectPlus=true;
			_x_start_data_fld->acceptTimeSubset=true;
			_x_start_data_fld->acceptTimeSlice=true;
			_x_start_data_fld->acceptTime=false;
			_x_start_data_fld->acceptTimeObject=false;
			_x_end_label_fld->hide();
			_x_end_data_fld->hide();
			_x_start_data_fld->setText("");
			_x_start_data_fld->reValidate();
		}
		
		// only tell list view to make changes to selected spectra items if this
		// is the tab being user modified (avoids infinite loop from signals as hidden tabs change)

		if(this->isVisible() ) 
		{
			// propogate to selected items if not hard linked
			if(I_Spectrum->linkedPlots.size() == 0) 
			{
				I_Plot->I_ListView->setPlotOption(USE_TIME_AS_SEC, setOn);
			}
		}
	
		// pass on to linked items	
		if(!(I_Spectrum->I_ListView->isLink)){
			I_Spectrum->I_ListView->isLink = true; // stops infinite recursion;
			list <QpPlotItem*>::iterator it;
			for( it=I_Spectrum->linkedPlots.begin(); it != I_Spectrum->linkedPlots.end(); it++)
			{
				if( *it == I_Spectrum) continue;
				if((*it)->DataTab->timeAsSec_toggle->isChecked() != setOn) (*it)->DataTab->timeAsSec_toggle->toggle();
			}
			I_Spectrum->I_ListView->isLink = false;
		}
		
		I_Spectrum->updateAxes();

	}
	

}
// --------------------------------------------

void QpDataTab::launchTypeSelector()
{
	// slot invoked by Change Type button
	Q_POPUP_MENU * typePopup = new Q_POPUP_MENU;
	typePopup->insertItem("Trace", QP_TRACE);
	typePopup->insertItem("Spectrogram", QP_SPECTROGRAM);
	typePopup->insertItem("Const X or T Trace", QP_CONST_XT);
	typePopup->insertItem("Const Y Trace", QP_CONST_Y);
	
	connect(typePopup, SIGNAL(activated(int)), I_Plot->I_ListView, SLOT(setPlotType(int)));

	typePopup->exec(QCursor::pos());
	
}

void QpDataTab::setLegendText(const QString &legend_text)
{
  	if( QuSlotDebug() ) cout << " QpDataTab::setLegendText = " << legend_text.latin1() << endl;

	if(_legend_text_fld->hasFocus()) _auto_legend_fld->setChecked(false);

	QpTraceItem *I_Trace = dynamic_cast<QpTraceItem*>(I_Plot);
	if( !I_Trace) return;
	
	// only tell list view to make changes to selected trace items if this
	// is the tab being user modified (avoids infinite loop from signals as hidden tabs change)

	if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_LEGEND_TEXT, 0, 0., legend_text);
}

void QpDataTab::setAutoLegend(bool autoOn)
{
	if ( QuSlotDebug() )
	{
		printf( "QpDataTab::setAutoLegend  %ld\n", autoOn );
		fflush(stdout);
	}

	if(autoOn)
	{
		_legend_text_fld->clearFocus();
		string legend_text;
		QpTraceItem * traceItem = dynamic_cast<QpTraceItem*>(I_Plot);
		if(traceItem) traceItem->getLegendText(legend_text);
		_legend_text_fld->setText(legend_text.c_str());
	}
	
	setSlotColour(!autoOn);
	_legend_text_fld->setPalette(GUIpalette);
	
	// only tell list view to make changes to selected frame items if this
	// is the tab being user modified (avoids infinite loop from signals as hidden tabs change)	
	if(this->isVisible() ) I_Plot->I_ListView->setTraceOption(LEGEND_AUTO_TEXT, 0, 0.0, "", autoOn);
	

} 


void QpDataTab::setShowLegend()
{
	if( QuSlotDebug() ) cout << " QpDataTab::setShowLegend = " <<  endl;

	QpTraceItem *I_Trace = dynamic_cast<QpTraceItem*>(I_Plot);
	if( !I_Trace) return;

  	// only tell list view to make changes to selected trace items if this
	// is the tab being user modified (avoids infinite loop from signals as hidden tabs change)
	int showLegend = (int) this->_show_legend_fld->isChecked();
	if(this->isVisible() ) I_Trace->I_ListView->setTraceOption(TRACE_SHOW_LEGEND, showLegend);
}

void QpDataTab::setXData()
{
	if( QuSlotDebug() ) cout << " QpDataTab::setXData  " << endl;

	QpTraceItem *I_Trace = dynamic_cast<QpTraceItem*>(I_Plot);
	if( !I_Trace) return;


	// is change to base object or component...
	QString BaseName = baseObjectText( _x_data_fld->text() ); // new value
	QString CompText = compText( _x_data_fld->text() ); // new value
	
	bool compOnly = false;
	if(old_BaseXName == BaseName && old_XComp != CompText) compOnly = true;
	
	old_BaseXName = QString(BaseName.latin1());
	old_XComp = QString(CompText.latin1());

	if(this->isVisible() ) {

		// propogate to selected items if not hard linked
		if(I_Trace->linkedPlots.size() == 0) 
		{
			if(compOnly) I_Trace->I_Panel->I_ListView->setTraceOption(TRACE_XCOMP, 0, 0.0, CompText);
			else I_Trace->I_Panel->I_ListView->setTraceOption(TRACE_XDATA, 0, 0.0, BaseName);
		}
	}
	
	// pass on the drop to linked items	: component selection should NOT propogate
	if(!(I_Trace->I_ListView->isLink)){
		I_Trace->I_ListView->isLink = true; // stops infinite recursion;
		list <QpPlotItem*>::iterator it;
		for( it=I_Trace->linkedPlots.begin(); it != I_Trace->linkedPlots.end(); it++)
		{
			if( *it == I_Trace) continue;
			if(!compOnly) (*it)->replaceXData(BaseName);
		}
		I_Trace->I_ListView->isLink = false;
	}
		
	I_Trace->updateAxes();

}

void QpDataTab::setYData()
{

	if( QuSlotDebug() ) cout << " QpDataTab::setYData  " << endl;

	// slot called when y data name field changed on widget
	QpTraceItem *I_Trace = dynamic_cast<QpTraceItem*>(I_Plot);
	if( !I_Trace) return;
	
	
	// is change to base object or component...
	QString BaseName = baseObjectText( _y_data_fld->text() ); // new value
	QString CompText = compText( _y_data_fld->text() ); // new value

	bool compOnly = false;
	if(old_BaseName == BaseName && old_Comp != CompText) compOnly = true;

	old_BaseName = QString(BaseName.latin1());
	old_Comp = QString(CompText.latin1());
	
	// if item uses data name change it
	if( !(I_Trace->nameLock) && !(_y_data_fld->text().isEmpty())) {
		I_Trace->setText(NAME_COL, _y_data_fld->text());
	}
	I_Trace->WLcontext = BaseName.latin1();
	if(this->isVisible() ) {

		// propogate to selected items 

		if(compOnly) I_Trace->I_Panel->I_ListView->setTraceOption(TRACE_YCOMP, 0, 0.0, CompText);
		else I_Trace->I_Panel->I_ListView->setTraceOption(TRACE_YDATA, 0, 0.0, BaseName);
	}
	
	// pass on the drop to linked items	: component selection should NOT propogate
	if(!(I_Trace->I_ListView->isLink)){
		bool discard = false; // used to find onceOnly linkage
		I_Trace->I_ListView->isLink = true; // stops infinite recursion;
		list <QpPlotItem*>::iterator it;
		for( it=I_Trace->linkedPlots.begin(); it != I_Trace->linkedPlots.end(); it++)
		{
			if( *it == I_Trace) continue;
			if(!compOnly) (*it)->replaceData(BaseName);
			if((*it)->text(LINK_COL).find("**") != -1) 
			{
				// once only linkage
				(*it)->setText(LINK_COL, "");
				(*it)->linkedPlots.clear();
				discard = true;
			}
		}
		if(discard) 
		{
			// also clear linkage on drop target
			I_Trace->setText(LINK_COL, "");
			I_Trace->linkedPlots.clear();
		}
		I_Trace->I_ListView->isLink = false;
	}
	
	I_Trace->updateAxes(); // also sets _ydo needed for legend
		
	if(_auto_legend_fld->isChecked())
	{
		string legend_text;
		I_Trace->getLegendText(legend_text);
		_legend_text_fld->setText(legend_text.c_str());
	}

}

void QpDataTab::setYTextOnly()
{
	// used for drops where data is not yet resolved
	if( QuSlotDebug() ) cout << " QpDataTab::setYTextOnly  " << endl;
	QpTraceItem *I_Trace = dynamic_cast<QpTraceItem*>(I_Plot);
	I_Trace->WLcontext = baseObjectText(_y_data_fld->text().latin1()).latin1();

	if(this->isVisible() ) {
		// propogate to selected items 
		I_Trace->I_Panel->I_ListView->setTraceOption(TRACE_YDATA, 0, 0.0, I_Trace->WLcontext);
	}

	if(!(I_Trace->I_ListView->isLink)){
		bool discard = false; // used to find onceOnly linkage
		I_Trace->I_ListView->isLink = true; // stops infinite recursion;
		list <QpPlotItem*>::iterator it;
		for( it=I_Trace->linkedPlots.begin(); it != I_Trace->linkedPlots.end(); it++)
		{
			if( *it == I_Trace) continue;
			(*it)->replaceData(I_Trace->WLcontext);
			if((*it)->text(LINK_COL).find("**") != -1) 
			{
				// once only linkage
				(*it)->setText(LINK_COL, "");
				discard = true;
			}

		}
		if(discard) 
		{
			I_Trace->setText(LINK_COL, "");
			I_Trace->linkedPlots.clear();
		}
		I_Trace->I_ListView->isLink = false;
	}

}


