#include "qsas_plplot_time_helpers.h"

#define MINUTE 60.
#define HOUR 60.*MINUTE
#define DAY 24.*HOUR
#define WEEK 7.*DAY
#define YEAR 365.25*DAY

/* Steve Schwartz additions to replace C mktime by mktimeutc. This avoids
*  problems due to mktime expecting its input in local time, which can also
*  depend on daylight saving etc.  The original recipe for determining the
*  offset from time_t epoch was found to be not always effective. */

time_t mktimeutc(struct tm *tm){
/* convert tm struct with entries in UTC to the calendar time secs from a
	*  fixed system epoch. Note that different implementations of C can use different
	*  epochs, so can't hardwire the value. Instead we calculate it. This is tedious
	*  but robust and portable. Relies on algorithms borrowed from NASA NSSDC cdf
	*  project to assemble a time since 0AD.
	*
	*  The returned time_t will be such that gmtime(returned value) gives tm
	*
	*  It needs to be robust against local time zones, daylight saving time, etc.
	*  and is necessary because there is a known issue with C-time handling. See
	*   http://www.cl.cam.ac.uk/~mgk25/time/c/ for example
	*
	*  History: Steve Schwartz s.schwartz@imperial.ac.uk 31 July 2008
	*
*  Bugs/features:
	*     tm.tm_isdst isn't checked; this is deliberate since tm is assumed UTC
	*     tm members must be within valid ranges; the C mktime will adjust its
*          members, but this one requires:
	*            (year+1900)  0->9999
	*            (mon+1)      1->12        note tm struct uses 0-11
	*            mday         1-31         no check against month
	*            hour         0->23
	*            min          0-59
	*            sec          0-59         no leap seconds
	*
	*     all other members (wday, yday, isdst) ignored. If you need them, you
	*     can call this routine to find the time_t, then pass that to gmtime to return
	*     tm. 
	*
*/
	time_t t;
	struct tm tm0;
	double tsecs, tsecs0;

	t=0;                     /* start somewhere */
	tm0 = *gmtime(&t);       /* and find its UTC */

	/* convert both tm structs to seconds since 0AD */

	tsecs  = mktimeEpochSecs(tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
	tsecs0 =  mktimeEpochSecs(tm0.tm_year+1900, tm0.tm_mon+1, tm0.tm_mday, tm0.tm_hour, tm0.tm_min, tm0.tm_sec);

	return (time_t)(tsecs - tsecs0);

}


/* *NSSDC cdf epoch algorithms - in modified form - follow                   */
/* ****************************************************************************
* JulianDay.
*    The year, month, and day are assumed to have already been validated.  This
* is the day since 0 AD (julian day may not be the proper term).
******************************************************************************/

static long JulianDay( long y, long m, long d )
{
	long jd;
	jd = (long)
			(367*y-7*(y+(m+9)/12)/4-3*((y+(m-9)/7)/100+1)/4+275*m/9+d+1721029);
	return (jd);
}

/******************************************************************************
* computeEPOCH.
*   Computes (and returns) an EPOCH in **SECS** since 0AD 
* based on its component parts.  -1.0
* is returned if an illegal component part is detected.
double computeEPOCH (year, month, day, hour, minute, second, msec)
long year, month, day, hour, minute, second;  cdf msec removed
******************************************************************************/

double mktimeEpochSecs( int year, int month, int day,
			int hour, int minute, int second )
{
	double epoch;
/*  printf("QEpochSecs got %d %d %d %d %d %d %d\n", year, month, day, hour,
	minute, second, msec);
*/
	if ((day < 1 || day > 31) ||
		    (month < 1 || month > 12) ||
		    (year < 0 || year > 9999) ||
		    (hour < 0 || hour > 23) ||
		    (minute < 0 || minute > 59) ||
		    (second < 0 || second > 59) )
		epoch = -1.0;
	else {
		epoch = (double) JulianDay (year, month, day) - 1721060;
		epoch = epoch * 24.0 + hour;
		epoch = epoch * 60.0 + minute;
		epoch = epoch * 60.0 + second;
	}
	return epoch;
}

void computeFormat(PLFLT delta_t, char* fmt_subticks, char* fmt_label, PLFLT * major_interval, PLINT * n_minor_intervals, int colonFormat)	
{
	char h_fmt[6], m_fmt[6], s_fmt[6];
	if(colonFormat)
	{
		sprintf(h_fmt, ":");
		sprintf(m_fmt, ":");
		sprintf(s_fmt, "");
	}
	else
	{
		sprintf(h_fmt, "#uh#d");
		sprintf(m_fmt, "#um#d");
		sprintf(s_fmt, "#us#d");
	}
	
	if(delta_t < 30.)
	{
		*major_interval=5.;
		*n_minor_intervals=5;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 2.*MINUTE) // 1 minute. 10 sec major, 2 sec minor
	{
		*major_interval=10.;
		*n_minor_intervals=5;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 5.*MINUTE) // 3 minutes. 30 sec major, 5 sec minor
	{
		*major_interval=30.;
		*n_minor_intervals=3;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 10.*MINUTE) // 10 minutes. 2 min major, 30 sec minor
	{
		*major_interval=1.*MINUTE;
		*n_minor_intervals=4;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 20.*MINUTE) // 10 minutes. 2 min major, 30 sec minor
	{
		*major_interval=4.*MINUTE;
		*n_minor_intervals=4;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 40.*MINUTE) // 10 minutes. 2 min major, 30 sec minor
	{
		*major_interval=5.*MINUTE;
		*n_minor_intervals=5;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 1.*HOUR)
	{
		*major_interval=10.*MINUTE;
		*n_minor_intervals=5;
		sprintf(fmt_subticks, "%%M%s%%S%s", m_fmt, s_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d %%H%s", h_fmt);
	}
	else if(delta_t < 3.*HOUR)
	{
		*major_interval=30.*MINUTE;
		*n_minor_intervals=6;
		sprintf(fmt_subticks, "%%H%s%%M%s", h_fmt, m_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d");
	}
	else if(delta_t < 6.*HOUR)
	{
		*major_interval=1.*HOUR;
		*n_minor_intervals=4;
		sprintf(fmt_subticks, "%%H%s%%M%s", h_fmt, m_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d");
	}
	else if(delta_t < 1.*DAY)
	{
		*major_interval=2.*HOUR;
		*n_minor_intervals=4;
		sprintf(fmt_subticks, "%%H%s", h_fmt);
		sprintf(fmt_label, "%%Y/%%m/%%d");
	}
	else if(delta_t < 3.*DAY)
	{
		*major_interval=12.*HOUR;
		*n_minor_intervals=6;
		sprintf(fmt_subticks, "%%d %%H%s", h_fmt);
		sprintf(fmt_label, "%%Y/%%m/");
	}
	else if(delta_t < 1.*WEEK)
	{
		*major_interval=1.*DAY;
		*n_minor_intervals=4;
		sprintf(fmt_subticks, "%%d");
		sprintf(fmt_label, "%%Y/%%m/");
	}
	else if(delta_t < 2.*WEEK)
	{
		*major_interval=1.*DAY;
		*n_minor_intervals=2;
		sprintf(fmt_subticks, "%%d");
		sprintf(fmt_label, "%%Y/%%m/");
	}
	else if(delta_t < 4.*WEEK)
	{
		*major_interval=1.*WEEK;
		*n_minor_intervals=7;
		sprintf(fmt_subticks, "%%m/%%d");
		sprintf(fmt_label, "%%Y/");
	}
	else //AUTOMATIC
	{
		*major_interval=0.;
		*n_minor_intervals=0;
		sprintf(fmt_subticks, "%%Y/%%m/%%d");
		sprintf(fmt_label, "");
	}
}

